 /*
  * rfc931() speaks a common subset of the RFC 931, AUTH, TAP and IDENT
  * protocols. The code queries an RFC 931 etc. compatible daemon on a remote
  * host to look up the owner of a connection. The information should not be
  * used for authentication purposes. This routine intercepts alarm signals.
  * 
  * Diagnostics are reported through stderr
  * 
  * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
  * modified by hvr
  */

static const char _rcsid[] = "$Id: rfc931.c,v 1.5.2.1 2002/01/02 21:28:42 soulshock Exp $";

/* System libraries. */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <glib.h>

#include "rfc931.h"
#include "inet_aton.h"

/* Local stuff. */

#define	RFC931_PORT 113	/* Semi-well-known port */
#define	ANY_PORT 0	/* Any local port will do */

#define RFC931_DEFAULT_TIMEOUT 10	
#define RFC931_DEFAULT_UNKNOWN "unknown"

gint rfc931_timeout = RFC931_DEFAULT_TIMEOUT;
const gchar *rfc931_unknown = RFC931_DEFAULT_UNKNOWN;

static jmp_buf timebuf;

/* fsocket - open stdio stream on top of socket */

static FILE *
fsocket (gint domain, gint type, gint protocol)
{
  gint s;
  FILE *fp = NULL;

  if ((s = socket (domain, type, protocol)) < 0)
    g_printerr (__FILE__ ":socket(): %s\n", g_strerror (errno));
  else if (!(fp = fdopen (s, "r+")))
    {
      g_printerr (__FILE__ ":fdopen(): %s\n", g_strerror (errno));
      close (s);
    }

  return fp;
}

/* bind_connect - bind both ends of a socket */

static gint
bind_connect (gint s, const struct sockaddr *local, 
	      const struct sockaddr *remote, gsize length)
{
  gint retval;

  if ((retval = bind (s, local, length)) < 0)
    g_printerr (__FILE__ ":bind(): %s\n", g_strerror (errno));
  else if ((retval = connect (s, remote, length)) < 0)
    {
#if defined (DEBUG)
      g_printerr (__FILE__ ":connect(): %s\n", g_strerror (errno));
#endif
    }
  
  return retval;
}

/* timeout - handle timeouts */

static void
timeout (gint sig)
{
  longjmp (timebuf, sig);
}

/* rfc931 - return remote user name, given socket structures */

static gint
_getprotobyname (const gchar name[])
{
  struct protoent *pe;

  g_return_val_if_fail (name != NULL, 0);

  setprotoent (0);
  
  pe = getprotobyname (name);

  endprotoent ();

  if (!pe)
    {
      g_printerr ("getprotobyname(%s): error occured\n", name);
      return 0;
    }

  return pe->p_proto;
}

static guint16
_getservbyname (const gchar name[], const gchar proto[],
		guint16 fallback)
{
  struct servent *se = NULL;

  g_return_val_if_fail (name != NULL, fallback);
  g_return_val_if_fail (proto != NULL, fallback);

  setservent (0);

  se = getservbyname (name, proto);

  endservent ();

  if (!se)
    {
      g_printerr ("getservbyname(%s): error occured\n", name);
      return fallback;
    }

  return g_ntohs (se->s_port);
}

const gchar *
rfc931 (const struct sockaddr_in *rmt_sin, const struct sockaddr_in *our_sin)
{
  guint rmt_port;
  guint our_port;
  struct sockaddr_in rmt_query_sin;
  struct sockaddr_in our_query_sin;
  static gchar user[256] = { 0, }; /* XXX */
  gchar buffer[512] = { 0, }; /* XXX */
  gchar *cp;
  const gchar *result; /* XXX */
  FILE *fp;

  /*
   * Use one unbuffered stdio stream for writing to and for reading from
   * the RFC931 etc. server. This is done because of a bug in the SunOS
   * 4.1.x stdio library. The bug may live in other stdio implementations,
   * too. When we use a single, buffered, bidirectional stdio stream ("r+"
   * or "w+" mode) we read our own output. Such behaviour would make sense
   * with resources that support random-access operations, but not with
   * sockets.
   */

  if ((fp = fsocket (PF_INET, SOCK_STREAM, _getprotobyname ("tcp") )) != 0)
    {
      setbuf (fp, (gchar *) 0);

      /*
       * Set up a timer so we won't get stuck while waiting for the server.
       */

      if (setjmp (timebuf) == 0)
	{
	  result = rfc931_unknown;
	  signal (SIGALRM, timeout);
	  alarm (rfc931_timeout);

	  /*
	   * Bind the local and remote ends of the query socket to the same
	   * IP addresses as the connection under investigation. We go
	   * through all this trouble because the local or remote system
	   * might have more than one network address. The RFC931 etc.
	   * client sends only port numbers; the server takes the IP
	   * addresses from the query socket.
	   */

	  our_query_sin = *our_sin;
	  our_query_sin.sin_port = g_htons (ANY_PORT);
	  rmt_query_sin = *rmt_sin;
	  rmt_query_sin.sin_port = g_htons (_getservbyname ("auth", "tcp", RFC931_PORT));

	  if (bind_connect (fileno (fp),
			    (struct sockaddr *) &our_query_sin,
			    (struct sockaddr *) &rmt_query_sin,
			    sizeof (our_query_sin)) >= 0)
	    {
	      /*
	       * Send query to server. Neglect the risk that a 13-byte
	       * write would have to be fragmented by the local system and
	       * cause trouble with buggy System V stdio libraries.
	       */

	      fprintf (fp, "%u,%u\r\n",
		       g_ntohs (rmt_sin->sin_port),
		       g_ntohs (our_sin->sin_port));
	      fflush (fp);

	      /*
	       * Read response from server. Use fgets()/sscanf() so we can
	       * work around System V stdio libraries that incorrectly
	       * assume EOF when a read from a socket returns less than
	       * requested.
	       */

	      if (fgets (buffer, sizeof (buffer), fp) != 0
		  && sscanf (buffer, "%u , %u : USERID :%*[^:]:%255s",
			     &rmt_port, &our_port, user) == 3
		  && g_ntohs (rmt_sin->sin_port) == rmt_port
		  && g_ntohs (our_sin->sin_port) == our_port)
		{

		  /*
		   * Strip trailing carriage return. It is part of the
		   * protocol, not part of the data.
		   */

		  if ((cp = strchr (user, '\r')))
		    *cp = 0;
		  result = user;
		}
	    }
	  alarm (0);
	}
      else
	result = rfc931_unknown;
      fclose (fp);
    }
  else
    result = rfc931_unknown;
  return result;
}
