/*
	$Id: site_vfs.c,v 1.7.2.12 2002/06/02 08:30:50 prime Exp $

	Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static const char _rcsid[] = "$Id: site_vfs.c,v 1.7.2.12 2002/06/02 08:30:50 prime Exp $";

#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#ifdef DIRCOMPLETION
#include <dirent.h>
#endif
#ifdef MACOSX
#include <machine/types.h>
#endif

#include <glib.h>

#include "site_vfs.h"
#include "pathutil.h"
#include "cdpath.h"
#include "userdb.h"
#include "pathutil.h"

/* public state information */
vfs_errno_t vfs_errno = VFS_OK;

/* private state information */
static gchar *__real_root_path = NULL;
static gchar *__cwd = NULL; 
static gchar *__chroot = NULL;
static gboolean __superuser = FALSE; /* filesystem godmode */

static gboolean strconvcasecmp(gchar* str1, gchar* str2) {
	gint i;
	gint size;
	gchar *cpy1;
	gchar *dummy1;
	gchar *cpy2;
	gchar *dummy2;

	if(strlen(str1)<strlen(str2)) {
		size=strlen(str1)*sizeof(gchar);
	}
	else {
		size=strlen(str2)*sizeof(gchar);
	}
	cpy1=(gchar *) malloc(size);
	cpy2=(gchar *) malloc(size);
	if((cpy1 == NULL) || (cpy2 == NULL)) {
		exit(-1);
	}
	memcpy(cpy1,str1,size);
	memcpy(cpy2,str2,size);
	dummy1=cpy1;
	dummy2=cpy2;
	for(i=0;i<(size / sizeof(gchar));i++) {
		*dummy1 = tolower(*dummy1);
		*dummy2 = tolower(*dummy2);
		if(*dummy1 != *dummy2) {
			if (isspace(*dummy1)) {
				*dummy1 = '_';
			}
			if (isspace(*dummy2)) {
				*dummy2 = '_';
			}
			if(*dummy1 != *dummy2) {
				return (*dummy1 - *dummy2);
			}
		}
		dummy1++;
		dummy2++;
	}
	free(cpy1);
	free(cpy2);
	return 0;
}

static gboolean
_filename_tabu (const gchar filename[])
{
  /* fixme: register tabu filenames dynamically */
  const gchar *_tabu_filenames[] = 
    {
      ".permissions",
      ".status",
      ".filled",
      ".locked",
      ".dircfg",
      NULL
    };

  const gchar *const *p = _tabu_filenames;

  while (*p) {
      if (!strcmp (*p, filename)) return TRUE;
      p++;
  }

  return FALSE;
}

const gchar *vfs_strerror (vfs_errno_t err) {
  switch (err)
    {
    case VFS_OK:
      return "OK";
    case VFS_UNKOWN:
      return "generic/unknown failure";
    case VFS_ACCESS:
      return "permission denied";
    case VFS_NOENT:
      return "file or directory not found";
    case VFS_EXIST:
      return "file or directory exists already";
    case VFS_NOTDIR:
      return "not a directory";
    case VFS_INVAL:
      return "invalid argument to function";
    default:
      g_return_val_if_fail (FALSE, "Ohoh... errno out of range");
    }
}

void
vfs_set_superuser (gboolean set)
{
  __superuser = set;
}

gint
vfs_chroot (const gchar path[])
{
  vfs_errno = VFS_OK;

  if (!path || !strlen (path))
    {
      g_free (__chroot);
      __chroot = NULL;
      return 0;
    }

  if (vfs_setcwd (path))
    return -1;

  g_free (__chroot);
  __chroot = g_strdup (vfs_getcwd ());
  return 0;
}

static gboolean 
_operation_permitted_p (const vfs_stat_t *st, guint32 op)
{
  if (__superuser)
    return TRUE;

  if (st->uid == userdb_get_uid ())
    return vfs_perm_get (st->mode, VFS_USER) & op ? TRUE : FALSE;

  if (userdb_have_group (st->gid))
    return vfs_perm_get (st->mode, VFS_GROUP) & op ? TRUE : FALSE;

  return vfs_perm_get (st->mode, VFS_OTHER) & op ? TRUE : FALSE;
}

static guint8
_perm_dir2file_helper (guint8 perm)
{
  guint8 result = 0;

  if (perm & VFS_DP_READ)
    result |= VFS_FP_READ;

  if (perm & VFS_DP_OVERWRITE)
    result |= VFS_FP_WRITE;

  if (perm & VFS_DP_CHMOD)
    result |= VFS_FP_CHMOD;

  return result;
}

static guint32 
_perm_dir2file (guint32 perm)
{
  guint32 p = vfs_perm_set (0, VFS_FLAG, VFS_DF_HIDDEN); /* ?? */

  p = vfs_perm_set (p, VFS_USER, 
		    _perm_dir2file_helper (vfs_perm_get (perm, VFS_USER)));

  p = vfs_perm_set (p, VFS_GROUP, 
		    _perm_dir2file_helper (vfs_perm_get (perm, VFS_GROUP)));

  p = vfs_perm_set (p, VFS_OTHER, 
		    _perm_dir2file_helper (vfs_perm_get (perm, VFS_OTHER)));

  return p;
}


/* permission bitset access utilities */

guint32 
vfs_perm_get (guint32 mode, vfs_perm_part_t part)
{
  g_return_val_if_fail (part == VFS_USER || part == VFS_GROUP
			|| part == VFS_OTHER || part == VFS_FLAG,
			0x00);

  return (mode >> part) & 0xff;
}

guint32
vfs_perm_set (guint32 mode, vfs_perm_part_t part, guint32 perm)
{
  g_return_val_if_fail (part == VFS_USER || part == VFS_GROUP
			|| part == VFS_OTHER || part == VFS_FLAG, mode);

  g_return_val_if_fail (perm <= 0xff, mode);

  /* mask away */
  mode &= ~((guint32) 0xff << part);

  mode |= (perm << part);

  return mode;
}

/****************************************************************************/

/* permission file */
typedef struct
{
  struct
  { 
    gint32 uid;
    gint32 gid;
    gint32 perm;
    gint32 dlc;
    gint32 dlct;
    gchar reserved[28];
  } dir;
  
  struct file_perm
  { 
    gchar name[124];
    gint32 crc;
    gint32 uid;
    gint32 gid;
    gint32 dlc;
    gint32 perm;
  } files[0];
} permfile_t;

static gchar *
_permfilename (const gchar dirname[])
{
  gchar *rp = NULL, *pfn = NULL;

  g_return_val_if_fail (dirname != NULL, NULL);

  rp = vfs_get_realpath (dirname);

  pfn = g_strconcat (rp, "/.permissions", NULL);

  g_free (rp);

  /* g_printerr ("_permfilename: %s -> %s\n", dirname, pfn); */

  return pfn;
}

/* file count <-> size conversion */
static gsize
_permfile_fcount2size (gint file_count)
{
  g_return_val_if_fail (file_count >= 0, 0);

  return sizeof (permfile_t) + (file_count * sizeof (struct file_perm));
}

static gint
_permfile_size2fcount (gsize size)
{
  gsize tmp;

  g_return_val_if_fail (size >= sizeof (permfile_t), 0);
  
  tmp = size - sizeof (permfile_t);
  
  if (tmp % sizeof (struct file_perm))
    g_printerr ("warning: permfile_size2fcount (%d): odd size\n", size);

  return tmp / sizeof (struct file_perm);
}

static permfile_t *
_permfile_load (const gchar permfilename[], gint *n)
{
  FILE *file;
  permfile_t *pf = NULL;
  struct stat st;
  gsize fcount;

  if (!(file = fopen (permfilename, "rt")))
    {
      g_printerr ("fopen(%s): %s\n", permfilename, strerror (errno));
      return pf;
    }

  if (fstat (fileno (file), &st) == -1)
    {
      g_printerr ("fstat(): %s\n", strerror (errno));
      fclose (file);
      return pf;
    }

  if (st.st_size < _permfile_fcount2size (0))
    {
      /* corrupted? */
      g_printerr ("_permfile_load: short .permissions\n");
      fclose (file);
      return pf;
    }

  fcount = _permfile_size2fcount (st.st_size);

  if (n)
    *n = fcount;

  pf = (permfile_t *) g_malloc0 (_permfile_fcount2size (fcount + 1));

  g_assert (pf != NULL);

  if (fread (pf, 1, st.st_size, file) != st.st_size)
    {
      g_printerr ("fread(): %s\n", strerror (errno));
      g_free (pf);
      pf = NULL;
    }

  fclose (file);
  return pf;
}

/****************************************************************************/


void
vfs_init (const gchar real_path[])
{
  g_assert (__real_root_path == NULL);
  g_assert (real_path != NULL);

  g_return_if_fail (real_path != NULL);

  __real_root_path = g_strdup (real_path);

  if (!__cwd)
    __cwd = g_strdup ("/");
}

gchar *
vfs_get_realpath (const gchar pathname[])
{
  gchar *tmp, *result;

  tmp = path_add (vfs_getcwd (), pathname);

  result = g_strconcat (__real_root_path, tmp, NULL);

  g_free (tmp);

  return result;
}

static gint
_stat (const gchar pathname[], struct stat *buf)
{
  gint rc = 0;
  gchar *rp = NULL;

  g_return_val_if_fail (buf != NULL, -1);
  g_return_val_if_fail (pathname != NULL, -1);

  rp = vfs_get_realpath (pathname);

  rc = stat (rp, buf);

  g_free (rp);

  return rc;
}

/* vfs_stat */
static gint _vfs_stat (const gchar _pathname[], vfs_stat_t *buf);

static gboolean 
_shrink_pathname (GString *pathname)
{
  gint idx;

  for (idx = pathname->len; idx >= 0; idx--)
    if (pathname->str[idx] == '/')
      break;

  if (!idx && pathname->len != 1)
    g_string_truncate (pathname, 1); /* exception for root */
  else
    g_string_truncate (pathname, (idx <= 0) ? 0 : idx);

  return pathname->len ? TRUE : FALSE;
}

gint
vfs_stat (const gchar _pathname[], vfs_stat_t *buf)
{
  gint rc;
  gchar *pathname;
  gboolean df_hidden = FALSE;

  if (!_pathname || !buf)
    {
      vfs_errno = VFS_INVAL;
      return -1;
    }

  pathname = path_add (vfs_getcwd (), _pathname);

  /* chroot check */
  if (__chroot && !path_child_p (__chroot, pathname))
    {
      vfs_errno = VFS_ACCESS;
      g_free (pathname);
      return -1;
    }

  {
    GString *subpath = g_string_new (pathname);
    gint sub_rc = 0;
    vfs_stat_t sub_st;

    while (_shrink_pathname (subpath))
      {
	if ((sub_rc = _vfs_stat (subpath->str, &sub_st)))
	  break;
	
	if (sub_st.type != VFS_DIR)
	  {
	    vfs_errno = VFS_NOTDIR;
	    sub_rc = -1;
	    break;
	  }

	if (!_operation_permitted_p (&sub_st, VFS_DP_CHDIR))
	  {
	    vfs_errno = VFS_ACCESS;
	    sub_rc = -1;
	    break;
	  }

	if (vfs_perm_get (sub_st.mode, VFS_FLAG) & VFS_DF_HIDDEN)
	  df_hidden = TRUE;
      }

    if (sub_rc)
      g_printerr ("(sub)vfs_stat (cwd=%s, %s) = %d (errno = %d)\n",
		  vfs_getcwd(), subpath->str, sub_rc, vfs_errno);

    g_string_free (subpath, TRUE);

    if (sub_rc)
      {
	g_free (pathname);
	return -1; /* vfs_errno was set already */
      }
  }

  rc = _vfs_stat (pathname, buf);

  if (df_hidden)
    buf->mode = vfs_perm_set (buf->mode, VFS_FLAG, 
			      vfs_perm_get (buf->mode, VFS_FLAG) | VFS_DF_HIDDEN_R);
  
  if (rc)
    g_printerr ("vfs_stat (cwd=%s, %s) = %d (errno = %d) (%d: 0x%8.8x %d.%d 0x%8.8x)\n",
		vfs_getcwd(), pathname, rc, vfs_errno, buf->cdpath_flag,
		buf->mode, buf->uid, buf->gid, buf->crc);

  g_free (pathname);
  return rc;
}

static gint
_vfs_stat (const gchar _pathname[], vfs_stat_t *buf)
{
  struct stat st;
  gchar *pathname;

  if (!_pathname || !buf)
    {
      vfs_errno = VFS_INVAL;
      return -1;
    }

  memset (buf, 0, sizeof (vfs_stat_t));

  pathname = path_add (vfs_getcwd (), _pathname);
  
  
  if (_stat (pathname, &st))
    {
      switch (errno)
	{
	case ENOENT:
	  vfs_errno = VFS_NOENT;
	  break;

	case ENOTDIR:
	  vfs_errno = VFS_NOTDIR;
	  break;

	default:
	  vfs_errno = VFS_UNKOWN;
	  break;
	}

      g_free (pathname);
      return -1;
    }

  /* stat()-call returned ok */

  if (S_ISREG (st.st_mode))
    buf->type = VFS_REG;
  else if (S_ISDIR (st.st_mode))
    buf->type = VFS_DIR;
  else				/* unexpected file type */
    {
      vfs_errno = VFS_UNKOWN;
      g_free (pathname);
      return -1;
    }

  buf->create_time = st.st_ctime;
  buf->change_time = st.st_mtime;
  buf->access_time = st.st_atime;

  switch (buf->type)
    {
      permfile_t *pf;
      gchar *pfn;
      gint n;
      const gchar *_cdpath;

    case VFS_REG:
      buf->size = st.st_size;
      buf->crc = 0;

      if ((_cdpath = cdpath_match (pathname)))
	{
	  cdpath_get (_cdpath, &buf->uid, &buf->gid, &buf->mode);
	  buf->mode = _perm_dir2file (buf->mode);
	  buf->cdpath_flag = TRUE;
	  break;
	}

      {
	gchar *dirname = g_dirname (pathname);
	
	pfn = _permfilename (dirname);
	g_free (dirname);

	pf = _permfile_load (pfn, &n);
	g_free (pfn);
      }
      
      if (pf)
	{
	  gchar *basename = g_basename (pathname); /* not-malloc'ed! */
	  gint idx;
	  
	  for (idx = 0; idx < n; idx++)
	    {
#if 0
	      g_printerr ("debug: %s %x %d %d %d %x\n",
			  pf->files[idx].name,
			  pf->files[idx].crc,
			  pf->files[idx].uid,
			  pf->files[idx].gid,
			  pf->files[idx].dlc,
			  pf->files[idx].perm);
#endif

	      if (!strcmp (basename, pf->files[idx].name))
		break;
	    }

	  buf->crc = pf->files[idx].crc;
	  buf->uid = pf->files[idx].uid;
	  buf->gid = pf->files[idx].gid;
	  buf->dlc = pf->files[idx].dlc;
	  buf->mode = pf->files[idx].perm;
	}

      g_free (pf);
      break;

    case VFS_DIR:

      if ((_cdpath = cdpath_match (pathname)))
	{
	  cdpath_get (_cdpath, &buf->uid, &buf->gid, &buf->mode);
	  buf->cdpath_flag = TRUE;
	  break;
	}

      pfn = _permfilename (pathname);
      pf = _permfile_load (pfn, NULL);
      g_free (pfn);

      if (pf)
	{
	  buf->uid = pf->dir.uid;
	  buf->gid = pf->dir.gid;
	  buf->mode = pf->dir.perm;
	}
      
      g_free (pf);
      break;

    default:
      g_assert_not_reached ();
      break;
    }

  g_free (pathname);
  return 0;
}


void is_link(char *fullp, char *split, int size)
{ 
    char link[256];
    char path[256];
    char *p;

    strncpy(path, fullp, sizeof(path) - 1);
    memset(split, 0, size);
    while(strcmp(path, "")) {
	memset(link, 0, sizeof(link));
	/* store link target in link */
	if (readlink(path, link, sizeof(link)) != -1) { 
	    if (link[0] != '/') {    /* target is an relative path */
		*(strrchr(path, '/') + 1) = 0; /* cut trailing / */
		strcat(path, link); /* append link to path */
	    } 
	    else { /* absolute path */
		strcpy(path, link);
	    }
	    strcat(path, split);
	    strcpy(split, "");
	} 
	else { /* error in readlink() */
	    strncpy(link, strrchr(path, '/'), sizeof(link) - 1);
	    strncat(link, split, sizeof(link) - strlen(link) - 1);
	    strncpy(split, link, sizeof(link) - 1);
	    *strrchr(path, '/') = 0;
	}
    }
    /* split var should'n have now sym links
       Now strip .. or . from split (Usefull when someone made a link
       ../../.... out of sitedir path) */
    while ((p = strstr(split,"/..")) != NULL) {
	*p = 0;
	strncpy(link, p+3, sizeof(link) - 1);
	*strrchr(split,'/') = 0;
	strcat(split,link);
    }
    while ((p = strstr(split,"/.")) != NULL) {
	*p=0;
	strcat(split,p+2);
    }
}


/* change the current working directory */
gint vfs_setcwd (const gchar new_path[]) 
{
    gchar *_new_path = NULL;
    vfs_stat_t st;
    char str[128];
    char wd[256];
    char *current;
    char resolved[PATH_MAX + 1];
#ifdef DIRCOMPLETION
    DIR *curdir;
    struct dirent *entry;
#endif

    g_return_val_if_fail (new_path != NULL, -1);

    _new_path = path_add (__cwd, new_path);

    if (vfs_stat (_new_path, &st)) {
#ifdef DIRCOMPLETION
	current = (char *) malloc((strlen(__cwd) + 3) * sizeof(char));
	sprintf(current, "./%s", __cwd);
	curdir = opendir(current);
	free(current);
	do {
	    entry = readdir(curdir);
	}
	while ((NULL != entry) && (0 != strconvcasecmp(entry->d_name, new_path)));
	closedir(curdir);
	if (NULL != entry) {
	    _new_path = path_add (__cwd, entry->d_name);
	    vfs_stat(_new_path, &st);
	} else {
#endif 
	    g_printerr ("vfs_setcwd (%s): %s\n", _new_path, vfs_strerror (vfs_errno));
	    return -1;
#ifdef DIRCOMPLETION
	}
#endif
    }

    if (st.type != VFS_DIR) {
	vfs_errno = VFS_NOTDIR;
	g_printerr ("vfs_setcwd (%s): %s\n", _new_path, vfs_strerror (vfs_errno));
	return -1;
    }
#define SYMLINK
#ifdef SYMLINK
    /* resolve symlinks */
    printf("SGV_DIR resolvesymlinks %s\n", _new_path);
    if (fgets(str, sizeof(str), stdin) && *(str + 1) == '1') {
	current = (char *) malloc((strlen(_new_path) + strlen(__real_root_path) + 1) * sizeof(char));
	sprintf(current, "%s%s", __real_root_path, _new_path);
	memset(wd, 0, sizeof(wd));
	is_link(current, wd, sizeof(wd) - 1);
	// readlink(current, wd, sizeof(wd) - 1); /* resolve links and allocate true path in wd */
	// realpath(wd, resolved);
	free(current);
	/* check if link targets outside the ftp root */
	if (strncmp(wd, __real_root_path, strlen(__real_root_path)) == 0) {
		_new_path = (char *) malloc (strlen(wd) - strlen(__real_root_path) + 1);
		if (_new_path) {
			strcpy(_new_path, wd + strlen(__real_root_path));
		}
	}
    }		       	 
#endif  

    if (!_operation_permitted_p (&st, VFS_DP_CHDIR)) {
	vfs_errno = VFS_ACCESS;
	g_printerr ("vfs_s	etcwd (%s): %s\n", _new_path, vfs_strerror (vfs_errno));
	g_free (_new_path);
	return -1;
    }

    /* 	everything's ok */
    vfs_errno = VFS_OK;
    g_free (__cwd);
    __cwd = _new_path;
    return 0;
}


const gchar *
vfs_getcwd (void)
{
  return __cwd;
}

gchar **
vfs_readdir (const gchar path[])
{
  vfs_stat_t st;
  gchar **result = NULL;
  
  if (vfs_stat (path, &st))
    return NULL;

  if (st.type != VFS_DIR)
    {
      vfs_errno = VFS_NOTDIR;
      return NULL;
    }

  if (!_operation_permitted_p (&st, VFS_DP_LIST))
    {
      vfs_errno = VFS_ACCESS;
      return NULL;
    }

  {
    DIR *dh = NULL;
    struct dirent *dentry = NULL;
    gsize entries = 0;
    gchar *rpath = vfs_get_realpath (path);

    if (!(dh = opendir (rpath)))
      {
	g_printerr ("vfs_readdir(%s): opendir(%s): %s\n",
		    path, rpath, strerror (errno));

	g_free (rpath);
	vfs_errno = VFS_UNKOWN;
	return NULL;
      }

    while ((dentry = readdir (dh)))
      {
	if (_filename_tabu (dentry->d_name))
	  continue;

	entries++;
	result = g_realloc (result, entries * sizeof (gchar *));

	result[entries-1] = g_strdup (dentry->d_name);
#if defined (DEBUG)
	g_printerr ("readdir = %s\n", dentry->d_name);
#endif
      }

    /* NULL-terminate list */
    entries++;
    result = g_realloc (result, entries * sizeof (gchar *));
    result[entries-1] = NULL;

    closedir (dh);
  }

  return result;
}

/* 
 * Local variables:
 *	c-file-style: "gnu"
 *	tab-width: 4
 *	indent-tabs-mode: t
 * End:
 */
