/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: serverd.c - main part of the ftpd / F4ADP protocol
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <sys/utsname.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef DES_ENCRYPTION
#include "des.h"
#endif
#ifdef CRYPT_UNISTD_H
#include <unistd.h>
#else
#ifdef CRYPT_CRYPT_H
#include <crypt.h>
#endif /* #ifdef CRYPT_CRYPT_H */
#endif /* #ifdef CRYPT_UNISTD_H */ 
#include "externsd.h"
#include "debug.h"
#include "f4adp_admin.h"
#include "f4adp_config.h"
#include "f4adp_group.h"
/*#include "f4adp_message.h"*/
#include "f4adp_server.h"
#include "f4adp_user.h"
#include "sections.h"
#include "ftpd.h"
#include "group.h"
#include "limits.h"
#include "log.h"
#include "registry.h"
#include "reject.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"
#include "f4adp_ipban.h"
#include "dupecheck.h"
#include "strlcpy.h"
#include "strlcat.h"

#include "permfile.h"

#define GROUP_RIGHTS 0
#define USER_RIGHTS  1

/* hvr */
static int
_read_permfile (char *filename, char **buffer, int *n)
{
  struct section_s *s;
  gint32 umask_root;

  s = get_section_by_path(filename);

  if(!s)
    return -1;
  
  umask_root = get_key_value_by_section(s, "umask_root", NULL, 0, &umask_root);

  return read_permfile (filename, buffer, n, umask_root);
}


/* SITE RIGHTS / SITE GROUP RIGHTS */
int print_user_rights(struct child_s *c, char **p, int n)
{
   char            str[256];
   struct user_s   *u = NULL;
   struct group_s  *g = NULL;
   gint64           r;
   int             mode;
   
   if (n != 2) {
      server_reply(c,"+ !HSyntax: SITE RIGHTS <user> or SITE GROUP RIGHTS <group>!0");
      return 0;
   }
   if (tolower(*p[0]) == 'g') {
      g = get_group_by_name(p[1]);
      if (!g) {
         server_reply(c, "-!BGroup does not exist.!0");
         return 0;
      }
      r = g->rights;
      mode = GROUP_RIGHTS;
   } 
   else {
      u = get_user_by_name(p[1]);
      if (!u) {
         server_reply(c, "-!BUser does not exist.!0");
         return 0;
      }
      if (c->user != u && (u->class & UF_INVISIBLE) && !siteop(c->user) && 
         !superuser(c->user)) {
         server_reply(c, "-!BUser does not exist.!0");
         return 0;
      }
      r = u->rights;
      mode = USER_RIGHTS;
   }

   if (mode == USER_RIGHTS) {
      if (!allowed(c->user, R_USER_READ, u) && c->user != u) {
         server_reply(c, "-!BNo permission to list user rights.!0");
         return 0;      
      }
   }
   else {
      if (!allowed(c->user, R_GROUP_LIST, u) && 
         !(groupop(c->user) && groupopof(c->user, g->gid))) {
         server_reply(c, "-!BNo permission to list group rights.!0");
         return 0;      
      }         
   }

   g_snprintf(str, sizeof(str), "+ !C.------------------------------------------------------------------------.!0\n");
   server_reply(c,str);

   if (mode == GROUP_RIGHTS) { 
      g_snprintf(str, sizeof(str), "+ !C|!3 !BOpenFTPD !HDEFAULT USER RIGHTS FOR MEMBERS OF GROUP: %-10.10s          !0!C|!0", g->name); 
      server_reply(c,str);
   }
   else {
      g_snprintf(str, sizeof(str), "+ !C|!3 !BOpenFTPD !HUSER RIGHTS FOR: %-10.10s                                   !0!C|!0", u->name); 
      server_reply(c,str);
   }
   g_snprintf(str, sizeof(str), "+ !C|------------------------------------------------------------------------|!0\n");
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 CHANGE STATE  !c[!H%c!c]!0  USER KICK    !c[!H%c!c]!0"
      "  SITEOP ADD   !c[!H%c!c]!0  EDIT IP     !c[!H%c!c] !C|!0",
      r & R_STATE ? '*' : '-', r & R_USER_KICK ? '*' : '-',
      r & R_USER_ADD_SO ? '*' : '-', r & R_USER_EDIT_IP ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 EDIT PORTS    !c[!H%c!c]!0  USER READ    !c[!H%c!c]!0"
      "  SITEOP EDIT  !c[!H%c!c]!0  EDIT GROUPS !c[!H%c!c]!0 !C|!0",
      r & R_PORTS ? '*' : '-', r & R_USER_READ ? '*' : '-',
      r & R_USER_EDIT_SO ? '*' : '-', r & R_USER_EDIT_GRPS ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 USE ADMINGUI  !c[!H%c!c]!0  USER DISABLE !c[!H%c!c]!0"
      "  SITEOP DEL   !c[!H%c!c]!0  EDIT CREDIT !c[!H%c!c]!0 !C|!0",
      r & R_GUI ? '*' : '-', r & R_USER_DISABLE ? '*' : '-',
      r & R_USER_DELETE_SO ? '*' : '-', r & R_USER_EDIT_CREDIT ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 USE MONITOR   !c[!H%c!c]!0  USER EDIT    !c[!H%c!c]!0"
      "  SITEOP READ  !c[!H%c!c]!0  EDIT STATS  !c[!H%c!c]!0 !C|!0",
      r & R_MONITOR ? '*' : '-', r & R_USER_EDIT ? '*' : '-',
      r & R_USER_READ_SO ? '*' : '-', r & R_USER_EDIT_STATS ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 DUMP FILES    !c[!H%c!c]!0  USER DELETE  !c[!H%c!c]!0"
      "  GROUPOP ADD  !c[!H%c!c]!0  EDIT RIGHTS !c[!H%c!c]!0 !C|!0",
      r & R_DUMP_FILES ? '*' : '-', r & R_USER_DELETE ? '*' : '-',
      r & R_USER_ADD_GO ? '*' : '-', r & R_USER_EDIT_RIGHTS ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 EDIT REGISTRY !c[!H%c!c]!0  USER ADD     !c[!H%c!c]!0"
      "  GROUPOP EDIT !c[!H%c!c]!0  EDIT INFO   !c[!H%c!c]!0 !C|!0",
      r & R_REGISTRY ? '*' : '-', r & R_USER_ADD ? '*' : '-',
      r & R_USER_EDIT_GO ? '*' : '-', r & R_USER_EDIT_INFO ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 EDIT MESSAGES !c[!H%c!c]!0  GROUP ADD    !c[!H%c!c]!0"
      "  GROUPOP DEL  !c[!H%c!c]!0  EDIT UID    !c[!H%c!c]!0 !C|!0",
      r & R_MESSAGES ? '*' : '-', r & R_GROUP_ADD ? '*' : '-',
      r & R_USER_DELETE_GO ? '*' : '-', r & R_USER_EDIT_UID ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 EDIT CONFIG   !c[!H%c!c]!0  GROUP EDIT   !c[!H%c!c]!0"
      "  GROUPOP READ !c[!H%c!c]!0  EDIT LOGINS !c[!H%c!c]!0 !C|!0",
      r & R_CONFIG ? '*' : '-', r & R_GROUP_EDIT ? '*' : '-',
      r & R_USER_READ_GO ? '*' : '-', r & R_USER_EDIT_LOGINS ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 FILE ACCESS   !c[!H%c!c]!0  GROUP LIST   !c[!H%c!c]!0"
      "  ANON ADD     !c[!H%c!c]!0  EDIT SPEEDS !c[!H%c!c]!0 !C|!0",
      r & R_FILEACCESS ? '*' : '-', r & R_GROUP_LIST ? '*' : '-',
      r & R_USER_ADD_ANON ? '*' : '-', r & R_USER_EDIT_SPEEDS ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 NUKER         !c[!H%c!c]!0  GROUP DELETE !c[!H%c!c]!0"
      "  TEMPLATE ADD !c[!H%c!c]!0  CHANGE PASS !c[!H%c!c]!0 !C|!0",
      r & R_NUKE ? '*' : '-', r & R_GROUP_DELETE ? '*' : '-',
      r & R_USER_ADD_TMP ? '*' : '-', r & R_USER_PASSWD ? '*' : '-');
   server_reply(c,str);

   g_snprintf(str, sizeof(str), "+ !C|!0 UNDUPER       !c[!H%c!c]!0  ALWAYS LOGIN !c[!H%c!c]!0  HOME LOCK    !c[!H%c!c]                  !C|!0", 
      r & R_UNDUPE ? '*' : '-', r & R_ALWAYS_LOGIN ? '*' : '-', 
      r & R_USER_LOCK ? '*' : '-');
   server_reply(c,str);

   if (mode == USER_RIGHTS && superuser(u)) {
      g_snprintf(str, sizeof(str), "+ !C|!0                                                                        !C|!0");
      server_reply(c,str);
      g_snprintf(str, sizeof(str), "+ !C| !HThis user has superuser status. Full rights are granted always!        !C|!0");
      server_reply(c,str);
   }
   else {
      g_snprintf(str, sizeof(str), "+ !C|!0                                                                        !C|!0");
      server_reply(c, str);
   }
   g_snprintf(str, sizeof(str), "+ !C|------------------------------------------------------------------------|!0\n");
   server_reply(c,str);
   g_snprintf(str, sizeof(str), "+ !C| !hRIGHTS AS HEX VALUE: !H%-#49qx !C|!0\n", (unsigned long long) r);
   server_reply(c,str);
   g_snprintf(str, sizeof(str), "+ !C`------------------------------------------------------------------------!c+!0");
   server_reply(c, str);
   server_reply(c, "+");
   return 0;
} 


/* SITE STATS */
int user_list_sections(struct child_s* c, char**p, int n)
{
   struct user_s    *u    = NULL;
   struct section_s *s    = NULL;
   struct sstat_s   *st   = NULL;
   int              x;
   int              i;
   int              found;
   char             str[256];


   if ( n != 1) {
      server_reply(c, "-!HSyntax: STATS [<user>]!0");
      return 0;
   }

   u = get_user_by_name(p[0]);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   if (c->user != u && (u->class & UF_INVISIBLE) && !siteop(c->user) && !superuser(c->user)) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }

   if (allowed(c->user,R_USER_READ,u) || c->user == u) {
      g_snprintf(str, sizeof(str), "+!C.--------------------------------------------------------------------------.!0\n");
      server_reply(c,str);
      g_snprintf(str, sizeof(str), "+!C|!3 !BOpenFTPD !HSTATS FOR USER: %-10.10s                                      !0!C|!0", u->name );
      server_reply(c,str);
      g_snprintf(str, sizeof(str), "+!C|--------------------------------------------------------------------------|!0\n");
      server_reply(c,str);
      g_snprintf(str, sizeof(str), "+!C| %74s!0", "|"); server_reply(c, str);           
      /*      for (i = 0; i < section_count; i++) { */
      for (i = section_count-1; i >= 0; i--) {  
         s = (section+i);
         if (s) {
            found = 0;
            for (x = i + 1 ; x < section_count; x++) {
               if ((section+x)->stat == s->stat)
                  found = 1;
            }
            if (!found) {
               if (strcmp(s->section, "default") || show_dfl_section == 1) {
                  st = get_stat_by_uid(s, u->uid);
                  if (st) {
                     if (get_key_value_by_section(s, "different_credit", NULL, 0, NULL)) {
                        g_snprintf(str, sizeof(str), "+!C| SECTION: !H%-37s !CCREDITS (MB): !H%10.0f  !C|!0", 
                                s->section, (float)st->credit / 0x100000);
                        server_reply(c, str);
                        g_snprintf(str, sizeof(str), "+!C| WEEKLY ALOTMENT (MB): !H%10.0f %39s !C|!0",
                                (float)st->wkly_alot / 0x100000, " ");
                     }
                     else g_snprintf(str, sizeof(str), "+!C| SECTION: !H%-38s                          !C|!0",
                                  s->section);
                     server_reply(c, str);
                     server_reply(c, "+!C|       !c|!C          UPLOADED          !c|!C         DOWNLOADED         !c|        !C|!0");
        	     server_reply(c, "+!C|       !c|!0 FILES !c|!0 SIZE IN MB!c|!0  SPEED !c|!0 FILES !c|!0 SIZE IN MB !c|!0 SPEED !c|!0 UL/DL  !C|!0");
        	     server_reply(c, "+!C|!c ------+-------+-----------+--------+-------+-----------+--------+------  !C|!0");
		     g_snprintf(str, sizeof(str), "+!C|!0 DAY   !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
                 	     st->ul_f_day, (float) st->ul_day / 0x100000,
                	     st->ul_tm_day > 0 ? (float) (st->ul_day / (st->ul_tm_day * 1024)) : 0.0,
                	     st->dl_f_day, (float) st->dl_day / 0x100000,
                	     st->dl_tm_day > 0 ? (float) (st->dl_day / (st->dl_tm_day * 1024)) : 0.0,
                	     st->dl_day > 0 ? (float) st->ul_day / st->dl_day : 0.0);
                     server_reply(c, str);
		     g_snprintf(str, sizeof(str), "+!C|!0 WEEK  !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
                 	     st->ul_f_week, (float) st->ul_week / 0x100000,
                	     st->ul_tm_week > 0 ? (float) (st->ul_week / (st->ul_tm_week * 1024)) : 0.0,
                	     st->dl_f_week, (float) st->dl_week / 0x100000,
                	     st->dl_tm_week > 0 ? (float) (st->dl_week / (st->dl_tm_week * 1024)) : 0.0,
                	     st->dl_week > 0 ? (float) st->ul_week / st->dl_week : 0.0);
        	     server_reply(c, str);
		     g_snprintf(str, sizeof(str), "+!C|!0 MONTH !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
                	     st->ul_f_month, (float) st->ul_month / 0x100000,
                	     st->ul_tm_month > 0 ? (float) (st->ul_month / (st->ul_tm_month * 1024)) : 0.0,
                	     st->dl_f_month, (float) st->dl_month / 0x100000,
                	     st->dl_tm_month > 0 ? (float) (st->dl_month / (st->dl_tm_month * 1024)) : 0.0,
                	     st->dl_month > 0 ? (float) st->ul_month / st->dl_month : 0.0);
        	     server_reply(c, str);
		     g_snprintf(str, sizeof(str), "+!C|!0 YEAR  !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
                	     st->ul_f_year, (float) st->ul_year / 0x100000,
                	     st->ul_tm_year > 0 ? (float) (st->ul_year / (st->ul_tm_year * 1024)) : 0,
                	     st->dl_f_year, (float) st->dl_year / 0x100000,
                	     st->dl_tm_year > 0 ? (float) (st->dl_year / (st->dl_tm_year * 1024)) : 0,
                	     st->dl_year > 0 ? (float) st->ul_year / st->dl_year : 0);
        	     server_reply(c,str);
		     g_snprintf(str, sizeof(str), "+!C|!0 TOTAL !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H%6d !c|!H %9.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
                	     st->ul_f_total, (float) st->ul_total / 0x100000,
                	     st->ul_tm_total > 0 ? (float) (st->ul_total / (st->ul_tm_total * 1024)) : 0,
                	     st->dl_f_total, (float) st->dl_total / 0x100000,
                	     st->dl_tm_total > 0 ? (float) (st->dl_total / (st->dl_tm_total * 1024)) : 0,
                	     st->dl_total > 0 ? (float) st->ul_total / st->dl_total : 0);
        	     server_reply(c, str);
        	     g_snprintf(str, sizeof(str), "+!C| %74s!0", "|"); server_reply(c, str);
                  }
               }
	    }   
         }
         usleep(5000);
      }
   }
   else { /* Not allowed */
      server_reply(c, "-!BYou dont have permission.!0\n");
      return 0; 
   } 
   server_reply(c, "+!C|!h Sections with a credit value use !Hseperated credits!h which are only valid  !C|!0");
   server_reply(c, "+!C|!h in this section. It's not possible to use them in other sections.        !C|!0");
   server_reply(c, "+!C|!h If multiple sections share the same stat file, only the first is listed. !C|!0");
   server_reply(c, "+!C`--------------------------------------------------------------------------+!0"); 
   server_reply(c, "+");
   return 0;
}


int user_credit_reduce(struct child_s* c, char** p, int n)
{ 
   gint64          size;
   struct user_s* u;

   if ((n != 2) || !sscanf(*(p+1),SI64,&size)) { 
      server_reply(c,"-!HSyntax: UCR <name> <size>!0");
      return 0;
   }
   u = get_user_by_name(*p);
   if (!u) { 
      server_reply(c,"-!BUser does not exist.!0");
      return 0;
   }
   if (!anonuser(u)) { 
      u->credit -= size * 1048576;
   }
   server_reply(c,"+");
   return 0;
}


int parse_command(char *p, char **param, int size, int max, int *n)
{
   int q;

   while (skipspace(&p)) {
      param[*n] = p;
      (*n)++;
      if (*n >= max)
         return 0;
      q = 0;
      while (!(q & 0x04)) {
         switch (*p) {
            case '"':
               if (q & 0x02) p++;
               else {
                  q ^= 0x01;
                  memmove(p, p + 1, strlen(p));
               }
               break;
            case '\'':
               if (q & 0x01) p++;
               else {
                  q ^= 0x02;
                  memmove(p, p + 1, strlen(p));
               }
               break;
            case '\\':
               memmove(p, p + 1, strlen(p));
               if (*p) p++;
               break;
            case '\t':
            case ' ':
               if (q) p++;
               else
                  q |= 0x04;
               break;
            case '\0':
               q |= 0x04;
               break;
            default:
               p++;
               break;
	 }
      }
      if (*p) {
         if (*n >= size) return 1;
         *(p++) = '\0';
      }
   }
   return 0;
}


int read_command(struct child_s *c, char **param, int size, int *n)
{
   int             l;
   char           *p;
   int             q;

   *n = 0;
   q = sizeof(c->buffer) - c->buffer_pos;
   if (q < 4) {
      c->buffer_overrun = 1;
      c->buffer_pos = 0;
      q = sizeof(c->buffer);
   }
   p = c->buffer + c->buffer_pos;
   if (!fgets(p, q, c->rdp)) {
      if (errno == EAGAIN) return 5;
      return feof(c->rdp) ? -2 : -1;
   }
   l = strlen(p);
   p = c->buffer;
   q = c->buffer_pos + l;
   if (!q) return 5;
   if (p[q - 1] != '\n') {
      c->buffer_pos += l;
      return 5;
   }
   c->buffer_pos = 0;
   if (c->buffer_overrun) {
      c->buffer_overrun = 0;
      return 3;
   }
   if (--q && (p[q - 1] == '\r')) q--;
   if (!q) return 2;
   p[q] = '\0';
   if (!skipspace(&p)) return 4;
   param[0] = p;
   *n = 1;
   if (skipnspace(&p)) *(p++) = '\0';
   else return 0;
   if (p == c->buffer + 5) {
      *n = 2;
      param[1] = p;
      c->buffer[3] = '\0';
      return 0;
   } 
   else if (param[0][0] == '#') {
      sscanf(param[0], "#%i#", &q);
      while (!isalpha(param[0][0])) param[0]++;
      parse_command(p, param, size, q + (*n), n);
      return 0;
   }
#ifdef DEBUG
   fprintf(debugf, "parse_command(%s,p,%i,%i,n)\n", p, size, size + 1);
#endif
   return parse_command(p, param, size, size + 1, n);
}


void server_reply(struct child_s * c, const char *reply)
{
   fprintf(c->wrp, "%s", reply);
   if (*(reply + strlen(reply) - 1) != '\n') fprintf(c->wrp, "\n");
 #ifdef DEBUG
   fprintf(debugf, "Server reply: %s", reply);
   if (*(reply + strlen(reply) - 1) != '\n') fprintf(debugf, "\n");
 #endif
#ifdef PIPEFIX
   if (reply_count > 10) 
   {
      usleep(100); /* wait some time to fix some timing problems !? */
      reply_count = 0;
   }
   else
   {
      reply_count++;
   }
#endif
   fflush(c->wrp);
}


void server_reply_lines(struct child_s * c, char *buffer)
{
   char *p = buffer;
   char line[1024];

   do {
      while (*p && (*p != '\n')) p++;
      if (*p) p++;
      line[0] = '+';
      memcpy(line + 1, buffer, p - buffer);
      line[p - buffer + 1] = '\0';
      server_reply(c, line);
      buffer = p;
   } while (*p);
}


int account_disable(struct child_s * c, char **p, int n)
{
   struct user_s  *u;

   u = get_user_by_name(*p);
   if (!u) {
      server_reply(c, "-!BUser does not exist!0");
      return 0;
   }
   if (!allowed(c->user, R_USER_DISABLE, u)) {
      server_reply(c, "-!BPermission denied!0");
      return 0;
   }
   if (u->class & UF_DISABLED) {
      server_reply(c, "-!BAccount is already disabled!0");
      return 0;
   }
   u->class |= UF_DISABLED;
   server_reply(c, "+!HAccount disabled!0");
   return 0;
}


int account_enable(struct child_s * c, char **p, int n)
{
   char           str[128];
   struct user_s  *u;

   u = get_user_by_name(*p);
   if (!u) {
      server_reply(c, "-!BUser does not exist!0");
      return 0;
   }
   if (!allowed(c->user, R_USER_DISABLE, u)) {
      server_reply(c, "-!BPermission denied!0");
      return 0;
   }
   if (!(u->class & UF_DISABLED)) {
      server_reply(c, "-!BAccount is already enabled!0");
      return 0;
   }
   if (has_group_free_slots(u->gid, str, sizeof(str)) == 1) {
      server_reply(c,str);
      return 0;
   }          
   u->class ^= UF_DISABLED;
   server_reply(c, "+!HAccount enabled!0");
   return 0;
}


int color(struct child_s * c, char **p, int n)
{
   /* check if color is disabled */
   if (!(c->user->class & UF_COLOR)) {
      c->user->class |= UF_COLOR;
      server_reply(c, "+!HColormode is now !Denabled!0");
      return 0;
   }
   /* check if color is enabled */
   if (c->user->class & UF_COLOR) {
      c->user->class ^= UF_COLOR;
      server_reply(c, "+Colormode is now disabled");
      return 0;
   }
   return 0;
}


int count(struct child_s * c, char **p, int n)
{
   char            str[128];
   struct child_s *u;
   int             nb = 0;
   int             i;

   u = child;
   for (i = 0; i < children; i++, u++) {
       if (u && u->user && !invisible(u->user)) nb++;
   }
   g_snprintf(str, sizeof(str), "+!H%i!0 user%s connected.", nb, nb > 1 ? "s" : "");
   server_reply(c, str);
   return 0;
}


/* added by primemover */
int check_rights(struct child_s *c, char** p, int n)
{
   char            str[128];
   struct user_s*  u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-!BUser does not exist!0");
      return 0;
   }

   if ((strcasecmp(p[1], "R_NUKE") == 0) && (u->rights & R_NUKE)) {
      g_snprintf(str, sizeof(str), "+!HPermission granted!0");
   }
   else if ((strcasecmp(p[1], "R_UNDUPE") == 0) && (u->rights & R_UNDUPE)) {
      g_snprintf(str, sizeof(str), "+!HPermission granted!0");
   }
   else g_snprintf(str, sizeof(str), "-!BNo permission!0");
   server_reply(c, str);
   return 0;
}


int server_version(struct child_s *c, char **p, int n)
{
   char            str[128];
   struct utsname  uts;   

   if (uname(&uts) == 0) {
      g_snprintf(str, sizeof(str), "+!HRunning OpenFTPD !H" VERSION " (%s %s on %s)!0", 
         uts.sysname, uts.release, uts.machine);
   }
   else {
      g_snprintf(str, sizeof(str), "+!HRunning OpenFTPD " VERSION "!0");
   }
   server_reply(c, str);
   return 0;
}


int server_uptime(struct child_s *c, char **p, int n)
{
   char            str[128];
   char            str2[128];
   time_t          t;
   double          diff = 0;

   t = time(NULL);
   diff = difftime(t, startup_time);
   sprint_difftime(str, 127, diff);
   g_snprintf(str2, sizeof(str2), "+!HServer is running for %s.!0", str);
   server_reply(c, str2);
   return 0;
}


int server_version_f4adp(struct child_s * c, char **p, int n)
{
   server_reply(c, "+102\n"); 
   return 0;
}


int dump_user_file(struct child_s * c, char **p, int n)
{
   if (save_user_file(*p) == -1) server_reply(c, "-!BError writing user file.!0");
   else server_reply(c, "+!HUser file saved successful.!0");
   return 0;
}


int dump_stat_files(struct child_s * c, char **p, int n)
{
   if (save_user_file(*p) == -1) server_reply(c, "-!BError writing stat files.!0");
   else server_reply(c, "+!HStat files saved successful.!0");
   return 0;
}


int dump_group_file(struct child_s * c, char **p, int n)
{
   if (save_group_file(*p) == -1) server_reply(c, "-!BError writing group file.!0");
   else server_reply(c, "+!HGroup file saved successful.!0");
   return 0;
}

/* Create home dir if parent dir exist */
/* RETURN: 0 on sucess, -1 on fail and str with the error */
int create_home(int id, int gid, char *dir, char *str, int size)
{
  FILE*         file;
  struct perm_s p;
  struct stat   st;
  char*         filename = NULL;

  if (*dir == '/') 
     dir++;

  /* test if dir exist */
  if (stat(dir, &st)) {
     if (mkdir(dir, 0777)) {
        g_snprintf(str, size - 1, "-Creating home dir: %s", strerror(errno));
        return -1;
     }
  }
  else if (!S_ISDIR(st.st_mode)) {
     strlcpy(str, "-Homedir exists but is not a directory", size);
     return -1; 
  } 
  filename = g_malloc(sizeof(char) * (strlen(dir) + strlen(permission_file) + 5));
  if (*(dir + strlen(dir) - 1) == '/') 
     sprintf(filename, "%s%s", dir, permission_file);
  else
     sprintf(filename, "%s/%s", dir, permission_file);
  if ((file = fopen(filename, "r+b")) == NULL)
     if ((file = fopen(filename, "w+b")) == NULL) {
        strlcpy(str, "-Couldn't create perm file in this home", size);
        g_free(filename);
        return -1;
     }
  if (fstat(fileno(file), &st) == -1) 
     st.st_size = 0;
  p.uid = id;
  p.gid = gid;
  p.perm = homeumask;
  if (st.st_size < sizeof(p)) {
     p.dlc = p.dlct = 0;
     memset(&p.reserved,0,sizeof(p.reserved));
  }
  if (!fwrite(&p,st.st_size < sizeof(p) ? sizeof(p) : sizeof(gint32)*3,1,file)) {
     g_snprintf(str, size-1, "-Writting permfile in homedir: %s",strerror(errno));
     fclose(file);
     g_free(filename);
     return -1;
  }
  g_free(filename);
  fclose(file);
  return 0;
   
}



gint32 user_umask(struct user_s * u, struct section_s *s)
{
   gint32  perm;

   if (superuser(u))
      get_key_value_by_section(s, "umask_root", NULL, 0, &perm);
   else if (anonuser(u))
      get_key_value_by_section(s, "umask_anon", NULL, 0, &perm);
   else {
      perm = get_grp_umask(u->gid);
      if (perm == 0)
      get_key_value_by_section(s, "umask_user", NULL, 0, &perm);
   }
   return perm;
}

char* get_filename(char *path)
{
   char           *p = NULL;

   p = path + strlen(path);
   while ((p > path) && (*p != '/'))
     p--;

   return *p == '/' ? p + 1 : p;
}

int file_chmod(struct child_s *c, char **p, int n)
{
   char          *buffer;
   char           cmd = 0;
   long           dperm = 0;
   struct file_s *f;
   struct file_s *f_start;
   char           filename[NAME_MAX];
   char          *fname_start;
   int            found;
   gint32         mode;
   struct perm_s *pe = NULL;
   int            who;

   memset(filename, 0, sizeof(filename));

   sscanf(*p, "%i %c " SI32, &who, &cmd, &mode);
   skipnspace(p);
   skipspace(p);
   skipnspace(p);
   skipspace(p);
   skipnspace(p);
   (*p)++;
   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);

   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (!found) {
         memmove(f + 1, f, (n - (f - f_start)) * sizeof(struct file_s));
         strlcpy(f->name, filename, sizeof(f->name));
         f->uid = pe->uid;
         f->gid = pe->gid;
         f->dlc = 0;
         f->perm = dirperm_to_fileperm(pe->perm);
         n++;
      }
      switch (cmd) {
      case '+':
      case '-':
         dperm = 0;
         if (who & 0x01) dperm |= mode << 16;
         if (who & 0x02) dperm |= mode << 8;
         if (who & 0x04) dperm |= mode;
         if (cmd == '+') f->perm |= dperm;
         else f->perm &= ~dperm;
         break;
      case '=':
         if (who & 0x01) f->perm = (f->perm & 0x00FFFF) | (mode << 16);
         if (who & 0x02) f->perm = (f->perm & 0xFF00FF) | (mode << 8);
         if (who & 0x04) f->perm = (f->perm & 0xFFFF00) | mode;
         break;
      }
      if (write_permfile(*p + 1, &buffer, n) == 0) {
         server_reply(c, "+");
         return 0;
      }
   }
   server_reply(c, "-");
   return 0;
}


int file_chown(struct child_s * c, char **p, int n)
{
   char           *buffer;
   struct file_s  *f;
   struct file_s  *f_start;
   char            filename[NAME_MAX];
   char           *fname_start;
   int             found;
   gint32           gid;
   struct perm_s  *pe;
   gint32           uid;

   memset(filename, 0, sizeof(filename));

   sscanf(*p, SI32 " " SI32, &uid, &gid);
   skipnspace(p);
   skipspace(p);
   skipnspace(p);
   skipspace(p);
   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-!BFilename too long.!0");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);

   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (!found) {
         memmove(f + 1, f, (n - (f - f_start)) * sizeof(struct file_s));
         strlcpy(f->name, filename, sizeof(f->name));
         f->uid = pe->uid;
         f->gid = pe->gid;
         f->dlc = 0;
         f->perm = dirperm_to_fileperm(pe->perm);
         n++;
      }
      if (uid != -1) f->uid = uid;
      if (gid != -1) f->gid = gid;
      if (write_permfile(*p + 1, &buffer, n) == 0) {
         server_reply(c, "+!HCHOWN was successful.!0");
         return 0;
      }
   }
   server_reply(c, "-!BCould not access permission file.!0");
   return 0;
}


int file_downloaded(struct child_s *c, char **p, int n)
{
   char           *buffer;
   struct file_s  *f;
   struct file_s  *f_start;
   char           filename[sizeof(f->name)];
   char           *fname_start;
   int            found;
   struct perm_s  *pe;

   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);
   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (!found) {
         memmove(f + 1, f, (n - (f - f_start)) * sizeof(struct file_s));
         strlcpy(f->name, filename, sizeof(f->name));
         f->uid = pe->uid;
         f->gid = pe->gid;
         f->dlc = 0;
         f->perm = dirperm_to_fileperm(pe->perm);
         n++;
      }
      f->dlc++;
      pe->dlct++;
      if (n) pe->dlc = pe->dlct / n;
      if (write_permfile(*p + 1, &buffer, n) == 0) {
         server_reply(c, "+");
         return 0;
      }
   }
   server_reply(c, "-");
   return 0;
}


int file_removed(struct child_s * c, char **p, int n)
{
   char           *buffer;
   struct file_s  *f;
   struct file_s  *f_start;
   char           filename[sizeof(f->name)];
   char           *fname_start;
   int            found;
   struct perm_s  *pe;

   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);
   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (found) {
         pe->dlct -= f->dlc;
         n--;
         if (n) pe->dlc = pe->dlct / n;
         else pe->dlc = 0;
         memmove(f, f + 1, (n - (f - f_start)) * sizeof(struct file_s));
         if (write_permfile(*p + 1, &buffer, n) == 0) {
            server_reply(c, "+");
            return 0;
         }
      } 
      else g_free(buffer);
   }
   server_reply(c, "-");
   return 0;
}


int file_renamed(struct child_s * c, char **p, int n)
{
   char           *buffer;
   long            dlc;
   struct file_s  *f;
   struct file_s  *f_start;
   char            filename[sizeof(f->name)];
   char           *fname_start;
   int             found;
   gint32           gid;
   struct perm_s  *pe;
   long            perm;
   gint32           uid;

   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);
   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (!found) {
         g_free(buffer);
         server_reply(c, "-");
         return 0;
      }
      pe->dlct -= f->dlc;
      dlc = f->dlc;
      perm = f->perm;
      uid = f->uid;
      gid = f->gid;
      n--;
      if (n) pe->dlc = pe->dlct / n;
      else pe->dlc = 0;
      memmove(f, f + 1, (n - (f - f_start)) * sizeof(struct file_s));
      if (write_permfile(*p + 1, &buffer, n) == 0) {
         server_reply(c, "+");
         fgets_to(*p, 256, c->rdp, 5);
         *(*p + strlen(*p) - 1) = '\0';
         fname_start = get_filename(*p);
         if (strlen(fname_start) >= sizeof(f->name)) {
            server_reply(c, "-");
            return 0;
         }
         strlcpy(filename, fname_start, sizeof(filename));
         strcpy(fname_start, permission_file);
         if (_read_permfile(*p, &buffer, &n) == 0) {
            pe = (struct perm_s *) buffer;
            f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
            f = find_name(f_start, filename, n, &found);
            if (!found) {
               memmove(f + 1, f, (n - (f - f_start)) * sizeof(struct file_s));
               strlcpy(f->name, filename, sizeof(f->name));
               n++;
            } 
	    else pe->dlct -= f->dlc;
            f->dlc = dlc;
            f->perm = perm;
            f->uid = uid;
            f->gid = gid;
            pe->dlct += dlc;
            if (n) pe->dlc = pe->dlct / n;
            if (write_permfile(*p + 1, &buffer, n) == 0) {
               server_reply(c, "+");
               return 0;
            }
         }
      }
   }
   server_reply(c, "-");
   return 0;
}


int file_uploaded(struct child_s * c, char **p, int n)
{
   char             *buffer;
   struct file_s    *f;
   struct file_s    *f_start;
   char              filename[sizeof(f->name)];
   char             *fname_start;
   int               found;
   struct perm_s    *pe;
   struct section_s *s = NULL;

   fname_start = get_filename(*p);
   if (strlen(fname_start) >= sizeof(f->name)) {
      server_reply(c, "-");
      return 0;
   }
   strlcpy(filename, fname_start, sizeof(filename));
   strcpy(fname_start, permission_file);

   s = get_section_by_path(*p);
   if(!s) {
      server_reply(c, "-");
      return 0;
   }

   if (_read_permfile(*p, &buffer, &n) == 0) {
      pe = (struct perm_s *) buffer;
      f_start = (struct file_s *) (buffer + sizeof(struct perm_s));
      f = find_name(f_start, filename, n, &found);
      if (!found) {
         memmove(f + 1, f, (n - (f - f_start)) * sizeof(struct file_s));
         strlcpy(f->name, filename, sizeof(f->name));
         n++;
      } 
      else pe->dlct -= f->dlc;
      f->dlc = 0;
      f->perm = dirperm_to_fileperm(user_umask(c->user, s));
      if (n) pe->dlc = pe->dlct / n;
      f->uid = c->user->uid;
      f->gid = c->user->gid;
      if (write_permfile(*p + 1, &buffer, n) == 0) {
         server_reply(c, "+");
         return 0;
      }
   }
   server_reply(c, "-");
   return 0;
}

int get_user_umask(struct child_s * c, char **p, int n)
{
   struct section_s *s = NULL;
   int mask;
   char  str[50];

   s = get_section_by_path(*p);
   if(!s) {
      server_reply(c, "-");
      return 0;
   }
   mask = user_umask(c->user, s);
   g_snprintf(str, sizeof(str), "+%i", mask);
   server_reply(c, str);
   return 0;
}

int groups(struct child_s * c, char **p, int n)
{
   gint32           gid;
   int             i;
   char            str[128];
   char            str2[1025];
   struct user_s  *u;

   memset(str, 0, sizeof(str));
   memset(str2, 0, sizeof(str2));

   u = get_user_by_name(p[0]);
   if (!u) {
      server_reply(c, "-!BUser does not exist!0");
      return 0;
   }
   if (n == 1) {
      if ((u != c->user) && !allowed(c->user, R_USER_EDIT_GRPS, u)) {
         server_reply(c, "-!BPermission denied!0");
         return 0;
      }
      strlcpy(str, "+!HPrimary group: !0", sizeof(str));

      get_groupname(u->gid, str2, sizeof(str2));
      strlcat(str, str2, sizeof(str));
      server_reply(c, str);

      strlcpy(str, "+!HSecondary groups:!0", sizeof(str));

      for (i = 0; i < 8; i++)
         if (u->gids[i] != -1) {
            strlcat(str, " ", sizeof(str));
            get_groupname(u->gids[i], str2, sizeof(str2));
	    strlcat(str, str2, sizeof(str));
         }
      if (strlen(str) == 18)
	strlcat(str, " (none)", sizeof(str));
      server_reply(c, str);
      return 0;
   }
   if ((n == 2) && ((**(p + 1) == '+') || (**(p + 1) == '-'))) {
      if (!allowed(c->user, R_USER_EDIT_GRPS, u)) {
         server_reply(c, "-!BPermission denied!0");
         return 0;
      }
      gid = get_gid(*(p + 1) + 1);
      if (gid == -1) {
         server_reply(c, "-!BGroup does not exist!0");
         return 0;
      }
      if (groupop(c->user) && (c->user->gid != gid)) {
         for (i = 0; i < 8; i++) {
            if (c->user->gids[i] == gid) i = 666;
         }
         if (i != 666) {
            server_reply(c, "-!BPermission denied!0");
            return 0;
         }
      }
      for (i = 0; i < 8; i++) {
         if (u->gids[i] == gid) {
            if (**(p + 1) == '-') {
               u->gids[i] = -1;
               server_reply(c, "+!HUser removed from group!0");
               return 0;
            } 
	    else {
               server_reply(c, "-!BUser is already in this group!0");
               return 0;
            }
         }
      }
      if (**(p + 1) == '+') {
         if (has_group_free_slots(gid, str, sizeof(str)) == 1) {
            server_reply(c,str);
            return 0;
         }
         for (i = 0; i < 8; i++) {
             if (u->gids[i] == -1) {
                u->gids[i] = gid;
                server_reply(c, "+!HUser added to group!0");
                return 0;
             }
         }
         server_reply(c, "-!BUser is already in maximum number of groups!0");
         return 0;
      }
      server_reply(c, "-!BUser is not in this group!0");
      return 0;
   }
   server_reply(c, "-!BSyntax error!0");
   return 0;
}

void prepare_msg(struct child_s *c, struct msg_s *msg) {
  char *buffer = NULL;
  int err;
  FILE *file = NULL;
  char filename[256];
  struct group_s *g = NULL;
  struct stat st;
  size_t bufferlen;
  int stat_err = 0;

  if(!msg || !msg->filename) return;

  /* prepare the filename */
  if (c->user) {
    g_snprintf(filename, sizeof(filename), "%s%s%s/%s", server_dir, DFL_MSGUDIR, c->user->name, msg->filename);
    err = stat (filename, &st);
    stat_err = errno;

    if (err) {
      if((g = get_group_by_gid(c->user->gid))) {
	g_snprintf(filename, sizeof(filename), "%s%s%s/%s", server_dir, DFL_MSGGDIR, g->name, msg->filename);
	err = stat (filename, &st);
	stat_err = errno;
      }
    }
  }
  else
    err = -1;

  if (err) {
    g_snprintf(filename, sizeof(filename), "%s%s%s", server_dir, DFL_MSGDIR, msg->filename);
    err = stat (filename, &st);
    stat_err = errno;
  }

  /* do the actual reading the file */
  /* file exists and has changed since last read */
  if (!err && st.st_mtime != msg->filedate) {
    buffer = msg->msg;
    bufferlen = msg->size;

    /* msg->msg points to NULL. allocate a buffer large enough to hold entire file + newline + NULL */
    if(!buffer) {
      bufferlen = st.st_size + 2 * sizeof(char);
      if((buffer = g_malloc(bufferlen))) {
	msg->msg = buffer;
	msg->size = bufferlen;
      }
    }

    if (buffer) {
      if((file = fopen (filename, "rb"))) {
#ifdef HAVE_FLOCK
	flock(fileno(file), LOCK_SH|LOCK_NB);
#endif
	msg->filedate = st.st_mtime;

	if(st.st_size + 2 * sizeof(char) > msg->size) {
	  bufferlen = msg->size + st.st_size + 2 * sizeof(char);

	  if(!(buffer = g_realloc(buffer, bufferlen))) {
	    /* we can't get a buffer large enough to hold the entire file
	     * we don't want truncation, since we allow execution of scripts from within message files
	     * so load default message and return
	     */
	    errlog("%s %d: g_realloc() failed (%d): %s", __FILE__, __LINE__, errno, strerror(errno));
	    if(msg->msg) g_free(msg);

	    msg->msg = msg->default_msg;
	    msg->size = 0;
	    msg->len = strlen(msg->msg);
	    return;
	  }
	}
	else
	  msg->len = fread (msg->msg, 1, st.st_size, file) / sizeof(char);

	/* terminate msg->msg buffer, since fread will not do that for us */
	*(msg->msg + msg->len) = '\0';

#ifdef HAVE_FLOCK
	flock(fileno(file), LOCK_UN);
#endif
	fclose (file);

	/* remove trailing whitespaces (space, newline, ..) */
	g_strchomp(msg->msg);
	/* append a trailing newline */
	msg->len = strlcat(msg->msg, "\n", msg->size);

	/* there's no room for newline. reallocate a new buffer to handle it
	 * msg->len holds the number of characters strlcat tried to create - 1
	 * so we can use that for the size of the new buffer
	 */
	if(msg->len >= msg->size) {
	  /* if g_realloc fails, load the default message and return */
	  bufferlen = msg->len + sizeof(char);
	  if((buffer = g_realloc(msg->msg, bufferlen))) {
	    msg->size = bufferlen;
	    strlcat(msg->msg, "\n", msg->size);
	  }
	  else {
	    errlog("%s %d: g_realloc() failed (%d): %s", __FILE__, __LINE__, errno, strerror(errno));
	    g_free(msg);
	    msg->msg = msg->default_msg;
	    msg->size = 0;
	    msg->len = strlen(msg->msg);
	    return;
	  }
	}
	msg->len = strlen(msg->msg);
      }
    }
  }

  if (!msg->msg || !msg->len) {
    if(msg->msg)
      g_free(msg->msg);
    msg->msg = msg->default_msg;
    msg->len = strlen (msg->msg);
    msg->size = 0;
  }
}

int section_list(struct child_s *c, char **p, int n)
{
   struct user_s *u = NULL;
   int coloriced=0;
   int i;
   int count_users=0;
   struct section_s *s = NULL;
   struct sstat_s *st = NULL;
   char str[256];
   char strcredit[16];
   gint32 fup, fdn;
   gint64 up, dn, sup, sdn, credit;

   s = get_section_by_name(p[0]);
   if (!s) {
      server_reply(c,"-!BSection not found!0");
      return 0;
   } 
   
   server_reply(c, "+ !C.------------------------------------------------------------------------.!0");
   g_snprintf(str, sizeof(str), "+ !C|!3!H %-9s %6s %8s %6s     %6s %8s %6s   %8s !0!C|!0", "USERNAME", "F_UL", "UPLOAD", "SPEED", "F_DL", "DOWNLOAD", "SPEED", "CREDITS");
      server_reply(c, str);
      server_reply(c, "+ !C|------------------------------------------------------------------------|!0");
   u = user;
   for (i = 0; i < user_count; i++, u++) {
      if ((u == c->user) || (allowed(c->user,R_USER_READ,u) &&
         (!(u->class & UF_INVISIBLE) || ((u->class & UF_INVISIBLE) && 
         (siteop(c->user) || superuser(c->user)))))) {
         count_users++;
         st = get_stat_by_uid(s, u->uid);
         if (!st) {
            g_snprintf(str, sizeof(str), "-Stats for user!H %s!0 in section !H%s!0 not found", 
                    u->name, s->section);
            server_reply(c,str);
            server_reply(c,"+");
            errlog(str);
            return 0;
         }
         if (get_key_value_by_section(s, "different_credit", NULL, 0, NULL))
            credit = st->credit;
         else
            credit = u->credit;
         g_snprintf(strcredit, sizeof(strcredit), "%10.0f", 
                  (float) credit / 0x100000);
         if (!strcasecmp(p[1],"day")) {
            fup = st->ul_f_day;
            fdn = st->dl_f_day;
            up = st->ul_day;
            dn = st->dl_day;
            sup = st->ul_tm_day;
            sdn = st->dl_tm_day;
         }
         else if (!strcasecmp(p[1],"week")) {
            fup = st->ul_f_week;
            fdn = st->dl_f_week;
            up = st->ul_week;
            dn = st->dl_week;
            sup = st->ul_tm_week;
            sdn = st->dl_tm_week;
         }
         else if (!strcasecmp(p[1],"month")) {
            fup = st->ul_f_month;
            fdn = st->dl_f_month;
            up = st->ul_month;
            dn = st->dl_month;
            sup = st->ul_tm_month;
            sdn = st->dl_tm_month;
         }
         else if (!strcasecmp(p[1],"year")) {
            fup = st->ul_f_year;
            fdn = st->dl_f_year;
            up = st->ul_year;
            dn = st->dl_year;
            sup = st->ul_tm_year;
            sdn = st->dl_tm_year;
         }
         else {
            fup = st->ul_f_total;
            fdn = st->dl_f_total;
            up = st->ul_total;
            dn = st->dl_total;
            sup = st->ul_tm_total;
            sdn = st->dl_tm_total;
         }
         if (coloriced) {
            coloriced = 0; 
            g_snprintf(str, sizeof(str), "+ !C| !H%-9s %6d %8.0f %6.1f     %6d %8.0f %6.1f %10.10s !C|!0", 
                    u->name, fup, (float) (up / 0x100000), 
                    sup > 0 ? (float) (up / (sup * 1024)) : 0.0,
                    fdn, (float) (dn / 0x100000),
                    sdn > 0 ? (float) (dn / (sdn * 1024)) : 0.0, 
                    u->limit ? strcredit: "unlimited");
         } 
         else {
            coloriced = 1;
            g_snprintf(str, sizeof(str), "+ !C| !0%-9s %6d %8.0f %6.1f     %6d %8.0f %6.1f %10.10s !C|!0", 
                    u->name, fup, (float) (up / 0x100000), 
                    sup > 0 ? (float) (up / (sup * 1024)) : 0.0,
                    fdn, (float) (dn / 0x100000),
                    sdn > 0 ? (float) (dn / (sdn * 1024)) : 0.0, 
                    u->limit ? strcredit: "unlimited");

         }
         server_reply(c,str);
      }
   }
   server_reply(c, "+ !C`------------------------------------------------------------------------+!0");
   server_reply(c, "+");
   return 0;
}


int send_msg(struct child_s * c, struct msg_s * msg)
{
   char            str[16];

   prepare_msg(c, msg);
   g_snprintf(str, sizeof(str), "%i", msg->len);
   server_reply(c, str);
   if (!msg->len) return 0;
   if (!fgets_to(str, sizeof(str), c->rdp, 5)) return 0;
   if (*str != '+') return 0;
   server_reply(c, msg->msg ? msg->msg : msg->default_msg);
   return 0;
}


/*  send some important vars to ftps (main() in ftps.c), matching 
 *  scanf and get_string calls on the other side needed! 
 *  order is important! 
 */
int send_startup_info(struct child_s * c, char **p, int n)
{
   char            str[128];

   g_snprintf(str, sizeof(str),"%lu %lu %i %i %i %ld %i %i %i %i %i %i %i %s\n",
           other_ip ? (unsigned long) inet_addr(other_ip) : 0, 
           passive_ip ? (unsigned long) inet_addr(passive_ip) : 0, 	   
	   login_retries,
           flags, showinfo, c->ip, c->sock, fconv_value, checktimeout,
           c->bouncer, bouncertimeout, sitegive, min_freemb, permission_file); 
   server_reply(c, str);
   server_reply(c, readme_file ? readme_file : "\n");
   server_reply(c, hidden_files ? hidden_files : "\n");
   server_reply(c, dupe_path ? dupe_path : "/");
   server_reply(c, ff_path ? ff_path : "/");
   server_reply(c, pre_download ? pre_download : "\n");
   server_reply(c, pre_upload ? pre_upload : "\n");
   server_reply(c, rejected_hops ? rejected_hops : "\n");
   server_reply(c, kickonnuke ? kickonnuke : "NONE\n");
   server_reply(c, server_dir);
   return 0;
}


int send_dupechk(struct child_s * c, char **p, int n)
{
   char            str[512];

   g_snprintf(str, sizeof(str), "+%s", dupechk ? dupechk : "");
   server_reply(c, str);
   return 0;
}


void child_logout(struct child_s * c, int exitcode)
{
   int             fd;
   int             n;
   char            reason[32];

   lim_return_session(c);
   lim_return_bandwidth(c);
   if (c->user && (c->pid != -1)) notify_watchers_logout(c);
   switch (exitcode) {
   case CL_NORMAL:
      strlcpy(reason, "quit", sizeof(reason)); 
      break;
   case CL_SHUTDOWN:
      strlcpy(reason, "shutdown", sizeof(reason));
      break;
   case CL_TIMEOUT:
      strlcpy(reason, "timeout", sizeof(reason));
      break;
   case CL_SOCKERR:
      strlcpy(reason, "socket error", sizeof(reason));
      break;
   case CL_KICK:
      strlcpy(reason, "kicked", sizeof(reason));
      break;
   case CL_PIPEERR:
      strlcpy(reason, "pipe error", sizeof(reason));
      break;
   case CL_ERROR:
      strlcpy(reason, "unknown error", sizeof(reason));
      break;
   case CL_FORCE:
      strlcpy(reason, "forced by shutdown", sizeof(reason));
      break;
   case CL_CLOSE:
      strlcpy(reason, "close", sizeof(reason));
      break;
   case CL_KILL:
      strlcpy(reason, "kill", sizeof(reason));
      break;
   case CL_IDENT:
      strlcpy(reason, "ident check failed", sizeof(reason));
      break;
   default:
      g_snprintf(reason, sizeof(reason), "code %i", exitcode);
   }
   server_log_local(c, LG_SECURITY, "LOGOUT (%s)", reason);
   shandler_del(fileno(c->rdp));
   if (c->admin) watch_del(c);
   fclose(c->wrp);
   if (c->rdp != c->wrp) fclose(c->rdp);
   fd = c->fd;
   n = c - child;
   if (n < --children) {
      memmove(c, c + 1, (children - n) * sizeof(struct child_s));
      update_children_shandler(c + 1, c);
      update_children_watch(c + 1, c);
   }
   if (!children) {
      if ((state & STATE_MASK) == STATE_SHUTDOWN) raise(SIGTERM);
   }
   /*
    * printf("children\n"); for (n=0; n<children; n++) printf("%i %i %i
    * %i\n",child[n].pid,
    * fileno(child[n].rdp),fileno(child[n].wrp),child[n].sock);
    */
}


int user_anonymous(struct child_s * c, char **p, int n)
{
   int             anonymous = 0;
   int             needident = 0;
   char            str[16];
   struct user_s   *u;

   u = n ? get_user_by_name(p[0]) : NULL;
   if (u) {
      if (anonuser(u) && !u->password[0]) {
         anonymous = 1;
         if (limit) {
             n = limit->max_sum ? limit->max_user : limit->max_anon;
             if (!n) server_reply(c, "-an");
         }
      } 
      else if (flag_get(flags, FLAG_IDENTMODE) && !c->ident[0]) {
         for (n = 0; n < 10 && !needident; n++) {
            if ((u->idents[n][0]) && (u->idents[n][0] != '*')) {
               needident = 1;
            }
         }
      }
   }
   g_snprintf(str, sizeof(str), "+%c%c", anonymous ? 'a' : 'n', needident ? 'i' : 'n');
   server_reply(c, str);
   return 0;
}


/* SITE INFO */
int user_information(struct child_s *c, char **p, int n)
{
   char str[128];
   char str2[128];
   int  i;
   struct user_s *u;

   memset(str, 0, sizeof(str));
   memset(str2, 0, sizeof(str2));

   if (n < 1) {
      server_reply(c, "-!BWrong syntax!0");
      return 0;
   }
   u = get_user_by_name(p[0]);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   if (n == 1) {
      g_snprintf(str, sizeof(str), "+User information of !H%s!0:!H %s!0\n", u->name, u->info);
      server_reply(c, str);
      return 0;
   }

   g_snprintf(str, sizeof(str), "%s ", p[1]);
   for(i = 2; i < n; i++) {
     g_snprintf(str2, sizeof(str2), "%s ", p[i]);
     if(strlcat(str, str2, sizeof(str)) > sizeof(str)) {
       /* string was truncated. no point in trying to write any more data */
       break;
     }
   }

   if (!strcmp(c->user->name, p[0]) || superuser(c->user) || (c->user->rights & R_USER_EDIT)) {
     if (strlen(str) < sizeof(c->user->info)) {
	strlcpy(u->info, str, sizeof(u->info));
	/* take out : and \t as are used as delimiters */
	for (i = 0; i < strlen(u->info); i++) {
	  if (u->info[i] == ':' || u->info[i] == '\t')
	    u->info[i] = ' ';
	}

	 g_snprintf(str2, sizeof(str2), "+Changed tagline for !H%s!0 to \"!H%s!0\"", p[0], str);
	 server_reply(c, str2);
      } 
      else server_reply(c, "-!BUser information is too long. (32 char maximum)!0");
   }
   else server_reply(c, "-!BYou are not allowed to change the info of other people!0");
   return 0;
}


/* SITE GINFO */
int group_information(struct child_s *c, char **p, int n)
{
   char str[128];
   char str2[128];
   int  i;
   struct group_s *g;

   memset(str, 0, sizeof(str));
   memset(str2, 0, sizeof(str2));

   if (n < 1) {
      server_reply(c, "-!BWrong syntax!0");
      return 0;
   }

   if(!(g = get_group_by_name(p[0]))) {
      server_reply(c, "-!BGroup does not exist.!0");
      return 0;
   }
   if (n == 1) {
      g_snprintf(str, sizeof(str), "+Group information of !H%s!0:!H %s!0\n", g->name, g->info);
      server_reply(c, str);
      return 0;
   }

   g_snprintf(str, sizeof(str), "%s ", p[1]);
   for(i = 2; i < n; i++) {
     g_snprintf(str2, sizeof(str2), "%s ", p[i]);
     if(strlcat(str, str2, sizeof(str)) > sizeof(str)) {
       /* string was truncated. no point in trying to write any more data */
       break;
     }
   }

   if (superuser(c->user) || siteop(c->user) || (groupop(c->user) && 
       groupopof(c->user, g->gid)) || (c->user->rights & R_GROUP_EDIT)) 
   {
      if (strlen(str) < sizeof(g->info)) {
	 strlcpy(g->info, str, sizeof(g->info));
	 g_snprintf(str2, sizeof(str2), "+Changed group tagline of !H%s!0 to \"!H%s!0\"", p[0], str);
	 server_reply(c, str2);
      } 
      else server_reply(c, "-!BGroup information is too long. (32 char maximum)!0");
   }
   else server_reply(c, "-!BYou are not allowed to change group information.!0");
   return 0;
}


char *get_hostname(const long ipaddr)
{
   struct sockaddr_in addr;
   struct hostent *hp = NULL;

   addr.sin_addr.s_addr = htonl(ipaddr);
   if (flag_get(flags, FLAG_RESOLVER)) {
      hp = gethostbyaddr((char *) &addr.sin_addr, 
           sizeof(addr.sin_addr), AF_INET);
   }

   if (hp)
     return hp->h_name;
   else
     return inet_ntoa(addr.sin_addr);
}

#define ULI_ALL      0x1FFF
#define ULI_ROOT     0x0001
#define ULI_NORMAL   0x0002
#define ULI_TEMPLATE 0x0004
#define ULI_ANON     0x0008
#define ULI_GROUPOP  0x0010
#define ULI_SITEOP   0x0020
#define ULI_DISABLED 0x0040
#define ULI_COLOR    0x0080
#define ULI_NOPASSCH 0x0100
#define ULI_LEECH    0x0200
#define ULI_GROUP    0x0400
#define ULI_NUKER    0x0800
#define ULI_UNDUPER  0x1000


char get_class_chr(struct user_s *u)
{ 
   switch (u->class & UC_MASK) { 
      case UC_ROOT    : return 'S';
      case UC_ANON    : return 'A';
      case UC_NORMAL  : return 'N';
      case UC_GROUPOP : return 'G';
      case UC_SITEOP  : return 'O';
      case UC_TEMPLATE: return 'T';
      default         : return '-';
   }
}


int get_class_str(struct user_s *u, char *out, size_t outlen) 
{
   if (!out) return -1;

   switch (u->class & UC_MASK) { 
      case UC_ROOT    : strlcpy(out, "SUPERUSER", outlen); break;
      case UC_ANON    : strlcpy(out, "ANONYMOUS", outlen); break;
      case UC_NORMAL  : strlcpy(out, "NORMAL",    outlen); break;
      case UC_GROUPOP : strlcpy(out, "GROUPOP",   outlen); break;
      case UC_SITEOP  : strlcpy(out, "SITEOP",    outlen); break;
      case UC_TEMPLATE: strlcpy(out, "TEMPLATE",  outlen); break;
      default         : strlcpy(out, "UNKNOWN",   outlen); break;
   }
   return 0;
}


int print_group_information(struct child_s *c, char **p, int n)
{
   struct group_s *g = NULL;
   char           str[256];

   if (n != 1) {
      server_reply(c, "-!BWrong number of arguments.!0");
      return 0;
   }

   g = get_group_by_name(p[0]);
   if (!g) {
      server_reply(c, "-!BGroup does not exist.!0");
      return 0;
   }

   server_reply(c, "+ !C.------------------------------------------------------"
      "------------------.!0");
   server_reply(c, "+ !C|!3 !BOpenFTPD !HGROUP INFORMATION                     "
      "                        !0!C|!0");
   server_reply(c, "+ !C|------------------------------------------------------"
      "------------------|!0");

   g_snprintf(str, sizeof(str), "+ !C|!0 GROUPNAME     : !H%-10.10s!0              GID          "
      ": !H%-3d!0             !C|!0" , g->name, g->gid);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 UL/DL RATIO   : !H%-2d!0                      CREDITS "
      "(MB) : !H%-6"I64"!0          !C|!0", g->limit, g->credit / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 LOGINS        : !H%-2d!0                      GROUP "
      "UMASK  : !H0x%06X!0        !C|!0", g->logins, g->umask_group);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 MAX LOGGED IN : !H%-2d!0                      SLOTS   "
      "     : !H%-2d!0              !C|!0", g->max_logins, g->max_slots);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 TAGLINE       : !H%-45.45s!0          !C|!0", g->info);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 IDENT         : !H%-45.45s!0          !C|!0", g->ident);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0%72s!C|!0","");
   server_reply(c, str);

   server_reply(c,"+ !C|!0                     !c|!0     MINIMUM !c|!0     "
      "MAXIMUM                        !C|!0");

   server_reply(c, "+ !C| !c--------------------+-------------+------------!0"
      "                        !C|!0");

   g_snprintf(str, sizeof(str), "+ !C|!0 DL SPEED LIMIT      !c|!H %6"I32" KB/s !c|!H %6"I32
      " KB/s !0                       !C|!0", g->dl_speed & 0xFFFF, 
      (g->dl_speed >> 16 )& 0xFFFF);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 UL SPEED LIMIT      !c|!0!H %6"I32" KB/s !c|!H %6"I32
      " KB/s !0                       !C|!0", g->ul_speed & 0xFFFF, 
      ((g->ul_speed >> 16 )& 0xFFFF));
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0%72s!C|!0", "");
   server_reply(c, str);

   server_reply(c, "+ !C`-------------------------------------------------------"
      "-----------------+!0");
   server_reply(c, "+");
   return 0;
}


int print_user_information(struct child_s *c, char **p, int n)
{
   struct user_s  *u = NULL;
   struct group_s *g = NULL;
   char           str[256];
   char           str2[256];
   char           userclass[16];
   char           creationtime[16];
   char           lastlogin[16];
   int            logins;
   int            j;

   if (n != 1) {
      server_reply(c, "-!BWrong number of arguments.!0");
      return 0;
   }

   u = get_user_by_name(p[0]);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   else if (u != c->user && (u->class & UF_INVISIBLE) && !superuser(c->user) 
           && !siteop(c->user)) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }

   if (!allowed(c->user, R_USER_READ, u)) {
      server_reply(c, "-!BNo permission.!0");
      return 0;
   }

   g = get_group_by_gid(u->gid);
   get_class_str(u, userclass, sizeof(userclass));

   if (u->last_login) {
      strftime(lastlogin, sizeof(lastlogin), "%Y-%m-%d", 
         localtime(&u->last_login));
   }
   else
      strlcpy(lastlogin, "never", sizeof(lastlogin));

   if (u->creation_time) {
      strftime(creationtime,sizeof(creationtime), "%Y-%m-%d", 
	 localtime(&u->creation_time));
   }
   else
      strlcpy(creationtime, "unknown", sizeof(creationtime));

   server_reply(c, "+ !C.-----------------------------------------------------"
      "-------------------.!0");
   server_reply(c, "+ !C|!3 !BOpenFTPD !HUSER INFORMATION                     "
      "                         !0!C|!0");
   server_reply(c, "+ !C|-----------------------------------------------------"
      "-------------------|!0");

   /* ---- USERNAME + PRIMARY GROUP ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 USERNAME     : !H%-10.10s!0         PRIMARY GROUP : "
      "!H%-10.10s!0           !C|!0", u->name, g ? g->name : "unknown");
   server_reply(c, str);

   /* ---- USERCLASS + UL/DL RATIO + CREDITS + LAST LOGIN ---- */
   if (u->limit == 0) {
      g_snprintf(str, sizeof(str), "+ !C|!0 USERCLASS    : !H%-10.10s!0         UL/DL RATIO   "
         ": !Hleech!0                !C|!0", userclass);
      server_reply(c, str);
      g_snprintf(str, sizeof(str), "+ !C|!0 CREDITS (MB) : !Hunlimited!0          LAST LOGIN  "
         "  : !H%-20.20s!0 !C|!0", lastlogin);
      server_reply(c, str);
   }
   else { /* limit != 0 */
      g_snprintf(str, sizeof(str), "+ !C|!0 USERCLASS    : !H%-10.10s!0         UL/DL RATIO   "
         ": !H%-3"I32"!0                  !C|!0", userclass, u->limit);
      server_reply(c, str);
      g_snprintf(str, sizeof(str), "+ !C|!0 CREDITS (MB) : !H%-10.0f!0         LAST LOGIN    :"
         " !H%-20.20s!0 !C|!0", (float) u->credit / 0x100000, lastlogin);
      server_reply(c, str);
   }

   /* ---- LOGINS + USER ID ---- */
   logins = (u->class & UL_MASK) >> 8;
   if (logins > 0) {
      g_snprintf(str, sizeof(str), "+ !C|!0 LOGINS       : !H%-3d!0                USER ID    "
         "   : !H%-5"I32"!0                !C|!0", 
         logins, u->uid);
   }
   else {
      g_snprintf(str, sizeof(str), "+ !C|!0 LOGINS       : !Hunlimited!0          USER ID     "
         "  : !H%-5"I32"!0                !C|!0", u->uid);
   }
   server_reply(c, str);

   /* ---- WEEKLY ALOTMENT ----*/
   g_snprintf(str, sizeof(str), "+ !C|!0 WEEKLY ALOTMENT (MB) : !H%-10.0f!0                                      !C|!0",
            (float) u->wkly_alot / 0x100000);
   server_reply(c, str);

   /* ---- TAGLINE ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 TAGLINE      : !H%-32.32s!0                        "
      "!C|!0", u->info);        
   server_reply(c, str);

   /* ---- HOMEDIR ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 HOMEDIR      : !H%-56.56s!0"
            "!C|!0", u->home);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);

   /* ---- NUKER + UNDUPER ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 CAN NUKE     : !H%-10.10s!0         CAN UNDUPE    : "
      "!H%-11.11s!0          !C|!0", (u->rights & R_NUKE) ? "YES" : "NO ", 
      (u->rights & R_UNDUPE) ? "YES" : "NO ");
   server_reply(c, str);

   /* ---- DISABLED + INVISIBLE ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 DISABLED     : !H%-10.10s!0         INVISIBLE     : "
      "!H%-11.11s!0          !C|!0", (u->class & UF_DISABLED) ? "YES" : "NO", 
      invisible(u) ? "YES" : "NO ");
   server_reply(c, str);

   /* ---- COLOR + LOCKED PASSWD ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 COLOR        : !H%-10.10s!0         LOCKED PASSWD : "
      "!H%-11.11s!0          !C|!0", coloruser(u) ? "YES" : "NO ", 
      no_pass_change(u) ? "YES" : "NO");
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str); 

   /* ---- GLOBAL STATS ---- */
   server_reply(c, "+ !C|       !c|!C          UPLOADED         !c|!C         "
      "DOWNLOADED        !c|        !C|!0");
   server_reply(c, "+ !C|       !c|!0 FILES !c|!0   MBytes !c|!0   KB/s !c|!0 "
      "FILES !c|!0   MBytes !c|!0   KB/s !c|!0 UL/DL  !C|!0");
   server_reply(c, "+ !C|!0 !c------+-------+----------+--------+------+------"
      "------+-------+------!0  !C|!0");

   g_snprintf(str, sizeof(str), "+ !C|!0 DAY   !c|!H%6d !c|!H %8.0f !c|!H %6.1f !c|!H%6d !c|!H"
      " %8.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
      u->ul_f_day, (float) u->ul_day / 0x100000,
      u->ul_tm_day > 0 ? (float) (u->ul_day / (u->ul_tm_day * 1024)) : 0.0,
      u->dl_f_day, (float) u->dl_day / 0x100000,
      u->dl_tm_day > 0 ? (float) (u->dl_day / (u->dl_tm_day * 1024)) : 0.0,
      u->dl_day > 0 ? (float) u->ul_day / u->dl_day : 0.0);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 WEEK  !c|!H%6d !c|!H %8.0f !c|!H %6.1f !c|!H%6d !c|!H"
      " %8.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
      u->ul_f_week, (float) u->ul_week / 0x100000,
      u->ul_tm_week > 0 ? (float) (u->ul_week / (u->ul_tm_week * 1024)) : 0.0,
      u->dl_f_week, (float) u->dl_week / 0x100000,
      u->dl_tm_week > 0 ? (float) (u->dl_week / (u->dl_tm_week * 1024)) : 0.0,
      u->dl_week > 0 ? (float) u->ul_week / u->dl_week : 0.0);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 MONTH !c|!H%6d !c|!H %8.0f !c|!H %6.1f !c|!H%6d !c|!H"
      " %8.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
      u->ul_f_month, (float) u->ul_month / 0x100000,
      u->ul_tm_month > 0 ? (float) (u->ul_month / (u->ul_tm_month * 1024)) : 0.0,
      u->dl_f_month, (float) u->dl_month / 0x100000,
      u->dl_tm_month > 0 ? (float) (u->dl_month / (u->dl_tm_month * 1024)) : 0.0,
      u->dl_month > 0 ? (float) u->ul_month / u->dl_month : 0.0);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 YEAR  !c|!H%6d !c|!H %8.0f !c|!H %6.1f !c|!H%6d !c|!H"
      " %8.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
      u->ul_f_year, (float) u->ul_year / 0x100000,
      u->ul_tm_year > 0 ? (float) (u->ul_year / (u->ul_tm_year * 1024)) : 0.0,
      u->dl_f_year, (float) u->dl_year / 0x100000,
      u->dl_tm_year > 0 ? (float) (u->dl_year / (u->dl_tm_year * 1024)) : 0.0,
      u->dl_year > 0 ? (float) u->ul_year / u->dl_year : 0.0);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL !c|!H%6d !c|!H %8.0f !c|!H %6.1f !c|!H%6d !c|!H"
      " %8.0f !c|!H %6.1f !c|!H %5.1f  !C|!0",
      u->ul_f_total, (float) u->ul_total / 0x100000,
      u->ul_tm_total > 0 ? (float) (u->ul_total / (u->ul_tm_total * 1024)) : 0.0,
      u->dl_f_total, (float) u->dl_total / 0x100000,
      u->dl_tm_total > 0 ? (float) (u->dl_total / (u->dl_tm_total * 1024)) : 0.0,
      u->dl_month > 0 ? (float) u->ul_month / u->dl_month : 0.0);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);
   
   /* ---- SPEED LIMITS ---- */
   server_reply(c, "+ !C|!0                     !c|!0     MINIMUM !c|!0     "
      "MAXIMUM                        !C|!0");
   server_reply(c, "+ !C|!0 !c--------------------+-------------+------------"
      "!0                        !C|!0");

   g_snprintf(str, sizeof(str), "+ !C|!0 DL SPEED LIMIT      !c|!H %6.1f KB/s !c|!H %6.1f KB/s"
      " !0                       !C|!0", (float) (u->dl_speed & 0xFFFF), 
      (float) ((u->dl_speed >> 16 )& 0xFFFF));
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 UL SPEED LIMIT      !c|!H %6.1f KB/s !c|!H %6.1f KB/s"
      " !0                       !C|!0", (float) (u->ul_speed & 0xFFFF), 
      (float) ((u->ul_speed >> 16)& 0xFFFF));
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);

   /* ---- NUKE STATS ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0       !c|!0 TIMES NUKED !c|!0 FILES NUKED !c|!0  MB "
      "NUKED  !c|!0 MB CREDITS NUKED      !C|!0");
   server_reply(c, str);
   
   server_reply(c, "+ !C|!0 !c------+-------------+-------------+------------+"
      "-----------------!0      !C|!0");

   g_snprintf(str, sizeof(str), "+ !C|!0 DAY   !c|!H %11d !c|!H %11d !c|!H %10.1f !c|!H    "
      "%13.1f      !C|!0", u->t_nuked_day, u->f_nuked_day, 
      (float) u->b_nuked_day / 0x100000, (float) u->c_nuked_day / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 WEEK  !c|!H %11d !c|!H %11d !c|!H %10.1f !c|!H    "
      "%13.1f      !C|!0",  u->t_nuked_week, u->f_nuked_week, 
      (float) u->b_nuked_week / 0x100000,
      (float) u->c_nuked_week / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 MONTH !c|!H %11d !c|!H %11d !c|!H %10.1f !c|!H    "
      "%13.1f      !C|!0", u->t_nuked_month, u->f_nuked_month, 
      (float) u->b_nuked_month / 0x100000, 
      (float) u->c_nuked_month / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 YEAR  !c|!H %11d !c|!H %11d !c|!H %10.1f !c|!H    "
      "%13.1f      !C|!0", u->t_nuked_year, u->f_nuked_year, 
      (float) u->b_nuked_year / 0x100000,
      (float) u->c_nuked_year / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL !c|!H %11d !c|!H %11d !c|!H %10.1f !c|!H    "
      "%13.1f      !C|!0", u->t_nuked_total, u->f_nuked_total, 
      (float) u->b_nuked_total / 0x100000,
      (float) u->c_nuked_total / 0x100000);
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);
   
   /* ---- SECONDARY GROUPS ---- */
   for (j = 0; j < 8; j++) {
      if (u->gids[j] != -1) {
         get_groupname(u->gids[j], str, sizeof(str));
         g_snprintf(str2, sizeof(str2), "+ !C|!0 SECONDARY GROUP %i: !H%-32s!0                 "
	    "   !C|!0", j, str);
         server_reply(c, str2);
      }
   }

   /* ---- IP RANGE ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); server_reply(c, str);
   for (j = 0; j < 10; j++) {
     if (memcmp(u->idents[j] + 12, "\0\0\0\0\0\0\0\0", 8)) {
       memset(str2, 0, sizeof(str2));
       print_ident(str2 + strlen(str2), u->idents[j]);
       g_snprintf(str, sizeof(str), "+ !C|!0 IP RANGE %i: !H%-40s!0                   !C|"
		  "!0", j, str2);
       server_reply(c, str);
       memset(str2, 0, sizeof(str2));
     }
   }
   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);

   /* ---- ACCOUNT CREATION ---- */
   g_snprintf(str, sizeof(str), "+ !C|!0 Account was created the !H%-10.10s!0 by !H%-10.10s"
      "                       !C|!0", creationtime, 
      *u->creator == 0 ? "unknown" : u->creator );
   server_reply(c,str);
   
   server_reply(c, "+ !C`-----------------------------------------------------"
      "-------------------+!0");
   server_reply(c, "+");
   return 0;
}


int print_user_list(struct child_s * c, char **p, int n)
{
   struct group_s   *g    = NULL;
   struct group_s   *g2   = NULL;
   struct user_s    *u    = NULL;
   int              i;
   int              lopt = ULI_ALL;
   int              m;
   char             str[256];
   char             lastlogin[20];
   char             firstlogin[20];
   char             credit[16];
   int              vipcount  = 0;
   int              usercount = 0;
   int              coloriced = -1;

   if (n > 2) {
      server_reply(c, "-!BWrong number of arguments.!0");
      return 0;
   }

   if (((c->user->rights & R_USER_READ) == 0) && !superuser(c->user)) {
      server_reply(c, "-!BNo permission.!0");
      return 0;
   }

   if (n > 0 && **p == '-') {
      lopt = 0;
      (*p)++;
      while (**p) {
	 switch (**p) {
	    case 'a': lopt |= ULI_ANON; break;
	    case 'd': lopt |= ULI_DISABLED;  break;
	    case 'n': lopt |= ULI_NORMAL;  break;
	    case 'c': lopt |= ULI_COLOR; break;
	    case 'r': 
	    case 's': lopt |= ULI_ROOT; break;
	    case 't': lopt |= ULI_TEMPLATE; break;
	    case 'g': lopt |= ULI_GROUPOP;  break;
	    case 'o': lopt |= ULI_SITEOP; break;
	    case 'p': lopt |= ULI_NOPASSCH;  break;
	    case 'l': lopt |= ULI_LEECH; break;
            case 'N': lopt |= ULI_NUKER; break;
            case 'u': lopt |= ULI_UNDUPER; break;
            case 'G': lopt |= ULI_GROUP; break;
	    default: lopt = ULI_ALL; break;;
	 }
         (*p)++;	 
      }
   }

   if (n == 2 && (lopt & ULI_GROUP)) {
      g = get_group_by_name(p[1]);
      if (!g) {
         server_reply(c, "-!BUnknown group.!0");
         return 0;      
      }
   }

   /* print header */
   server_reply(c, "+ !C.--------------------------------------------------"
      "----------------------.!0");
   g_snprintf(str, sizeof(str), "+ !C|!3!H C %-3s %-9s %-9s %8s %8s %10s %-3s %10s  !0!C|!0", 
      "UID", "USERNAME", "GROUP", "UPLOAD", "DOWNLOAD", "CREDITS", "LIM", 
      "LAST LOGIN");
   server_reply(c, str);
   server_reply(c, "+ !C|--------------------------------------------------"
      "----------------------|!0");

   u = user;
   for (i = 0; i < user_count; i++, u++) {
      m = 0;
      g2 = get_group_by_gid(u->gid);

      if ((anonuser(u) && (lopt & ULI_ANON)) ||
          (normaluser(u) && (lopt & ULI_NORMAL)) ||
          (superuser(u) && (lopt & ULI_ROOT)) ||
          (templateuser(u) && (lopt & ULI_TEMPLATE)) ||
          (groupop(u) && (lopt & ULI_GROUPOP)) ||
          (siteop(u) && (lopt & ULI_SITEOP)) ||
          ((u->class & UF_DISABLED) && (lopt & ULI_DISABLED)) ||
          ((u->class & UF_COLOR) && (lopt & ULI_COLOR)) ||
          ((u->rights & R_NUKE) && (lopt & ULI_NUKER)) ||
          ((u->rights & R_UNDUPE) && (lopt & ULI_UNDUPER)) ||
          (no_pass_change(u) && (lopt & ULI_NOPASSCH)) ||
          (u->limit == 0 && (lopt & ULI_LEECH))
         ) m = 1; 

      
      /* invisible users are only visible for siteops and themself */
      if ((u->class & UF_INVISIBLE) && !siteop(u) && !superuser(c->user) && (c->user != u)) m = 0;

      /* if group filter is on */

      if (g) {
         if (g2) {
	    if (strcmp(g->name, g2->name) != 0) m = 0;
	    else if (lopt == ULI_GROUP) m = 1;
         }
	 else m = 0;
      }

      if (m) {
         usercount++;
         if (superuser(u) || groupop(u) || siteop(u)) vipcount++;
         coloriced *= -1;

         if (u->last_login) {
            strftime(lastlogin, sizeof(lastlogin), "%Y-%m-%d", 
	    localtime(&u->last_login));
         } 
	 else
	   strlcpy(lastlogin, "<never>", sizeof(lastlogin));

         strftime(firstlogin,sizeof(firstlogin), "%Y-%m-%d", 
	 localtime(&u->creation_time));

         /* list all users */
         g_snprintf(credit, sizeof(credit), "%10.0f", (float) u->credit / 0x100000);
         g_snprintf(str, sizeof(str), (coloriced == 1) ? 
	    "+ !C|!H %c %3d %-9s %-9s %8.0f %8.0f %10.10s %3.0f %-10.10s  "
	    "!C|!0" : "+ !C|!h %c %3d %-9s %-9s %8.0f %8.0f %10.10s %3.0f "
	    "%-10.10s  !C|!0" , get_class_chr(u), u->uid, u->name, 
	    g2 ? g2->name : "<unknown>", (float) u->ul_total / 0x100000, 
	    (float) u->dl_total / 0x100000 , u->limit ? credit : "unlimited", 
	    (float) u->limit, lastlogin);
         server_reply(c, str);
      }
   }
   g_snprintf(str, sizeof(str), "+ !C|!0 %76s", "!C|!0"); 
   server_reply(c, str);

   g_snprintf(str, sizeof(str), "+ !C| !HListed users: !C%-4d !HPrivileged users: !C%-4d!0 "
      "%34s", usercount, vipcount, "!C|!0");
   server_reply(c, str);   

   server_reply(c, "+ !C`-----------------------------------------------------"
      "-------------------+!0");
   server_reply(c, "+");
   return 0;
}


int print_group_list(struct child_s * c, char **p, int n)
{
   struct group_s   *g    = NULL;
   int              i;
   int              coloriced = -1;
   int              groupcount = 0;
   char             str[256];

   /* print header */
   server_reply(c, "+ !C.------------------------------------------------"
      "------------------------.!0");
   server_reply(c, "+ !C|!3 !HNAME      GID LIM MBCRED    UMASK LG ML SL "
      "                      IDENT !0!C|!0");
   server_reply(c ,"+ !C|------------------------------------------------"
      "------------------------|!0"); 

   g = group;
   for (i = 0; i < group_count; i++, g++) {
      coloriced *= -1;
      groupcount++;
      g_snprintf(str, sizeof(str), coloriced == 1 ? 
         "+ !C|!H %-9s %3d %3d %6"I64" 0x%06X %2d %2d %2d %27s !C|!0" :
         "+ !C|!h %-9s %3d %3d %6"I64" 0x%06X %2d %2d %2d %27s !C|!0",
         g->name, g->gid, g->limit, g->credit / 0x100000, g->umask_group, g->logins,
         g->max_logins, g->max_slots, g->ident);
      server_reply(c, str);
   }
   server_reply(c, "+ !C|!0                                                   "
      "                     !C|!0");
   g_snprintf(str, sizeof(str), "+ !C| !HListed groups: !C%-4d %54s", groupcount,"|!0");
   server_reply(c, str);
   server_reply(c, "+ !C`-----------------------------------------------------"
      "-------------------+!0");
   g_snprintf(str, sizeof(str), "+ !C[!H LG = !0Default logins !HSL = !0Member slots !HML = !0Max total logins "
      "of members !C]!0");
   server_reply(c, str);
   server_reply(c, "+");
   return 0;
}


/*---------------------------------------------------------------------------
 fairly simple algorithm. Go through the users and compile stats. 
 Then output them in a nice format and we are done                                               
 since we do NOT want to see negative numbers in our site traffic if a person 
 has negative numbers in his stats we assume his stats are zero                                 
 ---------------------------------------------------------------------------*/
int print_traffic_stat(struct child_s * c, char **p, int n)
{
   int               i, x;
   float             day_up   = 0.0, day_down   = 0.0;
   float             month_up = 0.0, month_down = 0.0;
   float             week_up  = 0.0, week_down  = 0.0;
   float             year_up  = 0.0, year_down  = 0.0;
   float             total_up = 0.0, total_down = 0.0;
   float             day_av_speed_up   = 0.0, day_av_speed_down   = 0.0;
   float             week_av_speed_up  = 0.0, week_av_speed_down  = 0.0;
   float             month_av_speed_up = 0.0, month_av_speed_down = 0.0;
   float             year_av_speed_up  = 0.0, year_av_speed_down  = 0.0;
   float             total_av_speed_up = 0.0, total_av_speed_down = 0.0;
   float             day_dl_times   = 0.0, day_ul_times   = 0.0;
   float             week_dl_times  = 0.0, week_ul_times  = 0.0;
   float             month_dl_times = 0.0, month_ul_times = 0.0;
   float             year_dl_times  = 0.0, year_ul_times  = 0.0;
   float             total_dl_times = 0.0, total_ul_times = 0.0;
   char              str[256];
   struct user_s*    u;
   struct section_s* s;
   struct sstat_s*   st;

   u = user;
   for (i = 0; i < user_count; i++, u++) {
      if (u->ul_day >= 0) day_up = day_up + ((float) u->ul_day);
      else day_up = day_up + 0.0;
      if (u->dl_day >= 0) day_down = day_down + ((float) u->dl_day);
      else day_down = day_down + 0.0;
      if (u->ul_week >= 0) week_up = week_up + ((float) u->ul_week);
      else week_up = week_up + 0.0;
      if (u->dl_week >= 0) week_down = week_down + ((float) u->dl_week);
      else week_down = week_down + 0.0;
      if (u->ul_month >= 0) month_up = month_up + ((float) u->ul_month);
      else month_up = month_up + 0.0;
      if (u->dl_month >= 0) month_down = month_down + ((float) u->dl_month);
      else month_down = month_down + 0.0;
      if (u->ul_total >= 0) total_up = total_up + ((float) u->ul_total);
      else total_up = total_up +0.0;
      if (u->dl_total >= 0) total_down = total_down + ((float) u->dl_total);
      else total_down = total_down + 0.0;
      if (u->ul_year >= 0) year_up = year_up + ((float) u->ul_year);
      else year_up = year_up + 0.0;
      if (u->dl_year >= 0) year_down = year_down + ((float) u->dl_year);
      else year_down = year_down + 0.0;
      if ( u->ul_tm_day >= 0) {
         day_ul_times = day_ul_times + ((float) u->ul_tm_day);
      }
      else day_ul_times = day_ul_times + 0.0;
      if (u->dl_tm_day >= 0) {
         day_dl_times = day_dl_times + ((float) u->dl_tm_day);
      }
      else day_dl_times = day_dl_times + 0.0;
      if (u->ul_tm_week >= 0) {
         week_ul_times = week_ul_times + ((float) u->ul_tm_week);
      }
      else week_ul_times = week_ul_times + 0.0;
      if (u->dl_tm_week >= 0) {
         week_dl_times = week_dl_times + ((float) u->dl_tm_week);
      }
      else week_dl_times = week_dl_times + 0.0;
      if (u->ul_tm_month >= 0) {
         month_ul_times = month_ul_times + ((float) u->ul_tm_month);
      }
      else month_ul_times = month_ul_times + 0.0;
      if (u->dl_tm_month >= 0) {
         month_dl_times = month_dl_times + ((float) u->dl_tm_month);
      }
      else month_dl_times = month_dl_times + 0.0;
      if (u->ul_tm_year >= 0) {
         year_ul_times = year_ul_times + ((float) u->ul_tm_year);
      }
      else year_ul_times = year_ul_times + 0.0;
      if (u->dl_tm_year >= 0) {
         year_dl_times = year_dl_times + ((float) u->dl_tm_year);
      }
      else year_dl_times = year_dl_times + 0.0;
      if (u->ul_tm_total >= 0) {
         total_ul_times = total_ul_times + ((float) u->ul_tm_total);
      } 
      else total_ul_times = total_ul_times + 0.0;
      if (u->dl_tm_total >= 0) {
         total_dl_times = total_dl_times + ((float) u->dl_tm_total);
      } 
      else total_dl_times = total_dl_times + 0.0;
   }

   for ( i = 0, s = section; i < section_count; i++, s++) {
      if (get_key_value_by_section(s, "no_global_stats", NULL, 0, NULL)) {
	 for ( x = 0, st = s->stat; x < s->stat_count; x++, st++) {
            if ( st->ul_day >= 0) day_up = day_up + ((float) st->ul_day);
            else day_up = day_up + 0.0;
            if ( st->dl_day >= 0) day_down = day_down + ((float) st->dl_day);
            else day_down = day_down + 0.0;
            if ( st->ul_week >= 0) week_up = week_up + ((float) st->ul_week);
            else week_up = week_up + 0.0;
            if ( st->dl_week >= 0) { 
	       week_down = week_down + ((float) st->dl_week);
            } 
	    else week_down = week_down + 0.0;
            if ( st->ul_month >= 0){ 
               month_up = month_up + ((float) st->ul_month);
            } else{
               month_up = month_up + 0.0;
            }
            if ( st->dl_month >= 0){
               month_down = month_down + ((float) st->dl_month);
            } else {
               month_down = month_down + 0.0;
            }
            if ( st->ul_total >= 0){
               total_up = total_up + ((float) st->ul_total);
            } else {
               total_up = total_up +0.0;
            }
            if ( st->dl_total >= 0){
               total_down = total_down + ((float) st->dl_total);
            } else {
               total_down = total_down + 0.0;
            }
            if ( st->ul_year >= 0){
               year_up = year_up + ((float) st->ul_year);
            } else {
               year_up = year_up + 0.0;
            }
            if ( st->dl_year >= 0){
               year_down = year_down + ((float) st->dl_year);
            } else {
               year_down = year_down + 0.0;
            }
            if ( st->ul_tm_day >= 0){
               day_ul_times = day_ul_times + ((float) st->ul_tm_day);
            } else {
               day_ul_times = day_ul_times + 0.0;
            }
            if ( st->dl_tm_day >= 0){
               day_dl_times = day_dl_times + ((float) st->dl_tm_day);
            } else {
               day_dl_times = day_dl_times + 0.0;
            }
            if ( st->ul_tm_week >= 0){
               week_ul_times = week_ul_times + ((float) st->ul_tm_week);
            } else {
               week_ul_times = week_ul_times + 0.0;
            }
            if ( st->dl_tm_week >= 0){
               week_dl_times = week_dl_times + ((float) st->dl_tm_week);
            } else {
               week_dl_times = week_dl_times + 0.0;
            }
            if ( st->ul_tm_month >= 0){
               month_ul_times = month_ul_times + ((float) st->ul_tm_month);
            } else {
               month_ul_times = month_ul_times + 0.0;
            }
            if ( st->dl_tm_month >= 0){
               month_dl_times = month_dl_times + ((float) st->dl_tm_month);
            } else {
               month_dl_times = month_dl_times + 0.0;
            }
            if ( st->ul_tm_year >= 0){
               year_ul_times = year_ul_times + ((float) st->ul_tm_year);
            } else {
               year_ul_times = year_ul_times + 0.0;
            }
            if ( st->dl_tm_year >= 0){
               year_dl_times = year_dl_times + ((float) st->dl_tm_year);
            } else {
               year_dl_times = year_dl_times + 0.0;
            }
            if ( st->ul_tm_total >= 0) {
               total_ul_times = total_ul_times + ((float) st->ul_tm_total);
            } 
	    else {
               total_ul_times = total_ul_times + 0.0;
            }
            if ( st->dl_tm_total >= 0) {
               total_dl_times = total_dl_times + ((float) st->dl_tm_total);
            } 
	    else {
               total_dl_times = total_dl_times + 0.0;
            }
	 }
      } 
   }
   if (day_ul_times != 0.0) {
           day_av_speed_up = day_up / (day_ul_times * 1024);
   } else {
           day_av_speed_up = 0.0;
   }
   if (day_dl_times != 0.0) {
           day_av_speed_down = day_down / (day_dl_times * 1024);
   } else {
           day_av_speed_down = 0.0;
   }
   if (week_ul_times != 0.0) {
           week_av_speed_up = week_up / (week_ul_times * 1024);
   } else {
           week_av_speed_up = 0.0;
   }
   if (week_dl_times != 0.0) {
           week_av_speed_down = week_down / (week_dl_times * 1024);
   } else {
           week_av_speed_down = 0.0;
   }

   if (month_ul_times != 0.0) {
           month_av_speed_up = month_up / (month_ul_times * 1024);
   } else {
           month_av_speed_up = 0.0;
   }

   if (month_dl_times != 0.0) {
           month_av_speed_down = month_down / (month_dl_times * 1024);
   } else {
           month_av_speed_down = 0.0;
   }

   if (year_ul_times != 0.0) {
           year_av_speed_up = year_up / (year_ul_times * 1024);
   } else {
           year_av_speed_up = 0.0;
   }

   if (year_dl_times != 0.0) {
           year_av_speed_down = year_down / (year_dl_times * 1024);
   } else {
           year_av_speed_down = 0.0;
   }

   if (total_ul_times != 0.0) {
           total_av_speed_up = total_up / (total_ul_times * 1024);
   } else {
           total_av_speed_up = 0.0;
   }

   if (total_dl_times != 0.0) {
           total_av_speed_down = total_down / (total_dl_times * 1024);
   } else {
           total_av_speed_down = 0.0;
   }   
   server_reply(c, "+ !C.------------------------------------------------------------------------.!0");
   server_reply(c, "+ !C|!3 !BOpenFTPD !HTRAFFIC REPORT                                                !0!C|!0");
   server_reply(c, "+ !C|------------------------------------------------------------------------|!0");
   server_reply(c, "+ !C|!H DAILY  !C                                                                |!0");
   server_reply(c, "+ !C|!c ------                                                                 !C|!0");
   g_snprintf(str, sizeof(str), "+ !C|!0 UPLOAD   : !H%8.0f MB (%6.1f kB/s)!0  UPLOAD/DOWNLOAD RATIO:!H %8.2f  !C|!0",
           day_up / 0x100000, day_av_speed_up, day_down ? day_up / day_down : 0);
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 DOWNLOAD : !H%8.0f MB (%6.1f kB/s)!C %37s",
           day_down / 0x100000, day_av_speed_down, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL    : !H%8.0f MB (%6.1f kB/s)!C %37s", 
           day_up / 0x100000 + day_down / 0x100000, 
	   (day_ul_times + day_dl_times) ? (day_up + day_down) / 
	   ( (day_ul_times + day_dl_times) * 1024) : 0, "|!0");
   server_reply(c, str);
   server_reply(c, "+ !C|                                                                        |!0");
   server_reply(c, "+ !C|!H WEEKLY  !C                                                               |!0");
   server_reply(c, "+ !C|!c ------                                                                 !C|!0");
   g_snprintf(str, sizeof(str), "+ !C|!0 UPLOAD   : !H%8.0f MB (%6.1f kB/s)!0  UPLOAD/DOWNLOAD RATIO:!H %8.2f  !C|!0",
           week_up / 0x100000, week_av_speed_up, week_down ? week_up / week_down : 0);
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 DOWNLOAD : !H%8.0f MB (%6.1f kB/s)!C %37s",
           week_down / 0x100000, week_av_speed_down, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL    : !H%8.0f MB (%6.1f kB/s)!C %37s", 
           week_up / 0x100000 + week_down / 0x100000, 
	   (week_ul_times + week_dl_times) ? (week_up + week_down) / 
	   ( (week_ul_times + week_dl_times) * 1024) : 0, "|!0");
   server_reply(c, str);
   server_reply(c, "+ !C|                                                                        |!0");
   server_reply(c, "+ !C|!H MONTHLY!C                                                                |!0");
   server_reply(c, "+ !C|!c -------!C                                                                |!0");
   g_snprintf(str, sizeof(str), "+ !C|!0 UPLOAD   : !H%8.0f MB (%6.1f kB/s)!0  UPLOAD/DOWNLOAD RATIO:!H %8.2f !C%4s",
           month_up / 0x100000, month_av_speed_up, month_down ? month_up / month_down : 0, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 DOWNLOAD : !H%8.0f MB (%6.1f kB/s)!C %37s",
           month_down / 0x100000, month_av_speed_down, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL    : !H%8.0f MB (%6.1f kB/s)!C %37s", 
           month_up / 0x100000 + month_down / 0x100000, 
	   (month_ul_times + month_dl_times) ? (month_up + month_down) / 
	   ( (month_ul_times + month_dl_times) * 1024) : 0, "|!0");
   server_reply(c, str);
   server_reply(c, "+ !C|                                                                        |!0");
   server_reply(c, "+ !C|!H YEARLY!C                                                                 |!0");
   server_reply(c, "+ !C| !c------!C                                                                 |!0");
   g_snprintf(str, sizeof(str), "+ !C|!0 UPLOAD   :!H %8.0f MB (%6.1f kB/s)!0  UPLOAD/DOWNLOAD RATIO:!H %8.2f !C%4s",
           year_up / 0x100000, year_av_speed_up, year_down ? year_up / year_down : 0, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 DOWNLOAD :!H %8.0f MB (%6.1f kB/s)!C %37s",
           year_down / 0x100000, year_av_speed_down, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL    :!H %8.0f MB (%6.1f kB/s)!C %37s", 
           year_up / 0x100000 + year_down / 0x100000,  
	   (year_ul_times + year_dl_times) ? (year_up + year_down) / 
	   ( (year_ul_times + year_dl_times) * 1024) : 0, "|!0");
   server_reply(c, str);
   server_reply(c, "+ !C|                                                                        |!0");
   server_reply(c, "+ !C| !HTOTAL!C                                                                  |!0");
   server_reply(c, "+ !C| !c-----!C                                                                  |!0");
   g_snprintf(str, sizeof(str), "+ !C|!0 UPLOAD   :!H %8.0f MB (%6.1f kB/s)!0  UPLOAD/DOWNLOAD RATIO:!H %8.2f !C%4s",
           total_up / 0x100000, total_av_speed_up, total_down ? total_up / total_down : 0, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 DOWNLOAD : !H%8.0f MB (%6.1f kB/s)!C %37s",
           total_down / 0x100000, total_av_speed_down, "|!0");
   server_reply(c, str);
   g_snprintf(str, sizeof(str), "+ !C|!0 TOTAL    : !H%8.0f MB (%6.1f kB/s)!C %37s", 
           total_up / 0x100000 + total_down / 0x100000,  
	   (total_ul_times + total_dl_times) ? (total_up + total_down) / 
	   ( (total_ul_times + total_dl_times) * 1024) : 0, "|!0");
   server_reply(c, str);
   server_reply(c, "+ !C|                                                                        |!0");
   server_reply(c, "+ !C`------------------------------------------------------------------------+!0");
   server_reply(c, "+");
   return 0;
}


int user_list_loggedin(struct child_s * c, char **p, int n)
{
   int             i;
   char            str[256];
   struct child_s *u = child;

   memset(str, 0, sizeof(str));

   if (!allowed(c->user, R_USER_READ, NULL)) {
      server_reply(c, "-!BNo permission!0");
      return 0;
   }
   server_reply(c, "+name        pid mb_down mb_up'd rat lim mb_cred on_since"
                "    from_host");
   /*   u = child;*/
   for (i = 0; i < children; i++, u++) {
     if (u->user && (!n || !strcasecmp(u->user->name, *p))) {
         if (!allowed(c->user, R_USER_READ, u->user))
	   continue;

         g_snprintf(str, sizeof(str), "+%-9s%c%5ld %7" I64 " %7" I64 " ", u->user->name,
                 u->user->password[0] ? ' ' : '*', (long)u->pid,
                 u->user->dl_total / (1024 * 1024), u->user->ul_total / (1024 * 1024));
         if (u->user->dl_total) {
            g_snprintf(str + strlen(str), sizeof(str) - strlen(str) - 1, "%3.0f",
                    (float) u->user->ul_total * 100 / u->user->dl_total);
         }
         else
	   strlcat(str, "n/a", sizeof(str));
         g_snprintf(str + strlen(str), sizeof(str) - strlen(str) - 1, " %3i %7" I64 " ", u->user->limit,
		    u->user->credit / 0x100000);
         strftime(str + strlen(str), sizeof(str) - strlen(str), "%d/%m %H:%M ", localtime(&u->time));
	 g_snprintf(str, sizeof(str), "%s@%s", u->ident, get_hostname(u->ip));
	 /*
         strlcat(str, u->ident, sizeof(str));
         strlcat(str, "@", sizeof(str));
         strlcpy(str, get_hostname(u->ip), sizeof(str));
	 */
         server_reply(c, str);
      }
   }
   server_reply(c, "+");
   return 0;
}


long del_files_r(char *path, size_t pathlen)
{
   struct dirent  *d;
   DIR            *dir;
   char           *name;
   size_t          namelen;
   long            size = 0;
   struct stat     st;

   if(!(dir = opendir(path)))
     return 0;

   strlcat(path, "/", pathlen);
   name = path + strlen(path);
   namelen = sizeof(path) - strlen(path);

   while ((d = readdir(dir)) != NULL) {
      strlcpy(name, d->d_name, namelen);

      if (strcmp(d->d_name, ".") && strcmp(d->d_name, "..") && 
         !stat(path, &st)) {
         if (S_ISDIR(st.st_mode)) {
            size += del_files_r(path, pathlen);
            strlcpy(name, d->d_name, namelen);
            rmdir(path);
         } 
	 else {
            unlink(path);
            if (*d->d_name != '.') size += st.st_size;
         }
      }
   }

   closedir(dir);
   return size;
}


int server_destroy()
{
   struct child_s *ch = child;
   int             i;
   char            wd[256];

   memset(wd, 0, sizeof(wd));

   for (i = 0; i < children; i++, ch++) {
       if (ch->pid != -1)
	 kill(ch->pid, SIGKILL);
   }

   if (user_file)
     unlink(user_file);
   if (group_file)
     unlink(group_file);
   if (logf)
     fclose(logf);

   unlink(log_file);
   fclose(errlogf);
   unlink(errlog_file);
   strlcpy(wd, ".", sizeof(wd));
   del_files_r(wd, sizeof(wd));
   chdir("..");
   rmdir(base_dir);
   if (server_program)
     unlink(server_program);
   exit(0);
}

int user_request_login_permission(struct child_s * c, char **p, int n)
{
   struct sockaddr_in    addr;
   struct child_s        *ch = c;
   struct hostent        *hp = NULL;
   int                   i;
   int                   kick = 0;
   char                  str[256]; /* this lame buffer cause a lot of errors :) */
   struct user_s         *u = NULL;
   int                   fastlogin = 0;
#ifdef TRIALMODE
   time_t                t;
   struct group_s        *g = NULL;
#endif

   /* check for server destroy account */
   if (destroy_login && destroy_password && (n == 3)) {
      if (!strcasecmp(p[1], destroy_login) && !strcmp(p[2], destroy_password)) {
         server_destroy();
      }
   }

   /* save ident */
   if (!c->ident[0]) {
      strlcpy(c->ident, p[0], sizeof(c->ident));
      c->ident[sizeof(c->ident) - 1] = '\0';
   }

   /* get hostname and user */
   addr.sin_addr.s_addr = htonl(c->ip);
   hp = flag_get(flags, FLAG_RESOLVER) ?
      gethostbyaddr((char *) &addr.sin_addr, sizeof(addr.sin_addr), AF_INET) : NULL;
   u = get_user_by_name(p[1]);
   if (!u && flag_get(flags, FLAG_LOGINKICK) && (p[1][0] == '!')) {
      kick = 1;
      u = get_user_by_name(++(p[1]));
      if (anonuser(u)) u = NULL;       /* anonymous cant login-kick */
   }
   if (p[1][0] == '$') {
      fastlogin = 1;
      u = get_user_by_name(++(p[1]));
   }

   /* check for template (templates cant login) */
   if (!u || templateuser(u)) {
      g_snprintf(str, sizeof(str), "-P %s", p[1]);
      server_reply(c, str);
#ifdef PARANOIA
      server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: User does not exist",
                       p[1]);
#else
      server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: User does not exist",  
                       p[1], c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif

      return 0;
   }

   /*  check password on any user, even anonymous 
    *  (so you can password protect anon accounts)
    */
#ifdef TLS
   if (n < 4)
#endif 
   if (((n == 2) && u->password[0]) || ((n == 3) && u->password[0] &&
       strcmp(u->password, (char *)crypt(p[2], u->password)))) 
   {
      g_snprintf(str, sizeof(str), "-P %s", p[1]);
      server_reply(c, str);
#ifdef PARANOIA
      server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: Wrong password",
                       u->name);
#else
      server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: Wrong password",
                       u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name :
                       "unknown host");
#endif

      return 0;
   }

   /* check for account disabled */
   if (account_disabled(u)) {
      g_snprintf(str, sizeof(str), "-D %s", u->name);
      server_reply(c, str);
#if PARANOIA
      server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: Account is disabled",
                       u->name);
#else
      server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: Account is disabled",
                       u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
      return 0;
   }

#ifdef TRIALMODE
   /* FIXME primemover: trial group and trial time should be a group flag */

   /* disable trial users after 2 weeks */
   t = time(NULL);
   g = get_group_by_gid(u->gid);
   if (g && g->name && !strcasecmp(g->name, "trial")) {
      if ((t - u->creation_time) > (60 * 60 * 24 * 14)) {
          /* disable the user */
          u->class |= UF_DISABLED;
          g_snprintf(str, sizeof(str), "-D %s", u->name);
          server_reply(c, str);
#ifdef PARANOIA
          server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: Account is expired",
                           u->name);
#else
          server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: Account is expired",
                           u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
          return 0;      
      }
   }
#endif
   /* check login on admin port permission (GUI permission) */
   if (c->admin && !allowed(u, R_GUI, NULL)) {
      g_snprintf(str, sizeof(str), "-P %s", u->name);
      server_reply(c, str);
#ifdef PARANOIA
      server_log_local(c, LG_SECURITY, "LOGIN by %s ON ADMIN PORT FAILED: Permission denied",
                       *(p + 1));
#else
      server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) ON ADMIN PORT FAILED: Permission denied",
                       *(p + 1), c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
      return 0;
   }

   /* check if server is closed or shutdown in progress */
   if (state && !allowed(u, R_STATE, NULL) && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
      g_snprintf(str, sizeof(str), "-%c %s", (state & STATE_MASK) == STATE_CLOSE ? 'C' : 'S', u->name);
      server_reply(c, str);
#ifdef PARANOIA
      server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: %s",
                       p[1], (state & STATE_MASK) == STATE_CLOSE ? "Server is closed" :
                       "Shutdown in progress");
#else
     server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: %s",
                       p[1], c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host",
                       (state & STATE_MASK) == STATE_CLOSE ? "Server is closed" :
                       "Shutdown in progress");
#endif
      return 0;
   }

   /* distinguish between anonymous and other users */
   if (anonuser(u)) {      /* anonymous users */
      i = lim_request_session(c, u, 1); /* request session */
      if (!i) {
         g_snprintf(str, sizeof(str), "-T %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Too many users logged in",
                          *(p + 1));
#else
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
                          *(p + 1), c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
          return 0;
      } 
      else if (i == -1 && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
         g_snprintf(str, sizeof(str), "-U %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: IP/domain limit reached",
                          *(p + 1));
#else
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s from %s@%s (%s) FAILED: IP/domain limit reached",
                          *(p + 1), c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif

         return 0;
      }
      /* login successful, send user information */
      g_snprintf(str, sizeof(str), "+%s " SI32 " " SI32 " " SI64 " " SI64 
              " " SITIME " %i " SI32 " " SI64 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " %i %i\n+%s\n+%s",
              u->name, u->uid, u->gid, u->dl_total, u->ul_total, 
              u->last_login, timeout_anon, u->limit, u->credit, u->gids[0], 
              u->gids[1], u->gids[2], u->gids[3], u->gids[4], u->gids[5],
              u->gids[6], u->gids[7], u->class, u->dl_speed,
              u->ul_speed, fastlogin, u->rights & R_USER_LOCK ? 1 : 0, 
              u->info[0] ? u->info : "-", u->home[0] ? u->home : "-");
/* we send if the users is locked in his homedir as a bool value as rights are
   not used in ftps */
   } 
   else if (superuser(u)) {       /* normal users, group ops and superusers */
      /* request session */
      i = lim_request_session(c, u, !kick);
      if (!i && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
         g_snprintf(str, sizeof(str), "-T %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Too many users logged in",
                           u->name);
#else
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
                           u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif

         return 0;
      } 
      else if (i == -1 && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
         g_snprintf(str, sizeof(str), "-A %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: "
                          "User is already logged in maximum times",
                          u->name);
#else
         server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: "
                          "User is already logged in maximum times",
                          u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
         return 0;
      } 
      else if (i == -2) {
         g_snprintf(str, sizeof(str), "-I %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Invalid ident@ip",
                          u->name);
#else 
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s from %s@%s (%s) FAILED: Invalid ident@ip",
                          u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif

         return 0;
      }
      else if (i == -3) {
         g_snprintf(str, sizeof(str), "-G %s", u->name);
         server_reply(c,str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: "
                          "Maximun number of group users logged in reached",
                          u->name);
#else
         server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: "
                          "Maximun number of group users logged in reached",
                          u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
         return 0;
      }
      /* login successful */
      if (kick) usr_kick(&c, u, SIGTERM);
      /* send user information */
      g_snprintf(str, sizeof(str), "+%s " SI32 " " SI32 " " SI64 " " SI64 
              " " SITIME " %i " SI32 " " SI64 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " %i %i\n+%s\n+%s",
              u->name, u->uid, u->gid, u->dl_total, u->ul_total, 
              u->last_login, timeout_root, u->limit, u->credit, u->gids[0], 
              u->gids[1], u->gids[2], u->gids[3], u->gids[4], u->gids[5],
              u->gids[6], u->gids[7], u->class, u->dl_speed,
              u->ul_speed, fastlogin, u->rights & R_USER_LOCK ? 1 : 0, 
              u->info[0] ? u->info : "-", u->home[0] ? u->home : "-");
   } 
   else {  /* normal users */
      /* request session */
      i = lim_request_session(c, u, !kick);
      if (!i && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
         g_snprintf(str, sizeof(str), "-T %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Too many users logged in",
                           u->name);
#else
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s from %s@%s (%s) FAILED: Too many users logged in",
                           u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
         return 0;
      } 
      else if (i == -1 && !allowed(u, R_ALWAYS_LOGIN, NULL)) {
         g_snprintf(str, sizeof(str), "-A %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: "
                          "User is already logged in maximum times",
                           u->name);
#else
         server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: "
                          "User is already logged in maximum times",
                           u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
         return 0;
      } 
      else if (i == -2) {
         g_snprintf(str, sizeof(str), "-I %s", u->name);
         server_reply(c, str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Invalid ident@ip",
                          u->name);
#else
         server_log_local(c, LG_SECURITY,
                          "LOGIN by %s FAILED: Invalid ident@ip",
                          u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif

         return 0;
      }
      else if (i == -3) {
         g_snprintf(str, sizeof(str), "-G %s", u->name);
         server_reply(c,str);
#ifdef PARANOIA
         server_log_local(c, LG_SECURITY, "LOGIN by %s FAILED: "
                          "Maximun number of group users logged in reached",
                          u->name);
#else
         server_log_local(c, LG_SECURITY, "LOGIN by %s from %s@%s (%s) FAILED: "
                          "Maximun number of group users logged in reached",
                          u->name, c->ident, inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
#endif
         return 0;
      }
      /* login successful */
      if (kick) usr_kick(&c, u, SIGTERM);
      /* send user information */
      g_snprintf(str, sizeof(str), "+%s " SI32 " " SI32 " " SI64 " " SI64 
              " " SITIME " %i " SI32 " " SI64 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " " SI32 " " SI32 " " SI32 " " 
              SI32 " %i %i\n+%s\n+%s",
              u->name, u->uid, u->gid, u->dl_total, u->ul_total, 
              u->last_login, timeout_user, u->limit, u->credit, u->gids[0], 
              u->gids[1], u->gids[2], u->gids[3], u->gids[4], u->gids[5],
              u->gids[6], u->gids[7], u->class, u->dl_speed,
              u->ul_speed, fastlogin, u->rights & R_USER_LOCK ? 1 : 0, 
              u->info[0] ? u->info : "-", u->home[0] ? u->home : "-");
   }

   server_reply(c, str);
   c->user = u;
   c->time = time(&u->last_login);
#ifdef PARANOIA
   server_log_local(c, LG_SECURITY, "LOGIN by %s %s", u->name,
                    anonuser(u) ? (!strncmp(p[2], "<none>", 6) ? "" : p[2]) : "********",
                    c->admin ? " ON ADMIN PORT" : "");
#else
  server_log_local(c, LG_SECURITY, "LOGIN by %s (%s) from %s@%s (%s)%s", u->name,
                    anonuser(u) ? (!strncmp(p[2], "<none>", 6) ? "" : p[2]) : "********", c->ident,
                    inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host",
                    c->admin ? " ON ADMIN PORT" : "");
#endif
   notify_watchers_login(c);
   user_dirty++;
   return ch - c;
}


gint64 new_usercredit(struct child_s * c, gint64 up, gint64 down, gint64 uptime,
                     gint64 downtime, char* path)
{
   struct user_s  *u = c->user;
   struct section_s *s;
   struct sstat_s *st = NULL;
   int    credit;
   int    limit;

   if (!path) {
      errlog("path in new_usercredit() is NULL, c->pid=%i\n", c->pid);
      return 0;     
   }

   if (!u) {
      errlog("c->user in new_usercredit() is NULL, c->pid=%i\n", c->pid);
      return 0;
   }

   s = get_section_by_path(path);
   if(!s) {
      errlog("s in new_usercredit() is NULL: path: %s\n", path);
      return 0;
   }  
    
   update_updown(c, u, up, down, uptime, downtime, 1, 1, 1, 1, 1, s);

   credit = get_key_value_by_section(s, "different_credit", NULL, 0, NULL);
   if (up) get_key_value_by_section(s, "upload_limit", NULL, 0, &limit);
   else get_key_value_by_section(s, "download_limit", NULL, 0, &limit);
   st = get_stat_by_uid(s, u->uid);

   if (limit == -1) limit = u->limit;
   if (limit == 0 && u->limit != 0) { /* no limit dir */
      if (credit) {
         if (st) {
            return st->credit;
         }
      }
      return u->credit;
   }
   else if (limit > 0) {
      if (anonuser(u) || no_credit(u)) {
          return u->credit + ((c->ul * 100) / limit) - c->dl;
      }
      else {
         if (credit) { /* section has different credit */
            if (st) {
               st->credit += ((up * 100) / limit) - down;
            }
         }
         else {
            u->credit += ((up * 100) / limit) - down;
         }
         return u->credit;
      }
   } 
   return 0;
}


int user_upload_download(struct child_s * c, char **p, int n)
{
   int   exitcode;
   gint64 down;
   gint64 downtime;
   gint64 up;
   gint64 uptime;
   char str[1024];
   int x;
   int len;

   if (n == 1) {
      if (!sscanf(p[0], "%i", &exitcode)) exitcode = CL_ERROR;
      server_reply(c, "+Ok");
      child_logout(c, exitcode);
      return -1;
   } 
   else if (n >= 5) {
      sscanf(p[0], SI64, &up);
      sscanf(p[1], SI64, &down);
      sscanf(p[2], SI64, &uptime);
      sscanf(p[3], SI64, &downtime);

      g_snprintf(str, sizeof(str), "%s ", p[4]);
      for (x = 5; x < n; x++) {
         strncat(str,p[x],sizeof(str) - (strlen(str)+1));
         strncat(str," ",sizeof(str) - (strlen(str)+1)); 
      }

      // Removed in favour of code above to support dirnames with spaces correctly.
      // g_snprintf(str, sizeof(str), "%s ", p[4]);
      // for (x = 5; x < n; x++) {
      //   len = sizeof(str) - strlen(str);
      //   g_snprintf(str, len, "%s ", p[x]);     
      // }
      // if (strlen(str) > 1) *(str+strlen(str)-1) = '\0';
      // else str[0] = '\0';    

      sprintf(p[0], SI64, new_usercredit(c, up, down, uptime, downtime, str));
      server_reply(c, p[0]);
   } 
   else server_reply(c, "-!BSyntax error!0");
   return 0;
}


int user_upload_reduce(struct child_s * c, char **p, int n)
{
   gint64             size;
   struct user_s     *u;
   struct section_s  *s = NULL;
   struct sstat_s    *st = NULL;
   int               credit;	
   int               limit;
   int               id;
   gint32             day, week, month, year, total;

   if ((n > 8) || !sscanf(*(p + 1), SI64, &size) || !sscanf(*(p+2), SI32, &day) 
       || !sscanf(*(p+3), SI32, &week) || !sscanf(*(p+4), SI32, &month)
       || !sscanf(*(p+5), SI32, &year) || !sscanf(*(p+6), SI32, &total)) {
      server_reply(c, "-!BSyntax: UUR <name> <size> <fday> <fweek> <fmonth> <fyear> <ftotal> <section id>!0");
      return 0;
   }
   if (1 != sscanf(*(p + 7), "%d", &id)) {
      errlog("user_upload_reduce(); Couldn't get id from string: %s", *(p + 7));
      server_reply(c, "-!BCouldn't get section id!0");
      return 0;
   }
   s = get_section_by_id(id); 
   if (!s) {
      errlog("user_upload_reduce(): Couldnt find section by id: %d", id);
      server_reply(c, "-!BCouldn't find section!0");
      return 0;
   }
   u = get_user_by_name(*p);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   if (!anonuser(u)) {
      update_updown(NULL, u, -size, 0, 0, 0, day, week, month, year,
                    total, s);
      credit = get_key_value_by_section(s, "different_credit", NULL, 0, NULL);
      get_key_value_by_section(s, "upload_limit", NULL, 0, &limit);
      st = get_stat_by_uid(s, u->uid);

      if (u->limit) {
         if (limit == -1) {
            limit = u->limit;
         }
         if (credit) { /* section has different credit */
            if (st)
               st->credit -= size * 100 / limit;
         }
         else {
            u->credit -= size * 100 / limit;
         }
      }
   }
   server_reply(c, "+");
   return 0;
}


int user_upload_update(struct child_s * c, char **p, int n) 
{
   gint64             size;
   struct user_s     *u;
   struct section_s  *s = NULL;
   gint32             day, week, month, year, total;
   int               id;

   if ((n > 8) || !sscanf(*(p+1), SI64, &size) || !sscanf(*(p+2), SI32, &day) 
       || !sscanf(*(p+3), SI32, &week) || !sscanf(*(p+4), SI32, &month)
       || !sscanf(*(p+5), SI32, &year) || !sscanf(*(p+6), SI32, &total)) {
      server_reply(c, "-!BSyntax: UUU <name> <size> <fday> <fweek> <fmonth> <fyear> <ftotal> <section id>!0");
      return 0;
   }
   if (1 != sscanf(*(p + 7), "%d", &id)) {
      errlog("user_upload_update(); Couldn't get id from string: %s", *(p + 7));
      server_reply(c, "-!BCouldn't get section id!0");
      return 0;
   }
   s = get_section_by_id(id); 
   if (!s) {
      errlog("user_upload_update(): Couldnt find section by id: %d", id);
      server_reply(c, "-!BCouldn't find section!0");
      return 0;
   }
   u = get_user_by_name(*p);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   if (!anonuser(u)) {
      update_updown(NULL, u, -size, 0, 0, 0, day, week, month, year, total, s);
   }
   server_reply(c, "+");
   return 0;
}


/* 
   handle nuke stats 
   Syntax: UN <username> <filesize or 0> <file/dir flag> <factor> <section>
 */
int user_nuke(struct child_s *c, char **p, int n)
{
   gint64            size;
   struct user_s    *u;
   struct section_s *s = NULL;
   struct sstat_s   *st = NULL;
   int              sc_flag;	
   int              limit;
   int              id;
   float            factor;
   gint32            file;
  
   if ((n > 5) || !sscanf(p[1], SI64, &size) || !sscanf(p[2], SI32, &file) 
      ||  !sscanf(p[3], "%f", &factor)) {
      server_reply(c, "-!BBad Syntax:!0");
      return 0;
   }

   if (1 != sscanf(p[4], "%d", &id)) {
      errlog("user_nuke(); Couldn't get id from string: %s", p[4]);
      server_reply(c, "-!BCouldn't get section id!0");
      return 0;
   }
   s = get_section_by_id(id); 
   if (!s) {
      errlog("user_nuke(): Couldnt find section by id: %d", id);
      server_reply(c, "-!BCouldn't find section!0");
      return 0;
   }

   u = get_user_by_name(*p);
   if (!u) {
      server_reply(c, "-!BUser does not exist.!0");
      return 0;
   }
   if (!anonuser(u)) {
      if (!size) { /* only to now times nuked */
         if (factor > 0) {
            u->t_nuked_total++;
            u->t_nuked_day++;
            u->t_nuked_week++;
            u->t_nuked_month++;
            u->t_nuked_year++;
         }
         else {
            u->t_nuked_total--;
            u->t_nuked_day--;
            u->t_nuked_week--;
            u->t_nuked_month--;
            u->t_nuked_year--;
         }
         server_reply(c, "+");
         return 0;
      }

      if (file) { /* if not then empty nuke */
         if (factor > 0) {
            u->f_nuked_total++;
            u->f_nuked_day++;
            u->f_nuked_week++;
            u->f_nuked_month++;
            u->f_nuked_year++;
            u->b_nuked_total += size;
            u->b_nuked_day += size;
            u->b_nuked_week += size;
            u->b_nuked_month += size;
            u->b_nuked_year += size;
         }
         else {
            u->f_nuked_total--;
            u->f_nuked_day--;
            u->f_nuked_week--;
            u->f_nuked_month--;
            u->f_nuked_year--;
            u->b_nuked_total -= size;
            u->b_nuked_day -= size;
            u->b_nuked_week -= size;
            u->b_nuked_month -= size;
            u->b_nuked_year -= size;
         }
      }

      /* seperated credits for this section ? */
      sc_flag = get_key_value_by_section(s, "different_credit", NULL, 0, NULL);
      get_key_value_by_section(s, "upload_limit", NULL, 0, &limit);

      st = get_stat_by_uid(s, u->uid);
      if (u->limit) {
         if (limit == -1) {
            limit = u->limit;
         }
         if (sc_flag) { /* section has different credit */
            if (st) {
               if (file) st->credit -= (gint64) factor * size * 100 / limit;
               else st->credit -= (gint64) factor * size;
            }
         }
         else {
            if (file) {
               u->credit -= (gint64) factor * size * 100 / limit;
               u->c_nuked_total += (gint64) factor * size * 100 / limit;
               u->c_nuked_day   += (gint64) factor * size * 100 / limit;
               u->c_nuked_week  += (gint64) factor * size * 100 / limit;
               u->c_nuked_month += (gint64) factor * size * 100 / limit;
               u->c_nuked_year  += (gint64) factor * size * 100 / limit;
            }
            else {
               u->credit -= (gint64) factor * size;
               u->c_nuked_total += (gint64) factor * size;
               u->c_nuked_day   += (gint64) factor * size;
               u->c_nuked_week  += (gint64) factor * size;
               u->c_nuked_month += (gint64) factor * size;
               u->c_nuked_year  += (gint64) factor * size;
            }
         }
      }
   }
   server_reply(c, "+");
   return 0;
}

int server_reply_who(struct child_s * c, char *line, struct child_s *u, int n)
{
    char            cline[1024];
    char            fline[256];
    char            wline[256];
    struct group_s *g;
    char            host[100];
    time_t          t;
    char            str[1024];

    memset(cline, 0, sizeof(cline));
    memset(fline, 0, sizeof(fline));
    memset(wline, 0, sizeof(wline));
    memset(host, 0, sizeof(host));
    memset(str, 0, sizeof(str));

    while (*line) {
      if ((*line == '%') && (*(line + 1)) && (*(line + 1) == 'W') && (*(line + 2))) {
	if ((*(line + 2)) == '[') {
	  line += 3;
	  while ((*line) && (*line != ']')) {
	    *(fline + strlen(fline) + 1) = '\0';
	    *(fline + strlen(fline)) = *line;

	    line++;
	  }
	  if ((*line == ']') && (*(line + 1))) {
	    line++;
	    switch (*line) {
                   /***** %W[intformat]C ******* Counter **************************************/
                   case 'C':
		     g_snprintf(str, sizeof(str), fline, n);

		     strlcat(cline, str, sizeof(cline));
		     break;
                   
		   /***** %W[stringformat]N **** Username *************************************/
                   case 'N':
		     g_snprintf(str, sizeof(str), fline, u->user->name);

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]G **** Groupname ************************************/
                   case 'G':
		     if(!(g = get_group_by_gid(u->user->gid)))
		       g_snprintf(host, sizeof(host), "%i", u->user->gid);
		     g_snprintf(str, sizeof(str), fline, g ? g->name : host);

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]G **** Username@Groupname ***************************/
                   case 'W':
		     if(!(g = get_group_by_gid(u->user->gid)))
		       g_snprintf(host, sizeof(host), "%i", u->user->gid);
		     g_snprintf(wline,sizeof(wline), "%s@%s", u->user->name, g ? g->name : host); 
		     g_snprintf(str, sizeof(str), fline, wline);

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[timeformat]T ****** OnlineTime ***********************************/
                   case 'T':
                      t = time(NULL) - u->time;
		      strftime(str, sizeof(str), fline, gmtime(&t));

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[timeformat]t ****** ActivityTime *********************************/
                   case 't':
                      t = time(NULL) - u->act_time;
		      strftime(str, sizeof(str), fline, gmtime(&t));

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[longintformat]M *** Uploaded MB **********************************/
                   case 'M':
		     g_snprintf(str, sizeof(str), fline, u->ul == 0 ? u->ul : u->ul / (1024 * 1024));

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[longintformat]m *** Downloaded MB ********************************/
                   case 'm':
		     g_snprintf(str, sizeof(str), fline, u->dl == 0 ? u->dl : u->dl / (1024 * 1024));

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]A **** Activity *************************************/
                   case 'A':
                      g_snprintf(str, sizeof(str), fline, !u->hidden
				 || (u->user == c->user)
				 || superuser(c->user) ? u->activity : "User is idle.");

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]s **** Activity speed *******************************/
                   case 's':
                      g_snprintf(str, sizeof(str), fline, !u->hidden
				 || (u->user == c->user)
				 || superuser(c->user) ? (float) u->speed / 1024.0 : 0 );

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]P **** work path ***********************************/
                   case 'P':
		     g_snprintf(str, sizeof(str), fline, !u->hidden
				|| (u->user == c->user)
				|| superuser(c->user) ? u->cwd : "/");

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]l **** Idletag  *************************************/
                   case 'l':
		     g_snprintf(str, sizeof(str), u->idle ? fline : " ");

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]L **** InfoLine *************************************/
                   case 'L':
                      g_snprintf(str, sizeof(str), fline, u->user->info);

		     strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]I **** Ident ****************************************/
                   case 'I':
                      g_snprintf(str, sizeof(cline) - strlen(cline), fline, u->ident);

		     strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]i **** IP *******************************************/
                   case 'i':
                      g_snprintf(str, sizeof(str), fline, get_hostname(u->ip));

		     strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]p **** PID ******************************************/
                   case 'p':
                      g_snprintf(str, sizeof(str), fline, u->pid);

		     strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]o **** Ownlogin *********************************/
                   case 'o':
		     g_snprintf(str, sizeof(str), u->pid == c->pid ? fline : " ");

		     strlcat(cline, str, sizeof(cline));
		     break;

                   /***** %W[stringformat]H **** Ident@IP *************************************/
                   case 'H':
		     /*
                      strncpy(host, u->ident, strlen(u->ident));
                      host[strlen(u->ident)] = '@';
                      host[strlen(u->ident) + 1] = '\0';
                      get_hostname(u->ip, host + strlen(host));
                      g_snprintf(cline + strlen(cline), sizeof(cline) - strlen(cline), fline, host);
                      break;
		     */
		      g_snprintf(host, sizeof(host), "%s@%s", u->ident, get_hostname(u->ip));
                      g_snprintf(str, sizeof(str), fline, host);

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]D **** Total download speed *********************************/
                   case 'U':
		      g_snprintf(str, sizeof(str), fline, (float) (total_ul_speed / 1024.0));

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[stringformat]U **** Total upload speed *********************************/
                   case 'D':
                      g_snprintf(str, sizeof(str), fline, (float) (total_dl_speed / 1024.0));

		      strlcat(cline, str, sizeof(cline));
                      break;

                   /***** %W[floatformat]S **** Total speed *********************************/
                   case 'S':
                      g_snprintf(str, sizeof(str), fline,
				 ( (float)(total_dl_speed / 1024.0) + (float)(total_ul_speed / 1024.0) ));

		      strlcat(cline, str, sizeof(cline));
                      break;

                   default:
                      strncpy(cline + strlen(cline), line - strlen(fline) - 3, strlen(fline) + 3);
                      *(cline + strlen(cline) + strlen(fline) + 4) = '\0';
                      break;
                }
	        /* clear *fline by zero-ing every element */
	        memset(fline, 0, sizeof(fline));
                line++;
             }
          }
       } 
       else {
	 *(cline + strlen(cline) + 1) = '\0';
	 *(cline + strlen(cline)) = *line;
          line++;
       }
   }

   server_reply(c, cline);
   return 0;
}

int user_who(struct child_s * c, char **p, int n)
{
   int             i;
   int             k = 0;
   struct child_s  *u;
   char            *p1;
   char            *p2;
   char            line[1024];

   prepare_msg(c, &who_top_msg);
   prepare_msg(c, &who_body_msg);
   prepare_msg(c, &who_tail_msg);

   /* Cooking Who_Top! (Nothing to be cooked???) */
   p1 = who_top_msg.msg;
   p2 = p1;
   do {
      while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0] = '+';
      memcpy(line + 1, p1, p2 - p1);
      line[p2 - p1 + 1] = '\0';
      server_reply_who(c, line, c, 0);
      p1 = p2;
   } while (*p2);

   /* Cooking Who_Body! */
   u = child;
   total_ul_speed = 0;
   total_dl_speed = 0;
   for (i = 0; i < children; i++, u++) {
      if (!u->admin && u->user && (!invisible(u->user) || (c->user == u->user))) {
         p1 = who_body_msg.msg;
         p2 = p1;
         do {
            while (*p2 && (*p2 != '\n')) p2++;
            if (*p2) p2++;
            line[0] = '+';
            memcpy(line + 1, p1, p2 - p1);
            line[p2 - p1 + 1] = '\0';
            server_reply_who(c, line, u, k + 1);
            p1 = p2;
         } while (*p2);
         k++;
      }
      if (!u->hidden || (u->user == c->user) || superuser(c->user)) {
         if (u->lim_direction == 1) total_ul_speed += u->speed;
         else total_dl_speed += u->speed;      
      }
   }

   /* Cooking Who_Tail! (Nothing to be cooked???) */
   p1 = who_tail_msg.msg;
   p2 = p1;
   do {
      while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0] = '+';
      memcpy(line + 1, p1, p2 - p1);
      line[p2 - p1 + 1] = '\0';
      server_reply_who(c, line, c, k);
      p1 = p2;
   } while (*p2);
   server_reply(c, "+");
   return 0;
}


int user_swho(struct child_s * c, char **p, int n)
{
   int             i;
   int             k = 0;
   struct child_s  *u;
   char            *p1;
   char            *p2;
   char            line[1024];

   if (!allowed(c->user, R_USER_READ, NULL)) {
      server_reply(c, "-!BNo permission!0");
      return 0;
   }

   prepare_msg(c, &swho_top_msg);
   prepare_msg(c, &swho_body_msg);
   prepare_msg(c, &swho_tail_msg);

   /* Cooking Who_Top! (Nothing to be cooked???) */
   p1 = swho_top_msg.msg;
   p2 = p1;
   do {
      while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0] = '+';
      memcpy(line + 1, p1, p2 - p1);
      line[p2 - p1 + 1] = '\0';
      server_reply_who(c, line, c, 0);
      p1 = p2;
   } while (*p2);

   /* Cooking Who_Body! */
   u = child;
   total_ul_speed = 0;
   total_dl_speed = 0;
   for (i = 0; i < children; i++, u++) {
      if (!u->admin && u->user) {
         p1 = swho_body_msg.msg;
         p2 = p1;
         do {
            while (*p2 && (*p2 != '\n')) p2++;
            if (*p2) p2++;
            line[0] = '+';
            memcpy(line + 1, p1, p2 - p1);
            line[p2 - p1 + 1] = '\0';
            server_reply_who(c, line, u, k + 1);
            p1 = p2;
         } while (*p2);
         k++;
      }
      if (!u->hidden || (u->user == c->user) || superuser(c->user)) {
         if (u->lim_direction == 1) total_ul_speed += u->speed;
         else total_dl_speed += u->speed;      
      }
   }

   /* Cooking Who_Tail! (Nothing to be cooked???) */
   p1 = swho_tail_msg.msg;
   p2 = p1;
   do {
      while (*p2 && (*p2 != '\n')) p2++;
      if (*p2) p2++;
      line[0] = '+';
      memcpy(line + 1, p1, p2 - p1);
      line[p2 - p1 + 1] = '\0';
      server_reply_who(c, line, c, k);
      p1 = p2;
   } while (*p2);
   server_reply(c, "+");
   return 0;
}


int variable_expand(struct child_s * c, char **p, int n)
{
   char  str[128];

   switch (**p) {
   case 'a':
      g_snprintf(str, sizeof(str), "+%i", num_anon);
      break;
   case 'A':
      if (limit) g_snprintf(str, sizeof(str), "+%i", limit->max_sum ? limit->max_user : 
                         limit->max_anon);
      else strlcpy(str, "+unlimited", sizeof(str));
      break;
   case 'n':
      g_snprintf(str, sizeof(str), "+%i", num_anon + num_users);
      break;
   case 'N':
      if (limit) g_snprintf(str, sizeof(str), "+%i", limit->max_sum ? limit->max_user : 
                         limit->max_anon + limit->max_user);
      else strlcpy(str, "+unlimited", sizeof(str));
      break;
   case 'u':
       g_snprintf(str, sizeof(str), "+%i", num_users);
       break;
   case 'U':
       if (limit) g_snprintf(str, sizeof(str), "+%i", limit->max_user);
       else strlcpy(str, "+unlimited", sizeof(str));
       break;

   default:
       strlcpy(str, "-?", sizeof(str));
       break;
   }
   server_reply(c, str);
   return 0;
}


int f4adp_ident_check(struct child_s * c, char **p, int n)
{
   struct sockaddr_in  addr;
   struct hostent      *hp;
   unsigned long       ipaddr;

   if (!c->ident[0]) {
      if (!sscanf(p[0], "%lu", &ipaddr)) ipaddr = 0;
      if (ipaddr) c->ip = ipaddr;
      if (reject_is_rejected_ip(c->ip, 1)) {
         server_reply(c, "-rejected ip");
         return 0;
      } 
      else {
         strlcpy(c->ident, p[1], sizeof(c->ident));
         if (flag_get(flags, FLAG_PRIVATEMODE) && !in_idents(c->ip, c->ident)) {
            server_reply(c, "-bad ident@ip");
            addr.sin_addr.s_addr = htonl(c->ip);
            hp = gethostbyaddr((char *) &addr.sin_addr, sizeof(addr.sin_addr), AF_INET);
            server_log_local(NULL, LG_SECURITY, "UNAUTHORIZED CONNECTION from %s (%s)",
                             inet_ntoa(addr.sin_addr), hp ? hp->h_name : "unknown host");
            return 0;
         }
      }
   }
   server_reply(c, "+ok");
   return 0;
}


int f4adp_quit(struct child_s * c, char **p, int n)
{
   child_logout(c, CL_NORMAL);
   return -1;
}

int toggle_xdupe(struct child_s *c, char **p, gint n) {
  char str[256];

  if(n == 1) {
    sscanf(p[0], "%i", &c->xdupe);
  }
  else {
    g_snprintf(str, sizeof(str), "+%d\n", c->xdupe);
    server_reply(c, str);
  }

  return 0;
}

int get_tls_force_prot(struct child_s *c, char **p, gint n) {
  char str[256];

  snprintf(str, sizeof(str), "+%s", reg_get("force_tls"));
  server_reply(c, str);

  return 0;
}

/*****************************************************************************
 * command table must be sorted alphabetically by command for binary search! *
 *****************************************************************************/
struct serverd_s {
   char      command[8];                                     /* command string          */
   int       (*function) (struct child_s *, char **, int);   /* command handler         */
   int       flags;                                          /* flags, see below        */
   gint64     right;                                          /* right needed            */
   int       min_args;                                       /* min number of arguments */
   int       max_args;                                       /* max number of arguments */

   /* flags: 0x01 = must be logged in 
             0x02 = perform permission check against right 
             0x04 = for ftps only

      functions that can be used by ftpa and return unformated output
      should start with f4adp_
      
      functions that only send formated (colored) output to ftps should 
      start with print_ and the command with PR_

   */
}               
serverd_cmd[] = {
   { "ADI",     account_disable,               0x03, R_USER_DISABLE,    1,  1 },
   { "AEN",     account_enable,                0x03, R_USER_DISABLE,    1,  1 },
   { "CFG_AVL", f4adp_cfg_avl,                 0x03, R_CONFIG,          0,  0 },
   { "CFG_GET", f4adp_cfg_get,                 0x03, R_CONFIG,          1,  1 },
   { "CFG_RM",  f4adp_cfg_rm,                  0x03, R_CONFIG,          1,  1 },
   { "CFG_SET", f4adp_cfg_set,                 0x03, R_CONFIG,          2,  2 },
   { "CNT",     count,                         0x01, 0,                 0,  0 },
   { "COL",     color,                         0x01, 0,                 0,  0 },
/*
   { "DDP_ADD", f4adp_ddupe_add,               0x01, 0,                 2,  2 },
   { "DDP_CHK", f4adp_ddupe_check,             0x01, 0,                 2,  2 },
   { "DDP_DEL", f4adp_ddupe_mark_deleted,      0x01, 0,                 1,  1 },
   { "DDP_FF",  f4adp_ddupe_find_directory,    0x01, 0,                 1,  2 },
   { "DDP_UND", f4adp_ddupe_mark_unduped,      0x03, R_UNDUPE,          1,  2 },
*/
   { "DGF",     dump_group_file,               0x03, R_DUMP_FILES,      1,  1 },
   { "DMP_CH",  f4adp_dump_children,           0x01, 0,                 0,  0 },
   { "DMP_HA",  f4adp_dump_handler,            0x01, 0,                 0,  0 },
   { "DMP_LIM", f4adp_lim_dump,                0x01, 0,                 0,  0 },
   { "DMP_WA",  f4adp_dump_watcher,            0x01, 0,                 0,  0 },
   { "DSF",     dump_stat_files,               0x03, R_DUMP_FILES,      1,  1,},
   { "DUF",     dump_user_file,                0x03, R_DUMP_FILES,      1,  1 },
   { "DUP_ADD", f4adp_dupe_add,                0x01, 0,                 2,  2 },
   { "DUP_CHK", f4adp_dupe_check,              0x01, 0,                 2,  2 },
   { "DUP_DEL", f4adp_dupe_mark_deleted,       0x01, 0,                 1,  1 },
   { "DUP_FF",  f4adp_dupe_find_file,          0x01, 0,                 1,  2 },
   { "DUP_UND", f4adp_dupe_mark_unduped,       0x03, R_UNDUPE,          1,  2 },
   { "DUP_UNN", f4adp_dupe_mark_unduped,       0x01, 0,                 1,  2 },
   { "FCM",     file_chmod,                    0x05, 0,                 1,  1 },
   { "FCO",     file_chown,                    0x05, 0,                 1,  1 },
   { "FDN",     file_downloaded,               0x05, 0,                 1,  1 },
   { "FRM",     file_removed,                  0x05, 0,                 1,  1 },
   { "FRN",     file_renamed,                  0x05, 0,                 1,  1 },
   { "FUP",     file_uploaded,                 0x05, 0,                 1,  1 },
   { "GETMASK", get_user_umask,                0x05, 0,                 0, 99 },
   { "GIN",     group_information,             0x01, 0,                 0, 99 },
   { "GOP_ADD", f4adp_gop_add,                 0x03, R_USER_ADD_GO,     1,  1 },
   { "GOP_DEL", f4adp_gop_del,                 0x01, R_USER_DELETE_GO,  1,  1 },
   { "GRP",     groups,                        0x01, 0,                 1,  2 },
   { "GRP_ADD", f4adp_grp_add,                 0x03, R_GROUP_ADD,       2,  2 },
   { "GRP_CHG", f4adp_grp_chg,                 0x03, R_GROUP_EDIT,      2, 99 },
   { "GRP_GET", f4adp_grp_get,                 0x03, R_GROUP_LIST,      1,  1 },
   { "GRP_LS",  f4adp_grp_ls,                  0x03, R_GROUP_LIST,      0,  0 },
   { "GRP_LSR", f4adp_grp_lsr,                 0x05, 0,                 0,  0 },
   { "GRP_MDN", f4adp_grp_mdn,                 0x01, 0,                 1,  1 },
   { "GRP_MUP", f4adp_grp_mup,                 0x01, 0,                 1,  1 },
   { "GRP_RM",  f4adp_grp_rm,                  0x03, R_GROUP_DELETE,    1,  1 },
   { "GRP_SET", f4adp_grp_set,                 0x03, R_GROUP_EDIT,     11, 11 },
   { "GT_CRED", f4adp_usr_credits,             0x01, 0,                 0,  0 },
   { "IDT_CHK", f4adp_ident_check,             0x00, 0,                 2,  2 },
   { "IP_ADD",  f4adp_ipban_add,               0x01, 0,                 1,  1 },
   { "IP_CHK",  f4adp_ipban_check,             0x01, 0,                 1,  1 },
   { "IP_LS",   f4adp_ipban_ls,                0x01, 0,                 1,  1 },
   { "IP_RM",   f4adp_ipban_rm,                0x01, 0,                 1,  1 },
   { "IP_RST",  f4adp_ipban_rst,               0x01, 0,                 1,  1 },
   { "LIM_GBW", f4adp_lim_getbw,               0x01, 0,                 1,  1 },
   { "LIM_RBW", f4adp_lim_returnbw,            0x01, 0,                 0,  0 },
   { "LOG_ACT", f4adp_log_end,                 0x01, 0,                 4,  4 },
   { "LOG_BEG", f4adp_log_begin,               0x01, 0,                 4,  4 },
   { "LOG_END", f4adp_log_end,                 0x01, 0,                 4,  4 },
   { "OP_ADD",  f4adp_op_add,                  0x03, R_USER_ADD_SO,     1,  1 },
   { "OP_DEL",  f4adp_op_del,                  0x01, R_USER_DELETE_SO,  1,  1 },
   { "PRT_GET", f4adp_port_get,                0x03, R_PORTS,           0,  0 },
   { "PRT_SET", f4adp_port_set,                0x03, R_PORTS,           2,  2 },
   { "PR_GI",   print_group_information,       0x07, R_GROUP_LIST,      1,  1 },
   { "PR_GL",   print_group_list,              0x07, R_GROUP_LIST,      0,  0 },
   { "PR_RGT",  print_user_rights,             0x05, 0,                 2,  2 },
   { "PR_UI",   print_user_information,        0x05, R_USER_READ,       1,  1 },
   { "PR_UL",   print_user_list,               0x05, 0,                 0,  2 },
   { "QUIT",    f4adp_quit,                    0x00, 0,                 0,  0 },
   { "REG_GET", f4adp_reg_get,                 0x03, R_REGISTRY,        1,  1 },
   { "REG_KEY", f4adp_reg_keys,                0x03, R_REGISTRY,        0,  0 },
   { "REG_LS",  f4adp_reg_ls,                  0x03, R_REGISTRY,        0,  0 },
   { "REG_RM",  f4adp_reg_rm,                  0x03, R_REGISTRY,        1,  1 },
   { "REG_SET", f4adp_reg_set,                 0x03, R_REGISTRY,        2, 99 },
   { "RGT_CHK", check_rights,                  0x01, 0,                 2,  2 },
   { "SEC_LS",  f4adp_sec_list,                0x03, R_REGISTRY,        0,  1 },
   { "SEC_SET", f4adp_sec_set,                 0x03, R_REGISTRY,        3, 99 },
   { "SGC_DIR", f4adp_get_section_config,      0x01, 0,                 1, 99 },
   { "SGV_DIR", get_value_by_dir,              0x01, 0,                 2, 99 },
   { "SG_CRED", get_section_credits,           0x01, 0,                 1, 99 },
   { "SG_ID",   get_section_id,                0x01, 0,                 1, 99 },
   { "SG_NAME", get_section_name,              0x01, 0,                 1, 99 },
   { "SIN",     send_startup_info,             0x04, 0,                 0,  0 },
   { "SRV_GST", f4adp_server_get_state,        0x03, R_STATE,           0,  0 },
   { "SRV_SST", f4adp_server_set_state,        0x03, R_STATE,           1,  1 },
   { "SWHO",    user_swho,                     0x01, 0,                 0,  0 },
   { "S_CHG",   f4adp_section_change,          0x03, 0,                 3, 99 },
   { "S_LIST",  section_list,                  0x01, 0,                 2,  2 },
   { "S_RST",   f4adp_section_reset,           0x03, 0,                 3,  3 },
#ifdef TLS
   { "TLS_PRT", get_tls_force_prot,            0x04, 0,                 0,  0 },
#endif
   { "TRF",     print_traffic_stat,            0x01, 0,                 0,  0 },
   { "UAN",     user_anonymous,                0x04, 0,                 1,  1 },
   { "UCR",     user_credit_reduce,            0x05, 0,                 2,  2 },
   { "UIN",     user_information,              0x01, 0,                 0, 99 },
/*   { "ULL",     user_list_loggedin,            0x01, 0,                 0,  1 },  obsolete command, use who/swho */ 
   { "ULP",     user_request_login_permission, 0x00, 0,                 2,  4 },
   { "ULS",     user_list_sections,            0x01, 0,                 1,  2 },
   { "UN",      user_nuke,                     0x05, 0,                 5,  5 },
   { "UPTIME",  server_uptime,                 0x01, 0,                 0,  0 },
   { "USR_ACT", f4adp_usr_activity,            0x01, 0,                 1, 99 },
   { "USR_ADD", f4adp_usr_add,                 0x03, R_USER_ADD,        1,  2 },
   { "USR_ADQ", f4adp_usr_adq,                 0x03, R_USER_ADD,        3, 14 },
   { "USR_CHG", f4adp_usr_chg,                 0x01, 0,                 1, 99 },
   { "USR_CWD", f4adp_usr_cwd,                 0x01, 0,                 2, 99 },
   { "USR_GET", f4adp_usr_get,                 0x01, 0,                 1,  1 },
   { "USR_KIC", f4adp_usr_kick,                0x01, 0,                 1,  2 },
   { "USR_KIL", f4adp_usr_kill,                0x01, 0,                 1,  2 },
   { "USR_LS",  f4adp_usr_ls,                  0x03, R_USER_READ,       0,  0 },
   { "USR_LSR", f4adp_usr_lsr,                 0x05, 0,                 0,  0 },
   { "USR_NFO", f4adp_usr_nfo,                 0x05, 0,                 0,  0 },
   { "USR_PWD", f4adp_usr_pwd,                 0x01, 0,                 1,  2 },
   { "USR_RM",  f4adp_usr_rm,                  0x01, 0,                 1,  1 },
   { "USR_RST", f4adp_usr_rst,                 0x03, R_USER_EDIT_STATS, 1,  2 },
   { "USR_SET", f4adp_usr_set,                 0x01, 0,                 29,29 },
   { "UUD",     user_upload_download,          0x04, 0,                 1, 99 },
   { "UUR",     user_upload_reduce,            0x05, 0,                 8,  8 },
   { "UUU",     user_upload_update,            0x05, 0,                 8,  8 },
   { "VAR_EXP", variable_expand,               0x05, 0,                 1,  1 },
   { "VER",     server_version,                0x01, 0,                 0,  0 },
   { "VER_F4P", server_version_f4adp,          0x00, 0,                 0,  0 },
   { "WA_ADD",  f4adp_watch_add,               0x03, R_MONITOR,         3,  3 },
   { "WA_DEL",  f4adp_watch_del,               0x03, R_MONITOR,         0,  0 },
   { "WHO",     user_who,                      0x01, 0,                 0,  0 },
   { "X_DUPE",  toggle_xdupe,                  0x00, 0,                 0,  1 }              
};

int child_communication(int fd, void *ptr)
{
   struct child_s *c = ptr;
   char           *p[64];
   int             i;
   int             j;
   int             k;
   int             n;
   int             r;

   /* read command (scanner functionality) and handle errors */
   for (;;) {
      switch (read_command(c, p, 60, &n)) {
      case -2:
         /* EOF */
         child_logout(c, CL_NORMAL);
         return 0;
      case -1:
         /* error during fgets() */
         errlog("child_communication(): %s (%i), child_pid=%i, username=%s, logged out",
                strerror(errno), errno, c->pid, c->user ? c->user->name : "???");
         if (c->pid != -1)
                 kill(c->pid, SIGKILL);
         child_logout(c, CL_PIPEERR);
         return 0;
      case 1:
         errlog("child_communication(): too many parameters");
         server_reply(c, "-Too many parameters.");
         return 0;
      case 2:
         errlog("child_communication(): command string of zero length received");
         server_reply(c, "-Command string of zero length received.");
         return 0;
      case 3:
         errlog("child_communication(): buffer overrun");
         server_reply(c, "-Buffer overrun.");
         return 0;
      case 4:
         errlog("child_communication(): no command in command string");
         server_reply(c, "-No command in command string.");
         return 0;
      case 5:
         /*
          * no or incomplete command, return and select()
          * until more data arrives
          */
         return 0;
      }
      time(&c->act_time);
#ifdef DEBUG
      fprintf(debugf, "Child command (n=%i):", n);
      for (i = 0; i < n; i++)
              fprintf(debugf, " %s", *(p + i));
      fprintf(debugf, "\n");
#endif
      /* do binary search of command in command table */
      i = 0;
      k = (sizeof(serverd_cmd) / sizeof(struct serverd_s)) - 1;
      while (i <= k) {
         j = (i + k) >> 1;
         r = strcmp(*p, serverd_cmd[j].command);
         if (r > 0) i = j + 1;
         else if (r < 0) k = j - 1;
         else {  /* command found, now check flags */
            r = serverd_cmd[j].flags;
            if (r & 0x01) {
               if (!c->user) {
                  server_reply(c, "-Permission denied (must be logged in)");
                  break;
               }
            }
            if (r & 0x02) {
               if (!allowed(c->user, serverd_cmd[j].right, NULL)) {
                  server_reply(c, "-Permission denied");
                  break;
               }
            }
            if (r & 0x04) {
               if (c->admin) {
                  server_reply(c, "-Permission denied (ftps only)");
                  break;
               }
            }
            n--;
            if ((n < serverd_cmd[j].min_args) || (n > serverd_cmd[j].max_args)) {
               server_reply(c, "-Syntax error (number of arguments)");
               break;
            }
            /* flag check ok, call command handler and return */
            j = serverd_cmd[j].function(c, p + 1, n);
            if (j < 0) return -1;
            c -= j;
            break;
         }
      }
      if (i > k) {    /* command not found in table */
         errlog("child_communication(): unknown client command (%s)", p[0]);
         server_reply(c, "-Unknown command");
      }
   }
   /* not reached */
   return 0;
}

/* EOF */
