/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: transfer.c - upload / download data transfer
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <regex.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#else
#include <sys/param.h>
#include <sys/mount.h>                                                     
#endif

#ifdef HAVE_STATVFS
#include <sys/statvfs.h>
#define statfs statvfs
#endif

#include "externs.h"
#include "common.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "transfer.h"
#include "utils.h"
#include "cmd_protos.h"

#include "strlcpy.h"
#include "strlcat.h"

#include "crc32.h"
#include "site_vfs.h"

#ifdef TLS
#include "tlsutil.h"

#define FWRITE  tls_fwrite
#define FPUTS   tls_fputs
#define FGETS   tls_fgets
#define READ    tls_read
#define CLOSE   tls_close
#define FCLOSE  tls_fclose
#else
#define FWRITE  fwrite
#define FPUTS   fputs
#define FGETS   fgets
#define READ    read
#define CLOSE   close
#define FCLOSE  fclose
#endif

#define MSG_STORE_DENY       "!BWARNING: Transfers from this host are banned!!!0"
#define MSG_RETRIEVE_DENY    "!BWARNING: Transfers to this host are banned!!!0"
#define TRACEROUTE_ARGS      "-m 8"

int checkroute(unsigned long ip);


void alarmhandler(int sig)
{
   scriptabort = 1;
   longjmp(scriptabortjmp, 1);
}


int cmd_mode (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   if (n != 1) return send_reply(501, "Syntax: MODE <mode_id>");
   if (!strcasecmp(*p, "S")) return send_reply(200, "Mode set to STREAM.");
   if (!strcasecmp(*p, "B")) return send_reply(504, "Block mode not supported.");
   if (!strcasecmp(*p, "C")) return send_reply(504, "Compressed mode not supported.");
   return send_reply(501, "Invalid mode identifier.");
}


int cmd_stru (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   if (n != 1) return send_reply(501, "Syntax: STRU <structure_id>");
   if (!strcasecmp(*p, "F")) {
      structure = S_FILE;
      return send_reply(200, "Structure set to FILE.");
   }
   if (!strcasecmp(*p, "R")) {
      structure = S_RECORD;
      return send_reply(200, "Structure set to RECORD.");
   }
   if (!strcasecmp(*p, "P")) return send_reply(504, "Page structure not supported.");
   return send_reply(501, "Invalid structure identifier.");
}


int cmd_type (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   if (n < 1 || n > 2) return send_reply(501, "Syntax: TYPE <type_id>");
   if (!strcasecmp(*p, "A")) {
      type = T_ASCII;
      return send_reply(200, "Type set to ASCII.");
   }
   if (!strcasecmp(*p, "I")) {
      type = T_IMAGE;
      return send_reply(200, "Type set to IMAGE.");
   }
   if (!strcasecmp(*p, "E")) return send_reply(504, "EBCDI type not supported.");
   if (!strcasecmp(*p, "L")) return send_reply(504, "LOCAL type not supported.");
   return send_reply(501, "Invalid type identifier.");
}


int cmd_mdtm (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   char            dir[256];
   struct stat     st;

   if (n != 1) return send_reply(501, "Syntax: MDTM <filename>");
   if ((add_dir(vfs_getcwd (), *p, dir, sizeof(dir)) == -1) || (stat(dir + 1, &st) == -1)) {
      return send_error(550, "mdtm", errno);
   }
   if (user_locked && *usr.home) {
      if (strncmp(usr.home, dir, strlen(usr.home)))
         return send_reply(550, "!Bno permission.!0");
   }
   if (st.st_mode & S_IFREG) {
      strftime(dir, sizeof(dir), "%Y%m%d%H%M%S", localtime(&st.st_mtime));
      return send_reply(213, dir);
   } 
   else return send_message(550, 1, "%s: not a plain file.", *p);
}


int cmd_size (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   char            dir[256];
   struct stat     st;

   if (n != 1) return send_reply(501, "Syntax: SIZE <filename>");
   if ((add_dir(vfs_getcwd (), *p, dir, sizeof(dir)) == -1) || (stat(dir + 1, &st) == -1)) {
      return send_error(550, "size", errno);
   }
   if (user_locked && *usr.home) {
      if (strncmp(usr.home, dir, strlen(usr.home)))
         return send_reply(550, "!Bno permission.!0");
   }
   if (st.st_mode & S_IFREG) {
      g_snprintf(dir, sizeof(dir), "%u", (unsigned int) st.st_size);
      return send_reply(213, dir);
   } 
   else return send_message(550, 1, "%s: not a plain file.", *p);
}


int cmd_rest (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   int             se = 0;

   if (type == T_ASCII) {
      return send_reply(502, "REST not possible with ASCII type.");
   }
   if (n == 1) {
      if (!sscanf(*p, "%ld", &rest_offset)) se = -1;
   } 
   else se = -1;
   if (se == -1) return send_reply(501, "Syntax: REST <offset>");
   return send_reply(350, "REST command successful.");
}


int cmd_transferpermitted(char *prog, char *buffer, size_t buffersize)
{
   char           str[256];
   int            ec;
   char           *p[8];

   p[0] = prog;
   p[1] = buffer;

   if (run_prog_xt(prog, p, &ec, str, buffer) == -1) return 0;
   if (WIFEXITED(ec)) {
      if (WEXITSTATUS(ec) && (WEXITSTATUS(ec) != -1)) ec = 550;
      else ec = 150;
   } 
   else ec = 150;
   if (ec == 550) {
      g_snprintf(buffer, buffersize, "Transfer initial check failed (%s).", str);
      if (*buffer) {
         if (send_message(ec, ec == 550, buffer, NULL) == -1) return -1;
      }
   }
   return ec == 550 ? 1 : 0;
}


int cmd_retrieve(char *buffer, int buffersize, char *dir, char *name,
                 char *dirandname, FILE * file, FILE ** sockptr, 
		 int max_speed, char **p)
{
   long            down = 0;
   int             err;
   fd_set          fdset_r;
   fd_set          fdset_w;
   int             fd_ctrl;
   int             fd_file;
   int             fd_sock;
   int             hidden;
   int             len;
   int             maxfd;
   int             min_speed;
   int             n;
   FILE           *sock = *sockptr;
   struct sockaddr_in srv_addr;
   struct stat     st;
   struct timeval  t;
   struct timeval  start_time;
   struct timeval  last_check_time; 
   struct timeval  now_time; 
   time_t          t1;
   time_t          t2;
   time_t          t3;
   int             last_check_bytes = 0;
   float           transferspeed = 0.0;
   char            goodname[512];
   int             dir_dl_limit;
   gint64           needed;
#if defined(HIGHERPORT) && defined(LOWERPORT)
   int             i = 0;
   unsigned short  port;
   int             res;
#endif

   min_speed = dl_speed & 0xFFFF;
   if ((max_speed > 0) && (min_speed >= max_speed)) min_speed = max_speed - 1;
   if ((type == T_IMAGE) && rest_offset) {
      if (fseek(file, rest_offset, SEEK_SET) < 0) {
         return send_error(550, "seek", errno);
      }
      rest_offset = 0;
   }
   add_dir(vfs_getcwd (), dirandname, goodname, sizeof(goodname));

   printf("SGV_DIR download_limit %s\n", goodname);
   if (!fgets(buffer, buffersize, stdin)) dir_dl_limit = -1;
   if (sscanf(buffer+1, SI32, &dir_dl_limit) != 1) dir_dl_limit = -1;
   if ((usr.limit < 0) && dir_dl_limit) {
      return send_reply(550, "You have no download permission");
   }

   if (fstat(fileno(file), &st) == -1) st.st_size = 0;
   if (!(st.st_mode & S_IFREG)) {
      return send_message(550, 1, "%s: not a plain file.", name);
   }
   needed = limit_ok(st.st_size, goodname);
   if (needed < 0) {
      g_snprintf(buffer, buffersize, "Not enough credit (you have: " SI64 " KB / you need: %u KB)",
              (needed+st.st_size) / 1024, (unsigned int) st.st_size / 1024);
      return send_reply(550, buffer);
   }
   if (pasvsock == -1) {

      if (checkroute(ntohl(addr.sin_addr.s_addr))) {
         return send_reply(550, MSG_RETRIEVE_DENY);  
      }

      if ((fd_sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
         return send_error(425, "socket", errno);
      }
      if (other_ip_s) {
         memset(&srv_addr, 0, sizeof(srv_addr));
         srv_addr.sin_family = AF_INET;
#if defined(HIGHERPORT) && defined(LOWERPORT)
         port=rand()%(HIGHERPORT-LOWERPORT)+LOWERPORT;
         srv_addr.sin_port = htons(port);
         srv_addr.sin_addr.s_addr = other_ip_s;
         i = 0;

         while ((res = bind(fd_sock, (struct sockaddr *) & srv_addr, 
            sizeof(srv_addr)) < 0) && i <= (HIGHERPORT - LOWERPORT)) {
            i++;
            if(HIGHERPORT==port) port=LOWERPORT;
            srv_addr.sin_port = htons(++port);
         }
         if ((res < 0) || (i > (HIGHERPORT - LOWERPORT))) {
#else
         srv_addr.sin_port = 0;
         srv_addr.sin_addr.s_addr = other_ip_s;
         if (bind(fd_sock, (struct sockaddr *) & srv_addr, sizeof(srv_addr)) < 0) {
#endif
            return send_error(425, "bind", errno);
         }
      }
      if (connect_to(fd_sock, (struct sockaddr *) & addr, sizeof(addr), 30) < 0) {
         return send_error(425, "connect", errno);
      }

      g_snprintf(buffer, buffersize, "Opening %s data connection for %s (%lu bytes).",
        type == T_ASCII ? "ASCII" : "BINARY", name, (unsigned long) st.st_size);
      if (send_reply(150, buffer) == -1) return -1;
   } 
   else {
      len = sizeof(addr);
      fd_sock = accept_to(pasvsock, (struct sockaddr *) & addr, &len, 30);
      if (fd_sock < 0) return send_error(425, "accept", errno);
      if (checkroute(ntohl(addr.sin_addr.s_addr))) {
          close(fd_sock);
          return send_reply(550, MSG_RETRIEVE_DENY);
      }
   
      g_snprintf(buffer, buffersize, "Opening %s data connection for %s (%lu bytes).",
        type == T_ASCII ? "ASCII" : "BINARY", name, 
        (unsigned long) st.st_size);
      if (send_reply(150, buffer) == -1) return -1;
 
#ifdef TLS
      if (tls_on_data) {
        if (tls_accept_data(fd_sock)) {
          close(fd_sock);

          return send_reply(435, "Failed TLS negotiation on data channel");
        }
      }
#endif
   }
   if (type == T_ASCII) {
      err = fcntl(fd_sock, F_GETFL);
      if (err >= 0) err = fcntl(fd_sock, F_SETFL, err | O_NONBLOCK);
      if (err < 0) {
         err = errno;
         CLOSE(fd_sock);
         return send_error(425, "fcntl", err);
      }
   }
   sock = fdopen(fd_sock, type == T_ASCII ? "wt" : "wb");
   if (!sock) {
      err = errno;
      CLOSE(fd_sock);
      return send_error(425, "fdopen", err);
   }
  hidden = directory_flag_set_recursive(dir, DP_HIDDEN);
   *(name - 1) = '/';

   add_dir(vfs_getcwd (), dirandname, buffer, buffersize);
   g_snprintf(goodname, sizeof(goodname), "\"%s\"", buffer);

   log_begin(hidden, LG_TRANSFER, "DOWNLOAD", "%s (%lu bytes)", buffer,
             (unsigned long) st.st_size);

   *(name - 1) = '\0';

   time(&t1);
   FD_ZERO(&fdset_r);
   FD_ZERO(&fdset_w);
   fd_ctrl = fileno(sockfr);
   fd_file = fileno(file);
   maxfd = fd_ctrl > fd_sock ? fd_ctrl + 1 : fd_sock + 1;
   n = 1;
   err = 0;

   gettimeofday(&last_check_time, NULL);
   start_time = last_check_time;

   while (n > 0) {
      FD_SET(fd_ctrl, &fdset_r);
      FD_SET(fd_sock, &fdset_w);
      t.tv_sec = timeout;
      t.tv_usec = 0;
      n = select(maxfd, &fdset_r, &fdset_w, NULL, &t);
      if (!n) {
         send_reply(426, "Timeout, data connection closed");
         n = -1;
         break;
      }
      if (n < 0) {
         send_error(426, "select", err = errno);
         break;
      }
      if (FD_ISSET(fd_ctrl, &fdset_r)) {
         switch (read_ftp_command(buffer, buffersize, p, 4, &n)) {
         case 0:
            if (strcasecmp(*p, "ABOR")) {
               if (send_reply(503, "Unexpected command.") < 0)
               break;
            } else {
               send_reply(426, "Transfer interrupted.");
               send_reply(226, "Transfer successfully aborted.");
            }
         }
         err = 1;
	 n = -1;
	 break;
      } 
      else {
	 if (type == T_ASCII) {
            while (fgets(buffer, buffersize - 2, file)) {
               n = strlen(buffer);
               if (((n == 1) && (*buffer == '\n')) || ((n >= 2) &&
                  (buffer[n - 1] == '\n') && (buffer[n - 2] != '\r'))) {
                  buffer[n - 1] = '\r';
                  buffer[n] = '\n';
                  buffer[n + 1] = '\0';
               }
               if (n) {
                  if (FPUTS(buffer, sock) < 0) {
                     if (errno == EAGAIN) fseek(file, -n, SEEK_CUR);
                     else {
                        err = errno; n = -1;
                     }
                     break;
                  }
               }
	       down += n;
            }
            if (n >= 0) {
               if (feof(file)) n = 0;
               else {
                  if (errno == EAGAIN) {
                     down--;  n = 1;
                  } 
		  else n = -1;
               }
            }
	 } 
	 else {
            n = fread(buffer, 1, buffersize, file);
            if (n > 0) n = FWRITE(buffer, 1, n, sock);
            /* switched back to the pre 3.010 way because
               the download resume does not work on freebsd systems
	       with the two lines below!
	       
               n = read(fd_file, buffer, buffersize);
               if (n > 0) n = write(fd_sock, buffer, n);
            */
	 }
         if (n < 0) send_error(426, "send", err = errno);
         else if (n > 0) {
            down += n;
            /* enforce minimum and maximum speed, t2=time elapsed, t3=time limit */
            if (!fullspeed && (min_speed || max_speed)) {
               t2 = time(NULL) - t1;
               if (min_speed && (t2 >= 30)) {
                  t3 = (down / 1024) / min_speed;
                  if (t2 > t3) {
                     g_snprintf(buffer, buffersize, "Minimum speed of %i kb/s not reached, data connection closed", min_speed);
                     err=1;
                     send_reply(426, buffer);
                     n = -1;
                  }
               }
               if (max_speed) {
                  t3 = (down / 1024) / max_speed;
                  if (t2 < t3) {
                     t3 -= t2;
                     sleep(t3 > 5 ? 5 : t3);
                  }
               }
            }

            /* calculate transfer speed */
            gettimeofday(&now_time, NULL);
            if (((now_time.tv_sec - last_check_time.tv_sec) >= 2)) {
               transferspeed = (down * 1000.0) /
        	  (float) (((now_time.tv_sec - start_time.tv_sec) * 1000.0) +
        	  ((now_time.tv_usec - start_time.tv_usec) / 1000.0));
               update_user_activity(hidden, ACT_DOWNLOAD, transferspeed, goodname);
               last_check_time.tv_sec = now_time.tv_sec;
               last_check_time.tv_usec = now_time.tv_usec;
               last_check_bytes = down;
            }
         }
      }
   }   

   FCLOSE(sock);
   sock = NULL;

   CLOSE(fd_sock);
   
   gettimeofday(&now_time, NULL);

   bytes_downloaded += down;

   printf("UUD 0 %lu 0 %ld %s\n", down, now_time.tv_sec - start_time.tv_sec, goodname);
   scanf(SI64, &usr.credit);
   fgetc(stdin);

   len = path_in_cdpath(dir) == FALSE;
   *(name - 1) = '/';
   add_dir(vfs_getcwd (), dirandname, buffer, buffersize);

   if (err) {
      log_end(hidden, LG_TRANSFER, "DOWNLOAD", "%s FAILED: %s", buffer, strerror(err));
   }
   else {
      log_end(hidden, LG_TRANSFER, "DOWNLOAD", "%s (%lu bytes at %3.1f Kbps)",
              buffer, down, transferspeed / 1024.0);
   }

   /**********************************************
    * increase the download counter of the file  *
    **********************************************/

   if (len) {
      printf("#1#FDN %s\n", buffer);
      fgets(buffer, buffersize, stdin);
   }

   update_user_activity(hidden, ACT_IDLE, 0, vfs_getcwd ());
   return err ? 0 : send_message(226, 1, download_msg.msg, goodname);
}


int cmd_retr (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   char            buffer[8192];
   char           *dir;
   FILE           *file;
   FILE           *sock = NULL;
   char           *name;

   if (n != 1) return send_reply(501, "Syntax: RETR <filename>");
   n = add_dir(vfs_getcwd (), *p, buffer, sizeof(buffer));
   if (user_locked && *usr.home) {
      if (strncmp(usr.home, buffer, strlen(usr.home)))
         return send_reply(550, "!Bno permission.!0");
   }
   split_dir(*p, &dir, &name, 0);
   if (*pre_download) {
      switch (cmd_transferpermitted(pre_download, buffer, sizeof(buffer))) {
         case -1: return -1;
         case  1: return 0;
      }
   }
   file = file_open(dir, name, type == T_ASCII ? "rt" : "rb", PE_READ);
   if (!file) return send_error(550, "read", errno);

   printf("LIM_GBW 0\n");
   if (!fgets(buffer, sizeof(buffer), stdin)) {
      n = errno;
      fclose(file);
      return send_error(550, "pipe error", n);
   }
   if (*buffer == '+') {
      if (!sscanf(buffer + 1, "%i", &n)) n = 0;
   } 
   else n = 0;
   if (!n) {
      fclose(file);
      return send_reply(550, "Transfer not permitted (no bandwidth)");
   }
   if (n < 0) n = 0;

   n = cmd_retrieve(buffer, sizeof(buffer), dir, name, *p, file, &sock, n, p);
   fclose(file);
   if (sock) fclose(sock);

   printf("LIM_RBW\n");
   fgets(buffer, sizeof(buffer), stdin);

   return n;
}


int cmd_stor_check(char *buffer, int bufsize, char *dir, char *name, float speed, unsigned long crc)
{
   struct check_s *ch;
   int             i;
   char           *p[8];
   void           (*oldvec) () = SIG_IGN;
   char           spdstr[10];        
   char           crcstr[10];
   char           fullpath[1024];

   for (i = 0, ch = check; i < check_count; i++, ch++) {
      if (wildcard_matching(ch->wildcard, name, 0)) {
         if (add_dir(vfs_getcwd (), dir, buffer, 512) == -1) return 0;
         i = 0;
         p[i++] = ch->name;
         if (*ch->param) p[i++] = ch->param;

	 g_snprintf(spdstr, sizeof(spdstr), "%.0f", speed);
         g_snprintf(crcstr, sizeof(spdstr), "%08lX", crc);

	 /* arguments for the checker */
         p[i++] = name;
         /* p[i++] = usr.name; */ /* sending user is useless ($F4A_USER!!!) */
	 p[i++] = (char *) &spdstr;
	 if (calc_checksum) {
	    p[i++] = (char *) &crcstr;
         }

         if (checktimeout) {
            if (setjmp(scriptabortjmp)) {
               if (scriptabort == 1) {
                  scriptabort = 0;
                  return 2;
               }
            }
            oldvec = signal(SIGALRM, alarmhandler);
            alarm(checktimeout);
         }
         strlcpy(fullpath, buffer, sizeof(fullpath));
         if (run_program(ch->name, p, i, 0, 0, fullpath + 1, &i, buffer, bufsize
             , dir ? dir : "/") == -1) {
            return 0;
         }
         if (checktimeout) {
            alarm(0);
            signal(SIGALRM, oldvec);
         }
         if (WIFEXITED(i)) {
            if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550;
            else i = 226;
         } 
	 else i = 226;
         if ((i == 550) && !*buffer) strlcpy(buffer, "File is corrupt.", bufsize);
         if (*buffer) {
            if (send_message(i, i == 550, buffer, NULL) == -1) return -1;
         }
         if (i == 550) {
            add_dir(vfs_getcwd (), dir, buffer, bufsize);
            add_dir(buffer, name, buffer, bufsize);
            unlink(buffer + 1);
         }
         return i == 550 ? 1 : 0;
      }
   }
   return 0;
}


int cmd_stor_dcheck(char *buffer, int bufsize, char *name)
{
   struct check_s *ch = dcheck;
   int             i;
   char           *p[8];

   for (i = 0; i < dcheck_count; i++, ch++) {
      if (wildcard_matching(ch->wildcard, name, 0)) {
         i = 0;
         p[i++] = ch->name;
         if (*ch->param) p[i++] = ch->param;
         p[i++] = buffer + 1;
         p[i++] = usr.name;
         if (run_program(ch->name, p, i, 0, 0, NULL, &i, buffer, bufsize, buffer
             ) == -1) {
            return 0;
         }
         if (WIFEXITED(i)) {
            if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550;
            else i = 150;
         } 
	 else i = 150;
         if ((i == 550) && !*buffer) strlcpy(buffer, "Dupe check failed.", bufsize);
         if (*buffer) {
	    if (send_message(i, i == 550, buffer, NULL) == -1) return -1;
         }
         return i == 550 ? 1 : 0;
      }
   }
   return 0;
}


int internal_dcheck(char *buf, char *name, char *dir)
{
   regex_t        preg;
   char           str[128];
   FILE           *fp;

   printf("SGV_DIR internal_dupecheck %s\n", buf ? buf : "/");
   if (!fgets(str, sizeof(str), stdin)) return 0;
   if (*str == '-' || *(str+1) == '0') return 0;

   fp = file_open(dir, name, "r", 0);
   if (fp) {
      fclose(fp);
      if (send_message(550, 1, "!BWARNING: File already exists in this directory.!0", NULL) == -1) {
         return -1;
      }
      return 1;        
   }     
   
   printf("SGV_DIR allow_dupe %s\n", buf ? buf : "/");
   if (!fgets(str, sizeof(str), stdin)) return 0;
   if (*str == '-') return 0;

   str[strlen(str) - 1] = '\0';
   if (regcomp(&preg, str+1, REG_EXTENDED|REG_NEWLINE) == 0) {
      if (regexec(&preg, name, 0, NULL, 0) == 0) {
         regfree(&preg);
	 return 0;
      }
      regfree(&preg);
   }
   if (send_message(550, 1, " !BFiletype is not allowed in this section.!0", NULL) == -1) {
      return -1;
   }
   return 1;
}

int dupecheck(char *filename, unsigned long crc, FILE* sock)
{
   char str[1024];
   char str2[32];
   char dupe_user[10];
   char dupe_group[10];
   long dupe_age;
   char dupe_path[1024];

   printf("DUP_CHK %lu \"%s\"\n", crc, filename);
   if (fgets(str, sizeof(str), stdin)) {
      if (*str != '+') {
//	 if (sock) fclose(sock);
	 sscanf(str+1, "%9s\t%9s\t%ld\t%s", dupe_user, dupe_group, 
	    &dupe_age, dupe_path);
         sprint_difftime(str2, sizeof(str2), (double) dupe_age);

	 send_message(553, 0, "!BDupe found! Transfer was interrupted.!0", NULL);
         if (dupe_age != -1) {
  	    g_snprintf(str, sizeof(str), "!HFile was uploaded by %s (%s) %s ago.!0",
	       dupe_user, dupe_group, str2);
	 }
	 else {
  	    g_snprintf(str, sizeof(str), "!HFile is currently uploaded by %s (%s).!0",
	       dupe_user, dupe_group);
	 }	 
	 send_reply(553, str);
	 return 1;
      }
   }
   return 0;             
}


/************************************************************* 
 * first check if dupecheck is enabled for this section      *  
 * then check if the filename matches the regular expression * 
 * set for key dupe_types in the section config.             * 
 * if the name is matching 1 is returned in any other        *  
 * case including all errors, 0 is returned.                 * 
 *************************************************************/    
int is_dupecheck_needed(char *path, char *filename)
{
   regex_t  preg;
   char     str[128];

   printf("SGV_DIR internal_dupecheck %s\n", path ? path : "/");
   if (!fgets(str, sizeof(str), stdin)) return 0;
   if (*str == '-' || *(str+1) == '0') return 0;

   printf("SGV_DIR dupe_filetypes %s\n", path ? path : "/");
   if (!fgets(str, sizeof(str), stdin)) return 0;
   if (*str == '-') return 0;

   str[strlen(str) - 1] = '\0';
   if (regcomp(&preg, str+1, REG_EXTENDED|REG_NEWLINE) == 0) {
      if (regexec(&preg, filename, 0, NULL, 0) == 0) {
         regfree(&preg);
	 return 1;
      }
      regfree(&preg);
   }
   return 0;
}



int cmd_store(char *buffer, int buffersize, char *dir, char *name,
              char *dirandname, FILE **file, FILE **sockptr, int max_speed,
              char **p, struct stat *st)
{
   int             err;
   fd_set          fdset_r;
   int             fd_ctrl;
   int             fd_file;
   int             fd_sock;
   int             hidden;
   int             maxfd;
   int             min_speed;
   int             n;
   int             ok = 0;
   FILE            *sock = *sockptr;
   FILE            *fp   = *file;
   struct timeval  t;
   struct timeval  last_check_time; 
   struct timeval  now_time;
   struct timeval  start_time;
   int             last_check_bytes = 0;
   float           transferspeed = 0.0;
   unsigned long   up = 0;
   guint32         crc = 0;        /* crc of the whole file      */
   guint32         crc2 = 0;       /* crc of the first block     */
   char            str[128];
   int             firstloop = 1;  /* are we in the first loop ? */
   char            *fullpath = NULL;
   int             dupefound = 0;
   int             dflag = 1;
    
   min_speed = ul_speed & 0xFFFF;
   if ((max_speed > 0) && (min_speed >= max_speed)) min_speed = max_speed - 1;

   /* Added by Reboot */
   if ((type == T_IMAGE) && rest_offset) { 
      if (fseek(fp, rest_offset, SEEK_SET) < 0) {
         return send_error(550, "seek", errno);
      }
      rest_offset = 0;
      dflag = 0; /* since an offset was provided we are resuming hence no dupe check is needed */
   }

   if (pasvsock == -1) {
      /* ---- check if source host is allowed ---- */
      if (checkroute(ntohl(addr.sin_addr.s_addr))) {
         add_dir(vfs_getcwd (), name, buffer, buffersize);
         unlink(buffer + 1);
	      
         return send_reply(550, MSG_STORE_DENY);
      }	
      if ((fd_sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
         return send_error(425, "socket", errno);
      }
      if (connect_to(fd_sock, (struct sockaddr *) & addr, sizeof(addr), 30) < 0) {
         return send_error(425, "connect", errno);
      }

      g_snprintf(buffer, buffersize, "Opening %s data connection for %s",
        type == T_ASCII ? "ASCII" : "BINARY", name);
      if (send_reply(150, buffer) == -1) return -1;
   } 
   else {
      ok = sizeof(addr);
      fd_sock = accept_to(pasvsock, (struct sockaddr *) & addr, &ok, 30);
      if (fd_sock < 0) return send_error(425, "accept", errno);
      /* check if source hosts is allowed */
      if (checkroute(ntohl(addr.sin_addr.s_addr))) {
         close(fd_sock);
         add_dir(vfs_getcwd (), name, buffer, buffersize);
         unlink(buffer + 1);
         return send_reply(550, MSG_STORE_DENY);
      }
   
      g_snprintf(buffer, buffersize, "Opening %s data connection for %s",
        type == T_ASCII ? "ASCII" : "BINARY", name);
      if (send_reply(150, buffer) == -1) return -1;

#ifdef TLS
      if (tls_on_data) {
        if (tls_accept_data(fd_sock)) {
          close(fd_sock);

          return send_reply(435, "Failed TLS negotiation on data channel");
        }
      }
#endif
   }

   if (type == T_ASCII) {
      err = fcntl(fd_sock, F_GETFL);
      if (err >= 0) err = fcntl(fd_sock, F_SETFL, err | O_NONBLOCK);
      if (err < 0) {
         err = errno;
         CLOSE(fd_sock);
         return send_error(425, "fcntl", err);
      }
   }

   sock = fdopen(fd_sock, type == T_ASCII ? "rt" : "rb");
   if (!sock) {
      ok = errno;
      CLOSE(fd_sock);
      return send_error(425, "fdopen", ok);
   }

   hidden = directory_flag_set_recursive(dir, DP_HIDDEN);
   // *(name - 1) = '/';

   add_dir(vfs_getcwd (), name, buffer, buffersize);
   fullpath = g_strdup(buffer);
   if(fullpath == NULL)
	errlog("Not enough memory for g_malloc of filename in cmd_store()");
/*
   fullpath = g_malloc(strlen(buffer) + sizeof(char));
   if (fullpath) strcpy(fullpath, buffer);
   else errlog("Not enough memory for g_malloc of filename in cmd_store()");
*/

   /******************************************/
   /* get some flags from the section config */
   /******************************************/
   printf("SGV_DIR calc_checksum %s\n", buffer);
   if (fgets(str, sizeof(str), stdin)) {
      if (strncmp(str, "+0", 2) != 0) calc_checksum = 1;
   }
    
   /* only perform dupe check if the dflag hasn't been flipped to false
    * because the client provided an offset (we are resuming) */
   if (dflag) {
      dflag = is_dupecheck_needed(dir, name);
   }

   log_begin(hidden, LG_TRANSFER, "UPLOAD", "%s", fullpath);

   FD_ZERO(&fdset_r);
   fd_ctrl = fileno(sockf);
   fd_file = fileno(fp);
   maxfd = fd_ctrl > fd_sock ? fd_ctrl + 1 : fd_sock + 1;
   n = 1;
   err = 0;

   /* init timers */
   gettimeofday(&last_check_time, NULL);
   start_time = last_check_time;

   /* main transfer loop */
   while (n > 0) {
      FD_SET(fd_ctrl, &fdset_r);
      FD_SET(fd_sock, &fdset_r);
      t.tv_sec = timeout;
      t.tv_usec = 0;
      n = select(maxfd, &fdset_r, NULL, NULL, &t);
      if (!n) {
         send_reply(426, "Timeout, data connection closed");
#ifdef DELINCOMPLETE
	 /* delete if timeout added by poing */
         unlink(fullpath + 1);
#endif
         break;
      }
      if (n < 0) {
         send_error(426, "select", err = errno);
         n = -1;
         break;
      }

      if (FD_ISSET(fd_ctrl, &fdset_r)) {
         switch (read_ftp_command(buffer, buffersize, p, 4, &n)) {
            case 0:
               if (strcasecmp(*p, "ABOR")) {
        	  send_reply(503, "Unexpected command.");
#ifdef DELINCOMPLETE
	          unlink(fullpath + 1);
#endif
               } 
	       else {
        	  send_reply(426, "Transfer interrupted.");
        	  send_reply(226, "Transfer successfully aborted.");
#ifdef DELINCOMPLETE
	          unlink(fullpath + 1);
#endif
	       }
         }
         err = 1;
         n = -1;
         break;
      } 
      else {
         if (type == T_ASCII) {
            while (FGETS(buffer, buffersize, sock)) {
               n = strlen(buffer);
               if ((n >= 2) && (buffer[n - 1] == '\n') && (buffer[n - 2] == '\r')) {
                  buffer[--n] = '\0';
                  buffer[n - 1] = '\n';
               }
               if (n) {
                  if (fputs(buffer, fp) < 0) {
                     send_error(452, "write", err = errno);
                     if (err == EAGAIN) errno = EIO;
                     n = -1;
                     break;
                  }
               }
               up += n;
            }
            if (n >= 0) {
               if (feof(sock)) n = 0;
               else {
                  if (errno == EAGAIN) {
                     up--;
                     n = 1;
                  } 
		  else  n = -1;
               }
            }
         } 
	 else {
            n = READ(fd_sock, buffer, buffersize);
            if (n < 0) send_error(452, "recv", err = errno);
            else if (n > 0) {
               ok = write(fd_file, buffer, n);
               if (ok < 0) {
                  send_error(452, "write", err = errno);
                  n = -1;
               } 
	       else if (ok < n) {
                  send_error(452, "write", err = ENOSPC);
                  n = -1;
               }
               /*********************************** 
	        * on the fly checksum calculation * 
	        ***********************************/
	       if (calc_checksum || firstloop) {
		 crc = crc32_calc (crc, buffer, n);
               }
            }
         }

         if (n > 0) {
            up += n;

	    /*************************************************** 
	     * dupecheck: filename and crc of the first block  *
	     * are checked against dupe database, if dupe was  *
	     * found the transfer is aborted.                  * 
             ***************************************************/
	    if (firstloop) {
               if (dflag) {
        	  if (dupecheck(name, crc2, sock)) {	
                     err = 1;
		     dupefound = 1;  
		     break;
		  }             
               }
            }

	    /***************************************************
	     * enforce minimum and maximum speed limits        *
	     * t = elapsed time                                *  
             ***************************************************/
	    if (!fullspeed && (min_speed || max_speed)) {
               gettimeofday(&now_time, NULL);
	       t.tv_sec = now_time.tv_sec - start_time.tv_sec;
	       t.tv_usec = now_time.tv_usec - start_time.tv_usec;     
	       if (min_speed && (t.tv_sec >= 30)) {
	          if (t.tv_sec > ((up / 1024) / min_speed)) {
	  	     g_snprintf(buffer, buffersize, "Minimum speed of %i Kb/s not reached, "
		        "data connection closed", min_speed);
                     send_reply(426, buffer);
                     err = 1;
		     break;	     
		  }       
	       }
	       if (max_speed) {
	          if (t.tv_sec < ((up / 1024) / max_speed)) {
		      t.tv_sec =  ((up / 1024) / max_speed) - t.tv_sec;
		      usleep(t.tv_sec > 5 ? 5 * 1000000 : (t.tv_sec * 
		         1000000 + t.tv_usec));
		  }	       
	       }
            }

            /*****************************************************
	     * calculate transfer speed of the current transfer  *
	     * the speed is an average over the data transferred *
	     * yet and not of the last transferred block         *
	     *****************************************************/
            gettimeofday(&now_time, NULL);
            if (((now_time.tv_sec - last_check_time.tv_sec) >= 2)) {
               transferspeed = (up * 1000.0) /
        	  (float) (((now_time.tv_sec - start_time.tv_sec) * 1000.0) +
        	  ((now_time.tv_usec - start_time.tv_usec) / 1000.0));
               update_user_activity(hidden, ACT_UPLOAD, transferspeed, fullpath);
               last_check_time.tv_sec = now_time.tv_sec;
               last_check_time.tv_usec = now_time.tv_usec;
               last_check_bytes = up;
            }
         }
      }
      firstloop = 0;
   }   

   transferspeed = (up * 1000.0) /
      (float) (((now_time.tv_sec - start_time.tv_sec) * 1000.0) +
      ((now_time.tv_usec - start_time.tv_usec) / 1000.0));

   /* close the local file and the socket */
   fclose(fp);
   fp = NULL;
   FCLOSE(sock);
   sock = NULL;
   CLOSE(fd_sock);

#ifdef DELINCOMPLETE
   /* if filesize=0 then delete (added by poing) */
   if (up == 0) {unlink(fullpath + 1);}
#endif

   gettimeofday(&now_time, NULL);
  
   if (!err) {           /* upload succeeded */
     if ((ok = cmd_stor_check(buffer, buffersize, dir, name, transferspeed, crc)) == 2) 
       {
	 send_message(226, 0, "Upload checker timeout. File not tested.", NULL);
         ok = 0;
      }
      if (!ok) {
	/* add file to dupelist with the crc of the first block */
	printf("DUP_ADD %u \"%s\"\n", crc2, fullpath);
	fgets(str, sizeof(str), stdin);
      }
   } else {              /* upload failed */
      // nukebrokenupload is section setting now
      printf("SGV_DIR nukebrokenuploads %s\n", fullpath ? fullpath : "/");
      
      /* check the return value of the key */
      if (fgets(str, sizeof(str), stdin)) {
	 if ((*str == '+') && (*(str + 1) == '1')) {
	    ok = 1;
	 } else {
	    ok = 0;
	 }
      } else {
	 ok = 0;
      }
   }
   
   /****************************************************************
    * if an error occured while transfer and the nukebrokenupload  *
    * flag is enabled or if the upload checker failed we delete    *
    * the file else we update the user stats and credit            *
    ****************************************************************/
   if (ok || dupefound) {
      unlink(fullpath+1);
   }
   else {
      bytes_uploaded += up - st->st_size;
      if (up < (unsigned long) st->st_size) {  /* on REST */
	 printf("UUD -%lu 0 %ld 0 %s\n", (unsigned long) st->st_size - up, 
	    now_time.tv_sec - start_time.tv_sec, dir ? dir : "/");
      }
      else {
         printf("UUD %lu 0 %ld 0 %s\n", up - (unsigned long) st->st_size, 
	   now_time.tv_sec - start_time.tv_sec, dir ? dir : "/");
      }
      if (fgets(str, sizeof(str), stdin)) {
	 if (*str == '+') {
	    sscanf(str, SI64, &usr.credit);
	 }
      }
   }

   if (!n) st->st_size = -1;       /* OK FLAG */

   maxfd = path_in_cdpath(dir) == FALSE;

   /************************ 
    * add entry to logfile *
    ************************/
   if (err) {
      log_end(hidden, LG_TRANSFER, "UPLOAD", "%s FAILED: %s", fullpath, strerror(err));
   }
   else {
      log_end(hidden, LG_TRANSFER, "UPLOAD", "%s (%lu bytes at %3.1f Kbps [CRC=%08X])",
         fullpath, up, transferspeed / 1024.0, crc);
   }

   update_user_activity(hidden, ACT_IDLE, 0, vfs_getcwd ());

   if (ok) {
      g_free(fullpath);
      return ok;
   }

   /************************************************
    * update the permission file with the new file *
    ************************************************/
   if (maxfd) {
      printf("#1#FUP %s\n", fullpath);
      fgets(buffer, buffersize, stdin);
   }

   g_snprintf(buffer, buffersize, "\"%s\"", fullpath);
   g_free(fullpath);
   return err ? 0 : send_message(226, 1, upload_msg.msg, buffer);
}



int cmd_appe_stor(char **p, int n, int append)
{
   char           buffer[4096];
   char           *dir;
   FILE           *file;
   FILE           *sock = NULL;
   struct stat     st;
   char           *name;
   char           *hfile;
   char           fullpath[1024];
   int            size = 0;
   char           str[1024];

   struct statfs  sfs;

   fflush(stdin);
   printf("X_DUPE\n");
   if (!fgets(buffer, sizeof(str), stdin)) {
      return send_error(550, "pipe error", errno);
   }
   if (*buffer == '+') {
      if (!sscanf(buffer + 1, "%i", &n)) n = 0;
   }
   /* only send it if user enabled xdupe */
   if(n) send_xdupe(n);

   convert_filename(*p, fconv_mode & 0x0F);
   n = add_dir(vfs_getcwd (), *p, str, sizeof(str));
   if (n < 0) return send_error(550, "open", errno);
   if (user_locked && *usr.home) {
      if (strncmp(usr.home, str, strlen(usr.home)))
         return send_reply(550, "!Bno permission.!0");
   }
   if (append || (stat(str + 1, &st) < 0)) st.st_size = 0;
   memset(fullpath, 0, sizeof(fullpath));
   strlcpy(fullpath, str, sizeof(fullpath));

   split_dir(str, &dir, &name, 0);
   if (strlen(name) >= MAXFILENAMELENGTH)
      return send_error(550, "open", ENAMETOOLONG);

   if ((permission_ok(dir, PE_WRITE) == -1)
       || (!superuser() && (!strcmp(name, readme_file)))
       || (!strcmp(name, permission_file)))
     return send_error(550, "write", EACCES);

   strlcpy(buffer, hidden_files, sizeof(buffer));
   hfile = strtok(buffer, " \t");
   while (hfile != NULL) {
      if (!strcmp(name, hfile)) return send_error(550, "write", EACCES);
      hfile = strtok(NULL, " \t");
   }

   /* check for free space */
   if (min_freemb > 0 && statfs((dir && strlen(dir) > 1) ? dir+1 : ".", &sfs) != -1) {
      if (((sfs.f_bavail / 1024) * sfs.f_bsize / 1024) < min_freemb) {
         return send_reply(550, "!BTransfer not permitted (no diskspace)!0");
      }
   }

   if (*pre_upload) {
      switch (cmd_transferpermitted(pre_upload, str, sizeof(str))) {
         case -1: return -1;
         case  1: return 0;
      }
   }   
   if (!append) {
      /* buffer is destroyed, thats why we use it only temp. */
      strlcpy(buffer, fullpath, sizeof(buffer));
      switch (internal_dcheck(buffer, name, dir)) {
          case -1: return -1;
          case  1: return 0;
      }    
      strlcpy(buffer, fullpath, sizeof(buffer));
      switch (cmd_stor_dcheck(buffer, sizeof(buffer), name)) {
          case -1: return -1;
          case  1: return 0;
      }
   }
   fflush(stdin);
   printf("LIM_GBW 1\n");
   if (!fgets(buffer, sizeof(str), stdin)) {
      return send_error(550, "pipe error", errno);
   }
   if (*buffer == '+') {
      if (!sscanf(buffer + 1, "%i", &n)) n = 0;
   } 
   else {
     n = 0;
   }

   if (!n)
     return send_reply(550, "transfer not permitted (no bandwidth)");
   if (n < 0)
     n = 0;
   /* if (append) file = file_open(dir, name, type == T_ASCII ? "at" : "ab", PE_WRITE); The original line */

   /* Changed by Reboot */
   if (append || rest_offset)
     file = file_open(dir, name, type == T_ASCII ? "at" : "ab", PE_WRITE);
   else
     file = file_open(dir, name, type == T_ASCII ? "wt" : "wb", PE_WRITE);

   if (!file) {
       n = errno;
       printf("LIM_RBW\n");
       fgets(buffer, sizeof(buffer), stdin);
       return send_error(550, "write", n);
   }

   if (!path_in_cdpath (fullpath)) {
      printf("#1#FCO " SI32 " -2 %s\n", usr.uid, fullpath);
      fgets(buffer, sizeof(buffer), stdin);
   }

   n = cmd_store(buffer, sizeof(buffer), dir, name, fullpath, &file, &sock, n, p, &st);

   printf("LIM_RBW\n");
   fgets(buffer, sizeof(buffer), stdin);
   return n;
}


int cmd_stor (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   if (n != 1)
     return send_reply(501, "Syntax: STOR <filename>"); 
  
   return cmd_appe_stor(p, n, rest_offset ? 1 : 0);
}


int cmd_appe (gchar **p, gint n, gchar *_buffer, gint _buflen)
{
   if (n != 1) 
     return send_reply(501, "Syntax: APPE <filename>");
   
   return cmd_appe_stor(p, n, 1);
}


int cmd_port(char **p, int n, char *cmd, int cmdlen)
{
   int             a1;
   int             a2;
   int             a3;
   int             a4;
   int             p1;
   int             p2;
   char            str[32];

   if (n != 1) {
      return send_reply(501, "Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
   }
   rest_offset = 0;
   if (sscanf(*p, "%i,%i,%i,%i,%i,%i", &a1, &a2, &a3, &a4, &p1, &p2) != 6) {
      return send_reply(501, "Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
   }
   if (a1 < 0 || a1 > 255 || a2 < 0 || a2 > 255 || a3 < 0 || a3 > 255 || 
       a4 < 0 || a4 > 255 || p1 < 0 || p1 > 255 || p2 < 0 || p2 > 255)
   {       
      return send_reply(501, "Syntax: PORT <a1>,<a2>,<a3>,<a4>,<p1>,<p2>");
   }
   g_snprintf(str, sizeof(str), "%i.%i.%i.%i", a1, a2, a3, a4);
   memset(&addr, 0, sizeof(addr));
   addr.sin_addr.s_addr = inet_addr(str);
   if (addr.sin_addr.s_addr == -1) return send_reply(501, "Invalid address.");
   p1 = (p1 << 8) + p2;
   addr.sin_port = htons((u_short) p1);
   addr.sin_family = AF_INET;
   if (pasvsock != -1) close(pasvsock);
   pasvsock = -1;
   return send_reply(200, "PORT command successful.");
}


int cmd_pasv(char **p, int n, char *cmd, int cmdlen)
{
   struct sockaddr_in  addr;
   int                 i;
   unsigned long       ipaddr;
   unsigned short      port;
   char                str[128];
#if defined(HIGHERPORT) && defined(LOWERPORT)
   int                 res, j = 0;
#endif
   char                *g = NULL;

   if (!flag_get(flags, FLAG_PASSIVEMODE)) {
      return send_reply(502, "Passive mode is disabled.");
   }
   if (n) return send_reply(501, "Syntax: PASV");
   rest_offset = 0;
   if (pasvsock >= 0) close(pasvsock);
   pasvsock = -1;
   i = sizeof(addr);
   if (getsockname(sock, (struct sockaddr *) & addr, &i) < 0) {
      return send_error(550, "getsockname", errno);
   }
   pasvsock = socket(PF_INET, SOCK_STREAM, 0);
   if (pasvsock < 0) return send_error(550, "socket", errno);
   addr.sin_family = AF_INET;

#if defined(HIGHERPORT) && defined(LOWERPORT)
   port=rand()%(HIGHERPORT-LOWERPORT)+LOWERPORT;
   addr.sin_port = htons(port);
   while ((res = bind(pasvsock, (struct sockaddr *) & addr, 
      sizeof(addr)) < 0) && j <= (HIGHERPORT - LOWERPORT)) {
      j++;
      if(HIGHERPORT==port) port=LOWERPORT;
      addr.sin_port = htons(++port);
   }
   if ((res < 0) || (j > (HIGHERPORT - LOWERPORT))) {
#else
   addr.sin_port = 0;
   if (bind(pasvsock, (struct sockaddr *) & addr, sizeof(addr)) == -1) {
#endif
      i = errno;
      close(pasvsock);
      pasvsock = -1;
      return send_error(550, "bind", i);
   }

   if (listen(pasvsock, 1) < 0) {
      i = errno;
      close(pasvsock);
      pasvsock = -1;
      return send_error(550, "listen", i);
   }
   i = sizeof(addr);
   if (getsockname(pasvsock, (struct sockaddr *) & addr, &i) == -1) {
      i = errno;
      close(pasvsock);
      pasvsock = -1;
      return send_error(550, "getsockname", i);
   }

   /* should we use the special ip of the nat router ? */
   g = get_groupname(usr.gid);
   if (passive_ip_s) {
        /* user is not in the group "local" */
	if (!g || (g && strcmp(g, "local"))) {
   	   ipaddr = ntohl(passive_ip_s);
	}
	else {
   	   ipaddr = ntohl(addr.sin_addr.s_addr);	
	}
   }
   else {
   	ipaddr = ntohl(addr.sin_addr.s_addr);
   }

   port = ntohs(addr.sin_port);
   g_snprintf(str, sizeof(str), "%lu,%lu,%lu,%lu,%hu,%hu", (ipaddr >> 24) & 0xFF,
      (ipaddr >> 16) & 0xFF, (ipaddr >> 8) & 0xFF, ipaddr & 0xFF,
      (port >> 8) & 0xFF, port & 0xFF);
   if (send_message(227, 1, "Entering Passive Mode (%s).", str) < 0) {
      close(pasvsock);
      pasvsock = -1;
      return -1;
   }
   return 0;
}


/* ---------------------------------------------------------------------------
  checkroute - check for banned hosts using traceroute, checked ip's are
  stored in a good/bad log so we don't have to check them twice
----------------------------------------------------------------------------*/

int checkroute(unsigned long ip)
{
   char            str[256];
#ifdef TRACEROUTE_CHECK 
   FILE            *file;
   char            ppath[256];
   char            ipstr[16];
   int             pfd[2];
   pid_t           pid;
   unsigned char   *buffer;
   char            *hop;
   int             found = 0;  
#endif

   printf("IP_CHK %ld\n", ip);
   if (!fgets(str, sizeof(str), stdin)) {
      return send_error(550, "pipe error", errno);
   }
   if (strcmp(str, "+allowed\n") == 0) {
      return 0;
   } 
   else if (strcmp(str, "-denied\n") == 0) {
      return 1;
   }

#ifndef TRACEROUTE_CHECK 
   return 0;
#else   
   /* ip was never checked before */
   strlcpy(ppath, TRACEROUTE_BIN, sizeof(ppath));
   g_snprintf(ipstr, sizeof(ipstr), "%lu.%lu.%lu.%lu",(ip >> 24) & 0xFF,(ip >> 16) & 0xFF,(ip >> 8) & 0xFF,ip & 0xFF);
 
   buffer = g_malloc(20480);
   if(buffer == NULL) return -1;
   if(socketpair(AF_UNIX, SOCK_STREAM, 0, pfd) == -1) return -1;                                                       
   fflush(stdout);

   pid = fork();
   /* error on fork */
   if(pid == -1) {
      close(pfd[0]);
      close(pfd[1]);
      g_free(buffer);
      return -1;
   }

   /* parent process */
   if (pid) {
      close(pfd[1]);                                                                                               
      file = fdopen(pfd[0], "rt");
      if (!file) {
         close(pfd[0]);
         g_free(buffer);
         return -1;
      }
      ppath[0] = '\0';
      while(fgets(ppath, sizeof(ppath), file)) strcat(buffer, ppath);
      fclose(file);
      close(pfd[0]);
      wait((int *) &pid);
  
      /* search for all ip's set in reject_hops in the output from the traceroute scan */
      strlcpy(str, rejected_hops, sizeof(str));
      hop = strtok(str, " \t");
      while (hop) {
         if (strstr(buffer, hop)) found = 1;
 	 hop = strtok(NULL, " \t");
      }
     
      /* ip is not allowed ==> add to deny list */
      if(found) {
         printf("IP_ADD %s\n", ipstr);
         if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-pipe error", sizeof(str));
         /* if (*str == '-') send_reply(550, str+1); */
	 return 1;
      } 
      /* ip is allowed ==> remove from deny list (add to allow list) */
      else {
         printf("IP_RM %s\n", ipstr);
         if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-pipe error", sizeof(str));
         /* if (*str == '-') send_reply(550, str+1); */
         return 0;
      }
   } 
   /* child process */
   else {
      g_free(buffer);
      fclose(sockf);
      fclose(sockfr);
      close(pfd[0]);
      dup2(pfd[1], 0);
      dup2(pfd[1], 1);
      dup2(pfd[1], 2);
      close(pfd[1]);
      execl(ppath, ppath, TRACEROUTE_ARGS, ipstr, NULL); 
      exit(-1);
   }
#endif
}

 void send_xdupe(gint mode) {
   regex_t preg;
   DIR *dir = NULL;
   struct dirent *fname = NULL;
   gchar buffer[1025];
   gchar *dynbuffer = NULL;
   size_t buflen;
   gchar xdupe_mask[256];

   if((dir = opendir(vfs_get_realpath(""))) == NULL)
     return;

   if((fname = g_malloc(sizeof(struct dirent))) == NULL) {
     closedir(dir);
     return;
   }

   fflush(stdin);
   printf("REG_GET xdupe_mask\n");
   if (!fgets(buffer, sizeof(buffer), stdin)) {
     closedir(dir);
     return;
   }
   if (*buffer == '+') {
     if (!sscanf(buffer + 1, "%s", xdupe_mask)) {
       closedir(dir);
       return;
     }
   }

   memset(buffer, 0, sizeof(buffer));
   while((fname = readdir(dir))) {
     /* match global x-dupe pattern */
     if (regcomp(&preg, xdupe_mask, REG_EXTENDED|REG_NEWLINE))
       break;

     if (regexec(&preg, fname->d_name, 0, NULL, 0)) {
       regfree(&preg);
       continue;
     }
     regfree(&preg);

     if(mode == 1) {
       /* setup a buffer to hold 66 chars */
       buflen = sizeof(char) * 67;

       if(buffer[0] == '\0')
	 g_snprintf(buffer, buflen, "X-DUPE: ");

       /* there's not enough room to fit filename in buffer
	* so send current line and move to next filename
	* unless filename is too long to fit on line at all
	* in that case truncate it to fit and start new line
	*/
       /*
       if(strlen(fname->d_name) > buflen - sizeof(char) * 2 - strlen(buffer) && !(strlen(fname->d_name) > buflen - sizeof(char) * 2)) {
	 strlcat(buffer, fname->d_name, buflen - sizeof(char));
	 strlcat(buffer, "\n", buflen);
	 send_reply_full(553, FALSE, FALSE, FALSE, buffer);

	 g_snprintf(buffer, buflen, "X-DUPE: %s", fname->d_name);
	 continue;
       }
       */

       /* buffer is full, send it
	* since we have to add a trailing newline, reserve one character for that
	*/
       if(strlcat(buffer, fname->d_name, buflen - sizeof(char)) >= buflen) {
	 strlcat(buffer, "\n", buflen);
	 send_reply_full(553, FALSE, FALSE, TRUE, buffer);
	 memset(buffer, 0, sizeof(buffer));
       }
       else
	 strlcat(buffer, " ", buflen - sizeof(char));

       continue;
     }

     /* send one line pr filename. again filenames longer than 66 chars will be truncated */
     else if(mode == 2) {
       /* setup a buffer to hold 66 chars */
       buflen = sizeof(char) * 67;

       /* just to be sure..you never know */
       g_snprintf(buffer, buflen, "X-DUPE: %s\n", fname->d_name);
       send_reply_full(553, FALSE, FALSE, TRUE, buffer);
       memset(buffer, 0, sizeof(buffer));
       continue;
     }

     /* send one line pr filename. no truncating filenames, only system memory sets the limits */
     else if(mode == 3) {
       if(!(dynbuffer = g_strdup_printf("X-DUPE: %s\n", fname->d_name))) {
	 break;
       }	 
       send_reply_full(553, FALSE, FALSE, TRUE, dynbuffer);
       g_free(dynbuffer);
       continue;
     }

     /* cache up filenames up to 1024 bytes and discard the rest and send it in one line */
     else if(mode == 4) {
       buflen = sizeof(char) * 1025;

       if(buffer[0] == '\0')
	 g_snprintf(buffer, buflen, "X-DUPE: ");

       /* buffer's full. send it and skip the rest of the dir
	* since we have to add a trailing newline, reserve one character for that */
       if(strlcat(buffer, fname->d_name, buflen - sizeof(char)) >= buflen - sizeof(char)) {
	 strlcat(buffer, "\n", buflen);
	 send_reply_full(553, FALSE, FALSE, TRUE, buffer);
	 memset(buffer, 0, sizeof(buffer));
	 break;
       }
       else
	 strlcat(buffer, " ", buflen - sizeof(char));

       continue;
     }
   }
   free(fname);
   if(buffer[0] != '\0')
     send_reply_full(553, FALSE, FALSE, TRUE, buffer);

   closedir(dir);
   return;
 }
	   
/* EOF */
