#!/usr/bin/perl -w

##############################################################################
#                                                                            #
#  primeCHECK v2.2 - DIRECTORY DUPE CHECK (c) 2000 primemover                #
#                                                                            #
#  directory checker and newdir announcer for OpenFTPD v1.x                  #
#  primetools.pm is needed in the same path!                                 #
# ---------------------------------------------------------------------------#
#                                                                            #
#  This program is free software; you can redistribute it and/or modify      #
#  it under the terms of the GNU General Public License as published by      #
#  the Free Software Foundation; either version 2 of the License, or         #
#  (at your option) any later version.                                       #
#                                                                            #
##############################################################################



my $standarddir   = $ENV{'F4A_STD'};

# ----------------------------------------------------------------------------
#  the directory var/log/dupe file
# ----------------------------------------------------------------------------
my $directorylog = $standarddir."var/log/directory.log";


# ----------------------------------------------------------------------------
#  disable dupe and ban checks for siteops
# ----------------------------------------------------------------------------
my $siteop_disable = 0;


# ----------------------------------------------------------------------------
#  criterion what directories should be dupechecked (to prevent checking dirs
#  like "CD1" oder "Sample") by default at least one "-" is needed in the
#  directory name to be checked
# ----------------------------------------------------------------------------
my $dupecriterion = '\w+-\w+$';


##############################################################################
# end of configuration (don't edit below unless you know what you are doing  #
##############################################################################

use strict; # always
use Cwd;
use Time::Local;
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/";
use primetools;


my %dircfg = read_dircfg();

# ----------------------------------------------------------------------------
#  strdifftime($earlier, $later) - calc the diff between two times as string
# ----------------------------------------------------------------------------
sub strdifftime {
   my $time1 = $_[0]; my $time2 = $_[1];
   my $diffstr = "";  my $diff = $time2 - $time1;
   my $seconds = $diff % 60; $diff = ($diff - $seconds) / 60;
   my $minutes = $diff % 60; $diff = ($diff - $minutes) / 60;
   my $hours   = $diff % 24; $diff = ($diff - $hours) / 24;
   my $days    = $diff % 7; 
   my $weeks   = ($diff - $days) / 7;
   if ($weeks != 0) { $diffstr .= $weeks . " weeks "; }
   if ($days  != 0) { $diffstr .= $days . " days "; }
   if ($hours != 0 and $weeks == 0) { 
      $diffstr .= $hours . " hours "; 
   }
   if ($minutes != 0 and $days == 0 and $weeks == 0) { 
      $diffstr .= $minutes . " min "; 
   }
   if ($seconds != 0 and $hours == 0 and $days == 0) { 
      $diffstr .= $seconds . " sec "; 
   }
   return $diffstr;
}

# ----------------------------------------------------------------------------
#  is_siteop()
# ----------------------------------------------------------------------------
sub is_siteop {
   if (exists $ENV{F4A_CLASS} and $ENV{F4A_CLASS} =~ /^(s|o)$/) {
      return 1;
   }
   else { return 0; }      
}


# ----------------------------------------------------------------------------
#  add_to_dirlog() - add line to the directory log
# ----------------------------------------------------------------------------
sub add_to_dirlog {
   my ($username,$groupname,$relpath)=@_ or 
     die "Insufficient parameters for add_to_dirlog";
   my $section = "default";   
   if (exists $ENV{F4A_SECTION}) {
      $section = $ENV{F4A_SECTION};
   }
   if (exists $dircfg{show_newdirs} and $dircfg{show_newdirs} eq "yes") {
      if (-e $directorylog) {
         open(DIRLOG, ">> $directorylog") or die "ERROR: unable to open directory logfile.\n";
      }
      else {
         open(DIRLOG, "> $directorylog") or die "ERROR: unable to open directory logfile.\n";
      }
      printf DIRLOG "%s %-10s %-10s %-16s %s\n", timetostring(time), 
             $username, $groupname, $section, $relpath;
      close(DIRLOG);
   }      
}

# ----------------------------------------------------------------------------
#  readdirectory($dirname) - read directory listing to an array
# ----------------------------------------------------------------------------
sub readdirectory {
   opendir(DIR, $_[0]);
   my @files = grep {not /^\..*/} readdir(DIR);
   closedir(DIR);
   return @files;
}


###############################################################################
# main program                                                                #
###############################################################################


my $cwd = getcwd() or die "unable to get current working directory\n";

if ($#ARGV < 0) {
   print "primeDUPE - FTP4ALL Dupechecker (c) 2000 primemover\n";
   print "Syntax: primedirdupe.pl <directory>\n";
   exit 1;
}

# read the ftp4all username and groupname from environment variables
my ($username, $groupname);
if (exists $ENV{'F4A_USER'}) { $username = $ENV{'F4A_USER'}; }
else { $username = "unknown"; }
if (exists $ENV{'F4A_GROUP'}) { $groupname = $ENV{'F4A_GROUP'}; }
else { $groupname = "unknown"; }

my %registry = read_registry(); 

# path stuff
if (not exists $registry{basedir}) { die "unable to find basedir\n"; }
my $path = $registry{basedir} . '/' . $ARGV[0];
my $relpath = $ARGV[0];


my $topline   = " !C." . '-' x 72 . ".!0\n";
my $footer    = " !C\`" . '-' x 61 . "[DIRDUPE]--\'!0\n";
my $infoline  = sprintf(" !C|!3!H Directory: %-59.59s !0!C|!0\n", $relpath);
my $limiter   = " !C|" . '-' x 72 . "|!0\n";
my $header    = $topline . $infoline . $limiter;
my $message;

printcolor($header);

# dupecheck disabled for siteops
if (is_siteop() and $siteop_disable == 1) {
   $message = " !C|!H Dupecheck was disabled for siteops." . ' ' x 36 . "!C|!0\n" .
                 " !C|!c ==> !Hdirectory was successfully approved" . ' ' x 32 . "!C|!0\n";
   printcolor($message . $footer);
   add_to_dirlog($username,$groupname,$relpath);
   exit 0;
}

# check for unwanted directories (set in dircfg)
if (exists $dircfg{deny_dirs}) {
   my @unwanted = split(/\s/, $dircfg{deny_dirs});
   foreach my $item (@unwanted) {
      if ($relpath =~ /$item/i) {
         $message = " !C| !BWARNING!0: this directory is unwanted (banned keyword detected)" . ' ' x 10 . "!C|!0\n" .
                    " !C|!c ==> !Bdirectory was NOT approved!0" . ' ' x 41 . "!C|!0\n" .
                    " !C|       banned word: !B $item !0\n";
         printcolor($message . $footer);
	 exit 1;
      }	 
   }
}

if (exists $dircfg{allow_dirs}) {
   my @wanted = split(/\s/, $dircfg{allow_dirs});
   foreach my $item (@wanted) {
      if ($relpath !~ /$item/i) {
         $message = " !C| !BWARNING: this directory does not match the accept list." . 
	              ' ' x 16 . "!C|!0\n" .
                      " !C|!c ==> !Bdirectory was NOT approved!0" . ' ' x 41 . "!C|!0\n"; 
         printcolor($message . $footer);
	 exit 1;
      }	 
   }
}


my $dirname = basename($path);

# dupecheck disabled for directory
if (exists $dircfg{dir_dupecheck} and $dircfg{dir_dupecheck} ne "yes") {
   $message = " !C|!H Dupecheck was disabled for this directory." . ' ' x 29 . "!C|!0\n" .
              " !C|!c ==> !Hdirectory was successfully approved" . ' ' x 32 . "!C|!0\n";
   printcolor($message . $footer);
   add_to_dirlog($username, $groupname, $relpath);
   exit 0;
}

# case insensitive search in the dir
my $parentdir = dirname($path);
my @files=readdirectory($parentdir);
foreach my $file (@files) {
   if ($dirname =~ /^\Q$file\E$/i) {
       $message = " !C|!H Directory already exists in this dir." . ' ' x 34 . "!C|!0\n" .  
                  " !C|!c ==> !Bdirectory was NOT approved!0" . ' ' x 41 . "!C|!0\n"; 
       printcolor($message . $footer);
       exit 1;
   }
}

if ($dirname !~ /$dupecriterion/) {
   $message = " !C|!H Directory name does not match the dupecheck criterion (too common).    !C|!0\n";
   printcolor($message);
}
else {
   open(DIRLOG, "< $directorylog") or exit 0;
   while (<DIRLOG>) {
      if (/(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d) ([\w\-_]+) +([\w\-_]+) +(.+)/) {
         my $loggeddir = $9; my $usr = $7; my $grp = $8;
         if (basename($loggeddir) =~ /\Q$dirname\E/i) {
	    my $dupetime = stringtotime($_);
            my $diffstr = strdifftime($dupetime, time);
            $message = sprintf(" !C| %-84s !C|!0\n", 
  	       "!BDUPE FOUND: !hcreated !H$diffstr!hago by !H$usr !hof !H$grp.");
            printcolor($message);
            my $p = $registry{basedir} . '/' . $loggeddir;
	    if ( -d $p) {
               $message = sprintf(" !C| %-72.72s !C|!0\n", ' ' x 12 . 
                          'directory already exists in !H' . 
			  dirname($loggeddir));
	       printcolor($message);
            }
            $message = " !C|!c ==> !Bdirectory was NOT approved!0" . ' ' x 41 . "!C|!0\n"; 
            printcolor($message . $footer);
            exit 1;
         }      
      }
   }
   close(DIRLOG);
}

add_to_dirlog($username, $groupname,$relpath);

$message = " !C|!c ==> !Hdirectory was successfully approved!0" . ' ' x 32 . "!C|!0\n";
printcolor($message . $footer);
exit 0;

# --- end of file ---
